// //////////////////////////////////////////////////////////////////////////
//
//    map3d.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ////////////////////////////////////////////////////////////////////////////

#ifndef _MAP3D_H_
#define _MAP3D_H_ 1

#include <fftw3.h>

extern "C" {
	#include "spline.h"
}
#include "prob_dist.h"
#include "vec3.h"

// supported interpolation functions
enum InterpolationType {NEAREST, LINEAR, FFT, SPLINE2, SPLINE3, SPLINE4, SPLINE5};

template <class T> class map_3d;

////////////////////////////
//
// map_3d template class
//
////////////////////////////
template <class T>
class map_3d {
public:
	map_3d() {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(0,0,0);
	}
	map_3d(int x, int y, int z) {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(x,y,z);
	}
	map_3d(vec3<int> X) {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(X[0],X[1],X[2]);
	}
	map_3d(int x, int y, int z, T val) {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(x,y,z);
		setAll(val);
	}
	map_3d(vec3<int> X, T val) {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(X[0],X[1],X[2]);
		setAll(val);
	}
	~map_3d() {
		if (data) delete[] data; data = NULL;
		if (spline_coeffs) delete[] spline_coeffs; spline_coeffs = NULL;
	}

	map_3d(const map_3d<T>& X) {
		data = NULL; spline_coeffs = NULL;
		dims = vec3<int>(0,0,0);
		alloc(X.dims[0],X.dims[1],X.dims[2]);
		memcpy(data, X.data, dims[0]*dims[1]*dims[2]*sizeof(T));

		// copy spline coeffs
		spline_order = X.spline_order;
		if (X.spline_coeffs) {
			spline_coeffs = new double[dims[0]*dims[1]*dims[2]];
			memcpy(spline_coeffs, X.spline_coeffs, dims[0]*dims[1]*dims[2]*sizeof(double));
		}
	}
	map_3d<T>& operator=(const map_3d<T>& X) {
		if (dims[0] != X.dims[0] || dims[1] != X.dims[1] || dims[2] != X.dims[2]) {
			if (data) {free(data);data=NULL;}
			alloc(X.dims[0],X.dims[1],X.dims[2]);
		}
		memcpy(data, X.data, dims[0]*dims[1]*dims[2]*sizeof(T));

		// copy spline coeffs
		spline_order = X.spline_order;
		if (X.spline_coeffs) {
			if (spline_coeffs) {free(spline_coeffs);spline_coeffs=NULL;}
			spline_coeffs = new double[dims[0]*dims[1]*dims[2]];
			memcpy(spline_coeffs, X.spline_coeffs, dims[0]*dims[1]*dims[2]*sizeof(double));
		}
		
		return *this;
	}
	
	// (re)allocate storage space for this map
	void alloc(vec3<int> X) {
		alloc(X[0],X[1],X[2]);
	}
	void alloc(int x, int y, int z) {
		if (x != dims[0] || y != dims[1] || z != dims[2]) {   // only reallocate if necessary
			if (data) {
				delete data;   // reallocate space
				data = NULL;
			}
			if (x*y*z > 0) {
				data = new T[x*y*z];
				dims[0]=x;dims[1]=y;dims[2]=z;


				if (data == NULL)
					cerr << "ERROR: map_3d::alloc() returned NULL\n";
			} else {
				dims[0]=dims[1]=dims[2]=0;
			}
		}
		dimsProd = dims[0]*dims[1]*dims[2];
	}
	void reset() {
		if (spline_coeffs) delete[] spline_coeffs;
		spline_coeffs = NULL;
	}
	void clear() {
		alloc(0,0,0);
	}

	//
	// dumpDataToDisk
	//      make a raw dump of 'data' to disk, freeing up memory
	//
	void dumpDataToDisk(string filename) {
		ofstream out;
		out.open(filename.c_str(), ios::out);
		out.write(data, sizeof(T)*dims[0]*dims[1]*dims[2]);
		delete[] data; data = NULL;
	}

	//
	// loadDataFromDisk
	//      read in a previous dump of 'data'
	//
	void loadDataFromDisk(string filename) {
		ifstream in;
		in.open(filename.c_str(), ios::in);
		data = new T[dims[0]*dims[1]*dims[2]];
		in.read((char*)data, sizeof(T)*dims[0]*dims[1]*dims[2]);
	}

	//
	// read this (matlab-format) map from disk
	//
	bool readFromFile(string infile) {
		ifstream in;
		in.open(infile.c_str(), ios::in);

		if (!in.is_open())  {
			// cerr << "ERROR: Unable to open match map '" << infile << "'\n";
			return false;
		}

		// extract relevant bytes from header....assume that file was written
		//      on machine with same endianness as the one on which it is being read
		int ni,nj,nk;
		in.seekg(160);
		in.read((char*)&ni, 4);
		in.read((char*)&nj, 4);
		in.read((char*)&nk, 4);
		in.seekg(28, ios::cur);

		alloc(ni,nj,nk);

		float buff_f;
		for (int k=0; k<dims[2]; k++) {
			for (int j=0; j<dims[1]; j++) {
				for (int i=0; i<dims[0]; i++) {
					in.read((char*)&buff_f, 4);
					data[i*dims[1]*dims[2] + j*dims[2] + k] = buff_f;
				}
			}
		}

		in.close();
		return true;

	}

	//
	// map statistics
	//
	void testIfContainsNaN(string id) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			if (isnan(data[i])) {
				vec3<int> i_xyz;
				i_xyz[0] =   i / (dims[1]*dims[2]);
				i_xyz[1] = ( i / dims[2] ) % dims[1];
				i_xyz[2] = ( i % dims[2] );

				cerr << i_xyz.to_string() << " (" << data[i] << ")\n";
				exit(1);
			}
	}
	T sum() {
		T retval=0;
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			retval += data[i];
		return (retval);
	}
	T sum2() {
		T retval=0;
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			retval += data[i]*data[i];
		return (retval);
	}
	
	// mean/var
	T mean() {
		T sum=0;
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			sum += data[i];
		}
		return ( sum / (double)N );
	}
	T variance() {
		T sum=0, sum2=0;
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			sum += data[i];
			sum2 += data[i]*data[i];
		}
		double N2 = square((double)N);
		return ( (N*sum2-sum*sum) /  (N2) );
	}
	
	// mean/var masked
	template <class S>
	T mean(map_3d<S> &mask) {
		T sum=0;
		int N = dims[0]*dims[1]*dims[2], nMask=0;
		for (int i=0; i<N; i++) {
			if (mask.get(i) > (S)0) {
				sum += data[i];
				nMask++;
			}
		}
		return ( sum / (double)nMask );
	}
	template <class S>
	T variance(map_3d<S> &mask) {
		T sum=0, sum2=0;
		int N = dims[0]*dims[1]*dims[2], nMask=0;;
		for (int i=0; i<N; i++) {
			if (mask.get(i) > (S)0) {
				sum += data[i];
				sum2 += data[i]*data[i];
				nMask++;
			}
		}
		double N2 = square((double)nMask);
		return ( (nMask*sum2-sum*sum) /  (N2) );
	}
		
	void range(T& minX, T& maxX) {
		minX = maxX = data[0];
		int N = dims[0]*dims[1]*dims[2];
		for (int i=1; i<N; i++) {
			minX = min(minX,data[i]);
			maxX = max(maxX,data[i]);
		}
	}
	void argmax(T& maxX, vec3<int> &argmaxX) {
		maxX = data[0];
		int idxMax = 0;

		int N = dims[0]*dims[1]*dims[2];
		for (int i=1; i<N; i++) {
			if (maxX < data[i]) {
				maxX = data[i];
				idxMax = i;
			}
		}
		argmaxX[0] =   idxMax / (dims[1]*dims[2]);
		argmaxX[1] = ( idxMax / dims[2] ) % dims[1];
		argmaxX[2] = ( idxMax % dims[2] );
	}
	T getMax() {
		T maxX = data[0];

		int N = dims[0]*dims[1]*dims[2];
		for (int i=1; i<N; i++) {
			if (maxX < data[i]) {
				maxX = data[i];
			}
		}
		return maxX;
	}
	T getMin() {
		T minX = data[0];

		int N = dims[0]*dims[1]*dims[2];
		for (int i=1; i<N; i++) {
			if (minX > data[i]) {
				minX = data[i];
			}
		}
		return minX;
	}
	
	HistogramDist getDistribution(int M=1) {
		T minX, maxX;
		range(minX, maxX);

		// add some slack
		T delta = maxX-minX;
		minX -= delta/2;
		maxX += delta/2;

		HistogramDist retval;
		retval.setRange(minX , maxX, 1000, M);

		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			retval.addObservation(data[i]);

		return retval;
	}
	
	HistogramDist getDistribution(map_3d<double> &mask, int M=1) {
		T minX, maxX;
		
		int i=0;
		do {
			minX = maxX = data[i];
		} while (mask.get(i) == 0);
		
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			if (mask.get(i) != 0) {
				minX = min( minX , data[i] );
				maxX = max( maxX , data[i] );
			}
		}

		// add some slack
		T delta = maxX-minX;
		minX -= delta/2;
		maxX += delta/2;

		HistogramDist retval;
		retval.setRange(minX , maxX, 1000, M);

		for (int i=0; i<N; i++)
			if (mask.get(i) != 0) 
				retval.addObservation(data[i]);

		return retval;
	}
	
	// find the maximum belief constrained to ~ 3.8A from 'adjacent'
	void argmax_constrained(T& maxX, vec3<int> &argmaxX, vec3<double> &f_coord, mat33<double> f2c, map_3d<bool> &mask) {
		maxX = -INFINITY;
		vec3<double> del_f;
		double fi,fj,fk, dist_to_frag2;
		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(f2c).norm2() )
		);

		for (int i=0; i<dims[0]; i++) {
			fi = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				fj = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					fk = (double)k/dims[2];

					if (!mask.get(i,j,k)) {
						// for each atom in the fragment ...
						del_f[0] = fi-f_coord[0];
						del_f[1] = fj-f_coord[1];
						del_f[2] = fk-f_coord[2];
						del_f.minmod1();

						dist_to_frag2 = del_f.apply(f2c).norm2();

						if (dist_to_frag2 < square(3.8+K) && dist_to_frag2 > square(3.8-K)) {
							if (maxX < get(i,j,k)) {
								maxX = get(i,j,k);
								argmaxX = vec3<int>(i,j,k);
							}
						}
					}
				}
			}
		}
	}

	// find the maximum belief constrained to ~ 3.8A from 'adjacent'
	void argmax_constrained(T& maxX, vec3<int> &argmaxX, map_3d<bool> &mask) {
		maxX = -INFINITY;

		for (int i=0; i<dims[0]; i++) {
			for (int j=0; j<dims[1]; j++) {
				for (int k=0; k<dims[2]; k++) {
					if (!mask.get(i,j,k)) {
						if (maxX < get(i,j,k)) {
							maxX = get(i,j,k);
							argmaxX = vec3<int>(i,j,k);
						}
					}
				}
			}
		}
	}

	//
	// map transformation functions
	//
	void sqrtMap() {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] = sqrt(data[i]);
	}
	void invMap() {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] = 1/(data[i]);
	}
	
	void squareIntensities() {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			if (data[i] > 0) 
				data[i] = square(data[i]);
			else
				data[i] = -square(data[i]);
	}


	//
	// normalize this map so it sums to 1
	void normalize() {
		double dataSum=0;
		int N = dims[0]*dims[1]*dims[2];

		for (int i=0; i<N; i++) {
			//data[i] = max(data[i],0.0);
			dataSum += data[i];
		}
		for (int i=0; i<N; i++)
			data[i] /= dataSum;
	}

	// normalize map
	void normalize(double Mest) {
		double dataSum=0;
		int N = dims[0]*dims[1]*dims[2];

		for (int i=0; i<N; i++) {
			//data[i] = max(data[i],(T)0);
			dataSum += data[i];
		}

		double Mperpoint = Mest / N;
		for (int i=0; i<N; i++)
			data[i] = (1-Mest) * data[i]/dataSum + Mperpoint;
	}
	void normalize(double Mest, map_3d<double> &mask) {
		double dataSum=0;
		int N = dims[0]*dims[1]*dims[2];

		for (int i=0; i<N; i++) {
			if (mask.get(i) > 0) {
				//data[i] = max(data[i],(T)0);
				dataSum += data[i];
			}
		}

		double Mperpoint = Mest / N;
		for (int i=0; i<N; i++) {
			if (mask.get(i) > 0) {
				data[i] = (1-Mest) * data[i]/dataSum + Mperpoint;
			} else {
				data[i] = 0;
			}
		}
	}

	//
	// normalize this (log-space) map so its exponential sums to 1
	//
	void logNormalize() {   // normalize a log-transformed distribution
		int N = dims[0]*dims[1]*dims[2];
		double maxValue=data[0];
		for (int i=1; i<N; i++)
			maxValue = max(data[i],maxValue);

		// shift data by maxValue (==> multiplying by a constant to prevent underflow)
		// then take exp and sum
		double sum = 0;
		for (int i=0; i<N; i++) {
			data[i] -= maxValue;
			sum += exp(data[i]);
		}

		// now divide by sum (==> subtract log sum)
		sum = log(sum);
		for (int i=0; i<N; i++)
			data[i] -= sum;
	}

	// normalize log map, ensuring no value goes below log(Mest)
	void logNormalize(double Mest) {   // normalize a log-transformed distribution
		int N = dims[0]*dims[1]*dims[2];
		double maxValue=data[0];
		for (int i=1; i<N; i++)
			maxValue = max((T)data[i],(T)maxValue);

		// shift data by maxValue (==> multiplying by a constant to prevent underflow)
		// then take exp and sum
		double sum = 0;
		for (int i=0; i<N; i++) {
			data[i] = exp(data[i] - maxValue);
			sum += data[i];
		}

		double Mperpoint = Mest / N;
		for (int i=0; i<N; i++)
			data[i] = log( (1-Mest)*data[i] / sum + Mperpoint );
	}
	void logNormalize(double Mest, map_3d<double> &mask) {   // normalize a log-transformed distribution
		int N = dims[0]*dims[1]*dims[2];
		double maxValue=data[0];
		for (int i=1; i<N; i++)
			maxValue = max((T)data[i],(T)maxValue);

		// shift data by maxValue (==> multiplying by a constant to prevent underflow)
		// then take exp and sum
		double sum = 0;
		for (int i=0; i<N; i++) {
			if (mask.get(i) > 0) {
				data[i] = exp(data[i] - maxValue);
				sum += data[i];
			}
		}

		double Mperpoint = Mest / N;
		for (int i=0; i<N; i++) {
			if (mask.get(i) > 0) {
				data[i] = log( (1-Mest) * data[i]/sum + Mperpoint );
			} else {
				data[i] = -1000;
			}
		}
	}


	// transform to and from log space
	void log_transform() {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			data[i] = log(data[i]);
		}
	}
	void log_transform(map_3d<double> &mask) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			if (mask.get(i) > 0) 
				data[i] = log(data[i]);
			else 
				data[i] = -1000;
		}
	}
	void invlog_transform() {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] = exp(data[i]);
	}
	void invlog_transform(map_3d<double> &mask) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			if (mask.get(i) > 0)
				data[i] = exp(data[i]);
			else 
				data[i] = 0;
	}
	
	// build a map from a subset of this array's rows columns and planes
	void subarray(map_3d<T>& proj, vector<int> xs, vector<int> ys, vector<int> zs) {
		proj.alloc(xs.size(), ys.size(), zs.size());

		for (int i=0; i<xs.size(); i++)
			for (int j=0; j<ys.size(); j++)
				for (int k=0; k<zs.size(); k++)
					proj.set(i,j,k, data[xs[i]*dims[1]*dims[2] + ys[j]*dims[2] + zs[k]]);
	}

	// project the values of the points in 'mask'
	template <class U>
	vector<T> project(map_3d<U>& mask) {
		vector<T> retval;
		retval.reserve(mask.numNonZero());
		int N = dims[0]*dims[1]*dims[2];

		for (int i=0; i<N; i++)
			if (mask.get(i) != (U)0) retval.push_back(data[i]);
		return retval;
	}

	template <class U>
	vector<T> project(map_3d<U>& mask, vec3<int> center) {
		vector<T> retval;
		retval.reserve(mask.numNonZero());
		int N = dims[0]*dims[1]*dims[2];

		for (int i=0; i<dims[0]; i++)
		for (int j=0; j<dims[1]; j++)
		for (int k=0; k<dims[2]; k++)
			if (mask.get(i) != (U)0) retval.push_back( get(
			                                 pos_mod(center[0]+i,dims[0]) ,
			                                 pos_mod(center[1]+j,dims[1]) ,
			                                 pos_mod(center[2]+k,dims[2]) ) );
		return retval;
	}


	// set to 1 all points > dist angstroms from the origin
	void mask_distance(mat33<double>& f2c, double dist) {
		double dist2 = dist*dist;
		double fi,fj,fk;  // frac coords

		for (int i=0; i<dims[0]; i++) {
			fi = (double)i/dims[0];  if (fi>.5) fi -=1;
			for (int j=0; j<dims[1]; j++) {
				fj = (double)j/dims[1];  if (fj>.5) fj -=1;
				for (int k=0; k<dims[2]; k++) {
					fk = (double)k/dims[2];  if (fk>.5) fk -=1;

					set(i,j,k,(T)(vec3<double>(fi,fj,fk).apply(f2c).norm2() <= dist2));
				}
			}
		}
	}

	// set to 1 all points within 'dist' of an atom in 'coords'
	void mask_fragment( vector< vec3<double> > coords, mat33<double>& f2c, mat33<double>& c2f, double dist, map_3d<bool>* rot_mask = NULL) {
		double dist2 = dist*dist;
		int N = coords.size();

		double fi,fj,fk,del_fi,del_fj,del_fk;   // frac coords
		double dist_to_frag2;
		vector< vec3<double> > f_coords(N);

		// compute fractional coords of the fragment
		for (int atm=0; atm<N; atm++) {
			f_coords[atm] = coords[atm].apply(c2f);
			f_coords[atm][0] = pos_mod(f_coords[atm][0],1.0);
			f_coords[atm][1] = pos_mod(f_coords[atm][1],1.0);
			f_coords[atm][2] = pos_mod(f_coords[atm][2],1.0);
		}

		for (int i=0; i<dims[0]; i++) {
			fi = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				fj = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					fk = (double)k/dims[2];

					// for each atom in the fragment ...
					set(i,j,k,(T)0);
					if (!rot_mask || rot_mask->get(i,j,k)) {
						for (int atm=0; atm<N; atm++) {
							del_fi = fi-f_coords[atm][0]; if (del_fi>.5) del_fi -=1; if (del_fi<-.5) del_fi +=1;
							del_fj = fj-f_coords[atm][1]; if (del_fj>.5) del_fj -=1; if (del_fj<-.5) del_fj +=1;
							del_fk = fk-f_coords[atm][2]; if (del_fk>.5) del_fk -=1; if (del_fk<-.5) del_fk +=1;

							dist_to_frag2 = vec3<double>(del_fi,del_fj,del_fk).apply(f2c).norm2();

							if (dist_to_frag2 < dist2) {
								set(i,j,k,(T)1);
								//break;   // no need to look at other atoms
							}
						}
					}

				}
			}
		}
	}
	
	
	// set to 1 all points closer to 'include' than 'exclude'
	void mask_fragment( vector< vec3<double> > include, vector< vec3<double> > exclude, mat33<double>& f2c, mat33<double>& c2f) {
		int Ninc = include.size();
		int Nexc = exclude.size();
		
		cout << "include " << Ninc << "   ; exclude " << Nexc << "\n";

		vec3<double> f,del_f;   // frac coords
		double dist_to_inc2;
		double dist_to_exc2;
		
		vector< vec3<double> > f_inc(Ninc);
		vector< vec3<double> > f_exc(Nexc);

		// compute fractional coords of the fragments
		for (int atm=0; atm<Ninc; atm++) {
			f_inc[atm] = include[atm].apply(c2f);
			f_inc[atm][0] = pos_mod(f_inc[atm][0],1.0);
			f_inc[atm][1] = pos_mod(f_inc[atm][1],1.0);
			f_inc[atm][2] = pos_mod(f_inc[atm][2],1.0);
		}
		for (int atm=0; atm<Nexc; atm++) {
			f_exc[atm] = exclude[atm].apply(c2f);
			f_exc[atm][0] = pos_mod(f_exc[atm][0],1.0);
			f_exc[atm][1] = pos_mod(f_exc[atm][1],1.0);
			f_exc[atm][2] = pos_mod(f_exc[atm][2],1.0);
		}

		for (int i=0; i<dims[0]; i++) {
			f[0] = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				f[1] = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					f[2] = (double)k/dims[2];

					// for each atom in the fragment ...
					set(i,j,k,(T)0);
					dist_to_inc2 = dist_to_exc2 = INFINITY;
					
					for (int atm=0; atm<Ninc; atm++) {
						del_f = f-f_inc[atm]; del_f.minmod1();
						dist_to_inc2 = min(dist_to_inc2 , del_f.apply(f2c).norm2());
					}
					for (int atm=0; atm<Nexc; atm++) {
						del_f = f-f_exc[atm]; del_f.minmod1();
						dist_to_exc2 = min(dist_to_exc2 , del_f.apply(f2c).norm2());
					}
					
					if (dist_to_inc2 <= dist_to_exc2+1e-6 && dist_to_inc2 < 36) {
						//cout << sqrt(dist_to_inc2) << "," << sqrt(dist_to_exc2) << "\n";
						set(i,j,k,(T)1);
					}
				}
			}
		}
	}
	
	// set to 1 all points within 'dist' of an atom in 'coords'
	void or_mask( vector< vec3<double> > &f_coords, mat33<double>& f2c, double dist) {
		double dist2 = dist*dist;
		int nsymm = f_coords.size();

		double fi,fj,fk;   // frac coords
		vec3<double> del_f;
		double dist_to_frag2;

		for (int i=0; i<dims[0]; i++) {
			fi = (double)i/dims[0];
			for (int j=0; j<dims[1]; j++) {
				fj = (double)j/dims[1];
				for (int k=0; k<dims[2]; k++) {
					fk = (double)k/dims[2];

					// for each atom in the fragment ...
					for (int symm_idx = 0; symm_idx < nsymm; symm_idx++) {
						del_f[0] = fi-f_coords[symm_idx][0];
						del_f[1] = fj-f_coords[symm_idx][1];
						del_f[2] = fk-f_coords[symm_idx][2];
						del_f.minmod1();

						dist_to_frag2 = del_f.apply(f2c).norm2();

						if (dist_to_frag2 < dist2) {
							set(i,j,k,(T)1);
							//break;   // no need to look at other atoms
						}
					}
				}
			}
		}
	}

	// get a neighborhood centered on a point
	void get_neighborhood( map_3d<T> &newMap, vec3<double> pt, vec3<double> newCrystSize, mat33<double> c2f) {
		vec3<int> newDims = newMap.getDims();
		vec3<double> x, fX;

		computeSplineCoeffs(3);

		for (int i=0; i<newDims[0]; i++)
		for (int j=0; j<newDims[1]; j++)
		for (int k=0; k<newDims[2]; k++) {
			fX[0] = ((double)i)/newDims[0];  if (fX[0]>.5) fX[0] -=1;
			fX[1] = ((double)j)/newDims[1];  if (fX[1]>.5) fX[1] -=1;
			fX[2] = ((double)k)/newDims[2];  if (fX[2]>.5) fX[2] -=1;

			x[0] = pt[0] + fX[0]*newCrystSize[0]; x[1] = pt[1] + fX[1]*newCrystSize[1]; x[2] = pt[2] + fX[2]*newCrystSize[2];
			x = x.apply(c2f);
			x.mod(1);
			newMap.set(i,j,k, interp_spline(x));
		}
	}

	// interpolate the map at an arbitrary point (given in fractional coords)
	inline T interp_linear( vec3<double> pt ) {
		pt[0] *= dims[0]; pt[1] *= dims[1]; pt[2] *= dims[2];

		// out of bounds?
		//if (pt[0] < 0 || pt[1] < 0 || pt[2] < 0 ||
		//    pt[0] > (dims[0]-1) || pt[1] > (dims[1]-1) || pt[2] > (dims[2]-1) )
		//    return ((T) 0);

		int pt000[3], pt111[3];
		double fpart[3],neg_fpart[3];

		pt000[0] = pos_mod(int(floor(pt[0])),dims[0]);
		pt000[1] = pos_mod(int(floor(pt[1])),dims[1]);
		pt000[2] = pos_mod(int(floor(pt[2])),dims[2]);

		pt111[0] = (pt000[0]+1); if (pt111[0]==dims[0]) pt111[0] = 0;
		pt111[1] = (pt000[1]+1); if (pt111[1]==dims[1]) pt111[1] = 0;
		pt111[2] = (pt000[2]+1); if (pt111[2]==dims[2]) pt111[2] = 0;

		fpart[0] = pt[0]-floor(pt[0]); neg_fpart[0] = 1-fpart[0];
		fpart[1] = pt[1]-floor(pt[1]); neg_fpart[1] = 1-fpart[1];
		fpart[2] = pt[2]-floor(pt[2]); neg_fpart[2] = 1-fpart[2];

		//cout << fpart.to_string() << "   " << neg_fpart.to_string() << "\n";

		// trilinear interpolation
		T retval = neg_fpart[0]*neg_fpart[1]*neg_fpart[2]*get(pt000[0],pt000[1],pt000[2]) +
		           neg_fpart[0]*neg_fpart[1]*    fpart[2]*get(pt000[0],pt000[1],pt111[2]) +
		           neg_fpart[0]*    fpart[1]*neg_fpart[2]*get(pt000[0],pt111[1],pt000[2]) +
		           neg_fpart[0]*    fpart[1]*    fpart[2]*get(pt000[0],pt111[1],pt111[2]) +
		               fpart[0]*neg_fpart[1]*neg_fpart[2]*get(pt111[0],pt000[1],pt000[2]) +
		               fpart[0]*neg_fpart[1]*    fpart[2]*get(pt111[0],pt000[1],pt111[2]) +
		               fpart[0]*    fpart[1]*neg_fpart[2]*get(pt111[0],pt111[1],pt000[2]) +
		               fpart[0]*    fpart[1]*    fpart[2]*get(pt111[0],pt111[1],pt111[2]);

		//printf("linear interp: f(%8.4f,%8.4f,%8.4f) = %f\n", pt[0], pt[1], pt[2], retval);

		return retval;
	}

	inline T interp_spline( vec3<double> pt ) {
		pt[0] *= dims[0]; pt[1] *= dims[1]; pt[2] *= dims[2];

		// compute spline coeffs if necessary
		//if (spline_coeffs == NULL)
		//	computeSplineCoeffs(order);

		// call spline lib
		T retval = (T)interp3(spline_coeffs, dims.c_ptr(), pt.c_ptr(), spline_order);

		//if (isnan(retval)) {
		//	cerr << "spline interp: f(" << pt[0] << ", " << pt[1] << ", " << pt[2] << ") = " << retval << endl;
		//}
		
		return retval;
	}
	
	inline vec3<double> interp_grad( vec3<double> pt ) {
		pt[0] *= dims[0]; pt[1] *= dims[1]; pt[2] *= dims[2];

		// compute spline coeffs if necessary
		//if (spline_coeffs == NULL)
		//	computeSplineCoeffs(order);

		// call spline lib
		double retval[3];
		grad3(retval,spline_coeffs, dims.c_ptr(), pt.c_ptr(), spline_order);

		return vec3<double>(retval);
	}

	inline T interp_nearest( vec3<double> pt ) {
		return get(pos_mod(int(floor(pt[0]*dims[0]+.5)),dims[0]),
		           pos_mod(int(floor(pt[1]*dims[1]+.5)),dims[1]),
		           pos_mod(int(floor(pt[2]*dims[2]+.5)),dims[2]));
	}

	void computeSplineCoeffs(int order) {
		if (spline_coeffs)
			if (order == spline_order)
				return;                // already computed
			else
				delete spline_coeffs;  // recomputing

		// allocate space for the coefficients
		spline_coeffs = new double[dims[0]*dims[1]*dims[2]];
		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++) {
			spline_coeffs[i] = (double)data[i];
		}

		//clock_t S = clock();
		compute_coefficients(spline_coeffs, dims.c_ptr(), order);
		spline_order = order;

		//printf("\nmap_3d::computeSplineCoeffs(%d) took %f s\n", order, (double)(clock()-S)/CLOCKS_PER_SEC);
	}

	// number of non-zero elements in the map
	int numNonZero() {
		int retval = 0;
		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++) {
			if (data[i] != 0) {
				retval++;
			}
		}
		return retval;
	}

	// rescale the map s.t. inputs are mapped to -1 and +1, respectively
	void rescale(double negOneVal, double posOneVal) {
		int N = dims[0]*dims[1]*dims[2];
		double shift = -.5*(negOneVal+posOneVal),
		       scale =  .5*(posOneVal-negOneVal);
		for (int i=0; i<N; i++)
			data[i] = (data[i]+shift)/scale;
	}

	// multiply map by a constant factor
	void scale(double factor) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] *= factor;
	}

	// multiply map by a constant factor
	void maxEquals(T maxVal) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) data[i] = min(data[i], maxVal);
	}

	// multiply map by a constant factor, then shift by shift
	void scaleShift(double factor, T shift) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] = data[i]*factor + shift;
	}

	// set this map to the min of this map and passed-in map
	void minAggregate(map_3d<T>& X) {
		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++)
			data[i] = min(data[i],X.get(i));
	}

	// transformation / aggregation based on probabilities
	//     - given two gaussian distributions, compute the probability
	//       that input scores were drawn from the first
	//     - take the pointwise sum of probabilities (disjoint-OR)
	void probTransform(ProbDist &X_a, ProbDist &X_b) {
		int N = dims[0]*dims[1]*dims[2];
		T probA, probB;
		for (int i=0; i<N; i++) {
			probA = X_a.getPDF(data[i]);
			probB = X_b.getPDF(data[i]);

			data[i] = probA/probB;
		}
	}
	void probAggregate(map_3d<T>& X, double weight, ProbDist &X_a, ProbDist &X_b) {
		int N = dims[0]*dims[1]*dims[2];
		T probA, probB;
		for (int i=0; i<N; i++) {
			probA = X_a.getPDF(X.get(i));
			probB = X_b.getPDF(X.get(i));
			data[i] += weight * probA/(probA+probB);
		}
	}

	// compute the rank of a point
	int pointRank( vec3<double> pt , T &ptMaxValue) {
		pt[0] *= dims[0]; pt[1] *= dims[1]; pt[2] *= dims[2];

		// out of bounds?
		//if (pt[0] < 0 || pt[1] < 0 || pt[2] < 0 ||
		//    pt[0] > (dims[0]-1) || pt[1] > (dims[1]-1) || pt[2] > (dims[2]-1) )
		//    return ((T) 0);

		int pt000[3];
		int rank = 0;

		pt000[0] = pos_mod(int(floor(pt[0]+.5)),dims[0]);
		pt000[1] = pos_mod(int(floor(pt[1]+.5)),dims[1]);
		pt000[2] = pos_mod(int(floor(pt[2]+.5)),dims[2]);

		ptMaxValue = get(pt000[0],pt000[1],pt000[2]);
		
		//cout << setw(3) << pt000[0] << " , " << setw(3) << pt000[1] << " , " << setw(3) << pt000[2] << " , ";

		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++)
			if (ptMaxValue <= data[i]) rank++;   // count ties

		return rank;
	}

	int pointRank( vec3<int> pt) {
		T pt_value = get(pt[0],pt[1],pt[2]);
		int rank = 0;
		for (int i=0; i<dims[0]*dims[1]*dims[2]; i++)
			if (pt_value <= data[i]) rank++;     // count ties
		return rank;
	}

	//
	// accessor functions
	//
	inline T& get(int x, int y, int z) const { 
		//if (x < 0 || y < 0 || y < 0 || x >= dims[0] || y >= dims[1] || z >= dims[2] ) {
		//	cerr << "FATAL ERROR! get (" << x << "," << y << "," << z << ")\n";
		//	return (data[0]);
		//}
		return (data[x*dims[1]*dims[2] + y*dims[2] + z]); 
	}
	inline T& get(int xyz)             const { 
		//if (xyz < 0 || xyz >= dims[0]*dims[1]*dims[2] ) {
		//	cerr << "FATAL ERROR! get (" << xyz << ")\n";
		//	return (data[0]);
		//}		
		return (data[xyz]); 
	}
	inline void set(int x, int y, int z, T val) {
		//if (x < 0 || y < 0 || y < 0 || x >= dims[0] || y >= dims[1] || z >= dims[2] ) {
		//	cerr << "FATAL ERROR! set (" << x << "," << y << "," << z << ")\n";
		//	return;
		//}
		data[x*dims[1]*dims[2] + y*dims[2] + z] = val; 
	}
	inline void set(int xyz, T val)             { 
		//if (xyz < 0 || xyz >= dims[0]*dims[1]*dims[2] ) {
		//	cerr << "FATAL ERROR! set (" << xyz << ")\n";
		//	return;
		//}	
		data[xyz] = val; 
	}
	inline void inc(int x, int y, int z, T val) { 
		//if (x < 0 || y < 0 || y < 0 || x >= dims[0] || y >= dims[1] || z >= dims[2] ) {
		//	cerr << "FATAL ERROR! inc (" << x << "," << y << "," << z << ")\n";
		//	return;
		//}
		data[x*dims[1]*dims[2] + y*dims[2] + z] += val; 
	}
	inline void inc(int xyz, T val)             { 
		//if (xyz < 0 || xyz >= dims[0]*dims[1]*dims[2] ) {
		//	cerr << "FATAL ERROR! set (" << xyz << ")\n";
		//	return;
		//}	
		data[xyz] += val;
	}

	void setAll(T val) {
		for (int i=0; i < dims[0]*dims[1]*dims[2]; i++) data[i] = val;
	}

	vec3<int> getDims() { return dims; }
	int getDims(int i)  { return dims[i]; }
	int getDimsProd()   { return dimsProd; }

	// debugging info
	void debug() {
		printf("my address: %X\n", this);
		printf("   dims: [%d,%d,%d]\n", dims[0],dims[1],dims[2]);
		printf("   data: %X\n", data);
	}
	
	double* getSplineCoeffs() { return spline_coeffs; };

private:
	vec3<int> dims;
	int dimsProd;
	T* data;

	double* spline_coeffs;  // cache spline coeffs for spline interpolation
	int spline_order;
};

//
// read a single value from an M-file
bool readValueFromFile(string infile, vec3<double> pt, float &val) {
	ifstream in;
	in.open(infile.c_str(), ios::in);

	if (!in.is_open())  {
		// cerr << "ERROR: Unable to open match map '" << infile << "'\n";
		return false;
	}

	// extract relevant bytes from header....assume that file was written
	//      on machine with same endianness as the one on which it is being read
	int ni,nj,nk;
	in.seekg(160);
	in.read((char*)&ni, 4);
	in.read((char*)&nj, 4);
	in.read((char*)&nk, 4);
	in.seekg(28, ios::cur);

	vec3<int> dims(ni,nj,nk);
	pt[0] *= dims[0]; pt[1] *= dims[1]; pt[2] *= dims[2];
	int pt000[3];
	pt000[0] = pos_mod(int(floor(pt[0]+.5)),dims[0]);
	pt000[1] = pos_mod(int(floor(pt[1]+.5)),dims[1]);
	pt000[2] = pos_mod(int(floor(pt[2]+.5)),dims[2]);

	in.seekg((pt000[2]*dims[1]*dims[0] + pt000[1]*dims[0] + pt000[0])*4, ios::cur);
	in.read((char*)&val, 4);
	in.close();

	return true;
}



//
//  write a map to disk
//
template <class S>
void writeMapToDisk(map_3d<S> &M, string filestem) {
	string outfile = filestem + ".mat";
	vec3<int> dims = M.getDims();

	ofstream out;
	out.open(outfile.c_str(), ios::out);

	// write file header
	short buff_s;
	long buff_l;
	out.write("MATLAB 5.0 MAT-file, Platform: GLNX86   "
			  "                                        "
			  "                                        "
			  "    ", 124);
	buff_s=0x0100;            // endian-ness test
	out.write((char*)&buff_s, 2);
	buff_s=0x4D49;            // code(?)
	out.write((char*)&buff_s, 2);

	// write data block header
	buff_l = 0x0000000E;      // data type (array)
	out.write((char*)&buff_l, 4);
	buff_l = 4*dims[0]*dims[1]*dims[2] + 64;      // number of bytes
	out.write((char*)&buff_l, 4);

	// write array header
	buff_l = 0x00000006;      // array data type (single)
	out.write((char*)&buff_l, 4);
	buff_l = 0x00000008;      // # bytes in array data block
	out.write((char*)&buff_l, 4);
	buff_l = 0x00000007;      // global/logical flags
	out.write((char*)&buff_l, 4);
	buff_l = 0x00000000;      // padding
	out.write((char*)&buff_l, 4);

	// write dims array
	buff_l = 0x00000005;      // dims data type (int32)
	out.write((char*)&buff_l, 4);
	buff_l = 0x0000000C;      // # bytes in dims data block
	out.write((char*)&buff_l, 4);
	out.write((char*)&dims[0], 12); 
	buff_l = 0x00000000;      // padding
	out.write((char*)&buff_l, 4);

	// write name array
	buff_l=0x0001;            // name data type (int8)
	out.write((char*)&buff_l, 4);
	buff_l=0x0005;            // # bytes in name
	out.write((char*)&buff_l, 4);
	out.write("match   ", 8); // pad with spaces so next write is word-aligned

	// write array
	buff_l = 0x00000007;      // array data type (single)
	out.write((char*)&buff_l, 4);
	buff_l = 4*dims[0]*dims[1]*dims[2];
	out.write((char*)&buff_l, 4);

	for (int k=0; k<dims[2]; k++) {
		for (int j=0; j<dims[1]; j++) {
			for (int i=0; i<dims[0]; i++) {
				float buff_f = (float) M.get(i*dims[1]*dims[2] + j*dims[2] + k);
				out.write((char*)&buff_f, 4);
			}
		}
	}

	out.close();
}

void writeMapsToDisk(map_3d< complex<double> > &M, string filestem) {
	vec3<int> dims = M.getDims();
	map_3d<double> Mr(dims),Mi(dims);
	
	for (int i=0; i<dims[0]*dims[1]*dims[2]; ++i) {
		Mr.set(i, M.get(i).real() );
		Mi.set(i, M.get(i).imag() );
	}
	
	writeMapToDisk(Mr, filestem+"r");
	writeMapToDisk(Mi, filestem+"i");
}



// functions on maps
template <class S>
double KLdiv(map_3d<S> &A,map_3d<S> &B) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = A.getDimsProd();
	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "KLdiv(): map " << A.getDims().to_string() << " and mask "
		                        << B.getDims().to_string() << " dimensions do not match!\n";
		return -1.0;
	}

	double retval = 0.0;
	for (int i=0; i<N; ++i) {
		if (A.get(i) * B.get(i) > 0) 
			retval += A.get(i) * log( A.get(i) / B.get(i) );
	}
	retval /= log(2); // log base 2
	return retval;
}

template <class S>
double KLdiv(map_3d<S> &A,map_3d<S> &B, map_3d<double> &mask) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = A.getDimsProd();
	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "KLdiv(): map " << A.getDims().to_string() << " and mask "
		                        << B.getDims().to_string() << " dimensions do not match!\n";
		return -1.0;
	}

	double retval = 0.0;
	for (int i=0; i<N; ++i) {
		if (mask.get(i) > 0) {
			retval += A.get(i) * log( A.get(i) / B.get(i) );
		}
	}
	retval /= log(2); // log base 2
	return retval;
}


template <class S>
S map_dot(map_3d<S>& A, map_3d<S>& B) {
	S dotProd = 0.0;
	for (int i=0; i < A.getDims(0)*A.getDims(1)*A.getDims(2); i++)
		dotProd += A.get(i) * B.get(i);
	return dotProd;
}

template <class S>
S map_delta(map_3d<S>& A, map_3d<S>& B) {
	S dotProd = 0.0;
	for (int i=0; i < A.getDims(0)*A.getDims(1)*A.getDims(2); i++)
		dotProd += fabs(exp(A.get(i)) - exp(B.get(i)));    // TO DO??? make this take difference in exp()'s???
	return dotProd;
}


// fft r->c
/*
void fft(map_3d<double> &X, map_3d< complex<double> > &fX) {
	fX.alloc(X.getDims(0), X.getDims(1), X.getDims(2));

	fftw_plan p = fftw_plan_dft_r2c_3d(X.getDims(0), X.getDims(1), X.getDims(2),
	                                   &X.get(0), reinterpret_cast<fftw_complex*>(&fX.get(0)), FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);
}
void fft(map_3d<float> &X, map_3d< complex<double> > &fX) {
	static map_3d<double> Xdbl;
	static int dimsProd;
	if (Xdbl.getDims() != X.getDims()) {
		dimsProd = X.getDimsProd();
		Xdbl.alloc(X.getDims(0),X.getDims(1),X.getDims(2));
	}
	for (int i=0; i<dimsProd; ++i) {
		Xdbl.set(i, (double)X.get(i));
	}
	fft(Xdbl,fX);
}

// ifft, c->r
void ifft(map_3d< complex<double> > &fX, map_3d<double> &X) {
	X.alloc(fX.getDims(0), fX.getDims(1), fX.getDims(2));
	fftw_plan p = fftw_plan_dft_c2r_3d(fX.getDims(0), fX.getDims(1), fX.getDims(2),
	                                   reinterpret_cast<fftw_complex*>(&fX.get(0)), &X.get(0), FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);
	//X.scale(1.0/(fX.getDims(0)*fX.getDims(1)*fX.getDims(2)));
}
void ifft(map_3d< complex<double> > &fX, map_3d<float> &X) {
	static map_3d<double> Xdbl;
	int dimsProd = fX.getDimsProd();
	ifft(fX,Xdbl);
	X.alloc(Xdbl.getDims(0),Xdbl.getDims(1),Xdbl.getDims(2));
	for (int i=0; i<dimsProd; ++i) {
		X.set(i, (float)Xdbl.get(i));
	}
}
*/

void fft(map_3d<double> &X, map_3d< complex<double> > &fX) {
	vec3<int> dims = X.getDims();
	int dimsProd = dims[0]*dims[1]*dims[2];

	fftw_complex *in,*out;
	in  = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	out = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	
	for (int i=0; i<dimsProd; ++i) {
		in[i][0] = X.get(i);
		in[i][1] = 0.0;
	}
	
	fftw_plan p = fftw_plan_dft_3d( dims[0],dims[1],dims[2],
	                               in, out, 
								   FFTW_FORWARD, FFTW_ESTIMATE);
								   
	fftw_execute(p);
	fftw_destroy_plan(p);

	fX.alloc(dims);
	for (int i=0; i<dimsProd; ++i) {
		fX.set( i, complex<double> (out[i][0],out[i][1]) );
	}
	
	fftw_free(in); fftw_free(out);
}
void fft(map_3d<float> &X, map_3d< complex<double> > &fX) {
	vec3<int> dims = X.getDims();
	int dimsProd = dims[0]*dims[1]*dims[2];

	fftw_complex *in,*out;
	in  = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	out = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	
	for (int i=0; i<dimsProd; ++i) {
		in[i][0] = (double)X.get(i);
		in[i][1] = 0.0;
	}
	
	fftw_plan p = fftw_plan_dft_3d( dims[0],dims[1],dims[2],
	                               in, out, 
								   FFTW_FORWARD, FFTW_ESTIMATE);
						   
	fftw_execute(p);
	fftw_destroy_plan(p);

	fX.alloc(dims);
	for (int i=0; i<dimsProd; ++i) {
		fX.set( i, complex<double> (out[i][0],out[i][1]) );
	}
	
	fftw_free(in); fftw_free(out);
}

// ifft, c->r
void ifft(map_3d< complex<double> > &fX, map_3d<double> &X) {
	vec3<int> dims = fX.getDims();
	int dimsProd = dims[0]*dims[1]*dims[2];

	fftw_complex *in,*out;
	in  = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	out = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	
	for (int i=0; i<dimsProd; ++i) {
		out[i][0] = fX.get(i).real();
		out[i][1] = fX.get(i).imag();
	}
	
	fftw_plan p = fftw_plan_dft_3d( dims[0],dims[1],dims[2],
	                               out, in, 
								   FFTW_BACKWARD, FFTW_ESTIMATE);
								   
	fftw_execute(p);
	fftw_destroy_plan(p);

	X.alloc(dims);
	for (int i=0; i<dimsProd; ++i) {
		X.set( i, in[i][0] );
	}
	
	fftw_free(in); fftw_free(out);
	//X.scale(1.0/(fX.getDims(0)*fX.getDims(1)*fX.getDims(2)));
}
void ifft(map_3d< complex<double> > &fX, map_3d<float> &X) {
	vec3<int> dims = fX.getDims();
	int dimsProd = dims[0]*dims[1]*dims[2];

	fftw_complex *in,*out;
	in  = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	out = (fftw_complex*)fftw_malloc(dimsProd* sizeof(fftw_complex));
	
	for (int i=0; i<dimsProd; ++i) {
		out[i][0] = fX.get(i).real();
		out[i][1] = fX.get(i).imag();
	}
	
	fftw_plan p = fftw_plan_dft_3d( dims[0],dims[1],dims[2],
	                               out, in, 
								   FFTW_BACKWARD, FFTW_ESTIMATE);
								   
	fftw_execute(p);
	fftw_destroy_plan(p);

	X.alloc(dims);
	for (int i=0; i<dimsProd; ++i) {
		X.set( i, (float)in[i][0] );
	}
	
	fftw_free(in); fftw_free(out);
	//X.scale(1.0/(fX.getDims(0)*fX.getDims(1)*fX.getDims(2)));
}


// fft, c->c
void fft(map_3d< complex<double> > &X, map_3d< complex<double> > &fX) {
	fX.alloc(X.getDims(0), X.getDims(1), X.getDims(2));

	fftw_plan p = fftw_plan_dft_3d(X.getDims(0), X.getDims(1), X.getDims(2),
	                               reinterpret_cast<fftw_complex*>(&X.get(0)),
	                               reinterpret_cast<fftw_complex*>(&fX.get(0)), FFTW_FORWARD, FFTW_ESTIMATE);
 	fftw_execute(p);
	fftw_destroy_plan(p);
}

// ifft, c->c
void ifft(map_3d< complex<double> > &fX, map_3d< complex<double> > &X) {
	X.alloc(fX.getDims(0), fX.getDims(1), fX.getDims(2));
	fftw_plan p = fftw_plan_dft_3d(X.getDims(0), X.getDims(1), X.getDims(2),
	                               reinterpret_cast<fftw_complex*>(&fX.get(0)),
	                               reinterpret_cast<fftw_complex*>(&X.get(0)), FFTW_BACKWARD, FFTW_ESTIMATE);
	fftw_execute(p);
	fftw_destroy_plan(p);
	//X.scale(1.0/(fX.getDims(0)*fX.getDims(1)*fX.getDims(2)));
}




// element-by-element multiplication of A and B gives AB
template <class S, class T, class U>
void map_times(map_3d<S>& AB, map_3d<T>& A, map_3d<U>& B) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = dimsA[0]*dimsA[1]*dimsA[2];

	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "map_3d::map_times(): map " << dimsA.to_string() << " and map "
		                                    << dimsB.to_string() << " dimensions do not match!\n";
		return;
	}
	AB.alloc(dimsA);
	for (int i=0; i < N; i++)
		AB.set(i , A.get(i) * B.get(i));
}
template <class S, class T, class U>
void map_plus(map_3d<S>& AB, map_3d<T>& A, map_3d<U>& B) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = dimsA[0]*dimsA[1]*dimsA[2];

	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "map_3d::map_plus(): map " << A.getDims().to_string() << " and mask "
		                                  << B.getDims().to_string() << " dimensions do not match!\n";
		return;
	}
	AB.alloc(dimsA);
	for (int i=0; i < N; i++)
		AB.set(i , A.get(i) + B.get(i));
}
template <class S, class T, class U>
void map_minus(map_3d<S>& AB, map_3d<T>& A, map_3d<U>& B) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = dimsA[0]*dimsA[1]*dimsA[2];

	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "map_3d::map_minus(): map " << A.getDims().to_string() << " and mask "
		                                  << B.getDims().to_string() << " dimensions do not match!\n";
		return;
	}
	AB.alloc(dimsA);
	for (int i=0; i < N; i++)
		AB.set(i , A.get(i) - B.get(i));
}
// element-by-element linear combination x*A + y*B gives AB
template <class S, class T, class U>
void map_linear_combine(map_3d<S>& AB, double x, map_3d<T>& A, double y,  map_3d<U>& B) {
	vec3<int> dimsA = A.getDims(), dimsB = B.getDims();
	int N = dimsA[0]*dimsA[1]*dimsA[2];

	if (dimsA[0] != dimsB[0] || dimsA[1] != dimsB[1] || dimsA[2] != dimsB[2]) {
		cerr << "map_3d::map_linear_combine(): map " << A.getDims().to_string() << " and mask "
		                                  << B.getDims().to_string() << " dimensions do not match!\n";
		return;
	}
	AB.alloc(dimsA);
	for (int i=0; i < N; i++)
		AB.set(i , x*A.get(i) + y*B.get(i));
}


// multiply by complex conjugate
void conj_map_times(map_3d< complex<double> >& AB, map_3d< complex<double> >& A, map_3d< complex<double> >& B, double C=1) {
	if (A.getDims(0) != B.getDims(0) || A.getDims(1) != B.getDims(1) | A.getDims(2) != B.getDims(2)) {
		cerr << "map_3d::conj_map_times(): map " << A.getDims().to_string() << " and mask "
		                                  << B.getDims().to_string() << " dimensions do not match!\n";
		return;
	}
	AB.alloc(A.getDims(0), A.getDims(1), A.getDims(2));

	if (C==1) {
		for (int i=0; i < A.getDimsProd(); i++)
			AB.set(i , conj(A.get(i)) * B.get(i));
	} else {
		for (int i=0; i < A.getDimsProd(); i++)
			AB.set(i , C * conj(A.get(i)) * B.get(i));
	}
}

// apply a matrix transformation to a masked map
void map_transform(map_3d<double>& dest, map_3d<double>& src, mat33<double>& trans_inv, map_3d<bool> &mask, InterpolationType iType) {
	dest.alloc(src.getDims(0), src.getDims(1), src.getDims(2));

	vec3<double> src_coord, dest_coord;

	if (iType == SPLINE2) src.computeSplineCoeffs(2);
	if (iType == SPLINE3) src.computeSplineCoeffs(3);
	if (iType == SPLINE4) src.computeSplineCoeffs(4);
	if (iType == SPLINE5) src.computeSplineCoeffs(5);

	for (int i=0; i<dest.getDims(0); i++) {
		dest_coord[0] = (double)i/dest.getDims(0); if (dest_coord[0]>0.5) dest_coord[0] -= 1;
		for (int j=0; j<dest.getDims(1); j++) {
			dest_coord[1] = (double)j/dest.getDims(1); if (dest_coord[1]>0.5) dest_coord[1] -= 1;
			for (int k=0; k<dest.getDims(2); k++) {
				dest_coord[2] = (double)k/dest.getDims(2);
				if (mask.get(i,j,k)) {
					if (dest_coord[2]>0.5) dest_coord[2] -= 1;

					src_coord = dest_coord.apply(trans_inv);

					switch (iType) {
					case SPLINE2:
					case SPLINE3:
					case SPLINE4:
					case SPLINE5:
						dest.set(i,j,k, src.interp_spline(src_coord) );
						break;
					case LINEAR:
						dest.set(i,j,k, src.interp_linear(src_coord) );
						break;
					default:
						dest.set(i,j,k, src.interp_nearest(src_coord) );
					}

					//cout << src_coord.to_string() << " => " << dest_coord.to_string() << endl;
				} else {
					dest.set(i,j,k,0.0);
				}
			}
		}
	}
}



// resize a map (using FFT-interpolation)
void resize_fft(map_3d<double>& oldMap, map_3d<double>& newMap) {
	if (newMap.getDims(0) == oldMap.getDims(0) && newMap.getDims(1) == oldMap.getDims(1) && newMap.getDims(2) == oldMap.getDims(2)) {
		memcpy(&newMap.get(0), &oldMap.get(0), oldMap.getDimsProd()*sizeof(double));
		return;
	}

//	cout << "Resizing " << oldMap.getDims(0) << "x" << oldMap.getDims(1) << "x" << oldMap.getDims(2) << " to "
//	                    << newMap.getDims(0) << "x" << newMap.getDims(1) << "x" << newMap.getDims(2) << endl;

	// convert this map to complex<double>
	map_3d< complex<double> > oldMap_cplx(oldMap.getDims(0),oldMap.getDims(1),oldMap.getDims(2)),
	                          Foldmap    (oldMap.getDims(0),oldMap.getDims(1),oldMap.getDims(2)),
	                          Fnewmap    (newMap.getDims(0),newMap.getDims(1),newMap.getDims(2)),
	                          newMap_cplx(newMap.getDims(0),newMap.getDims(1),newMap.getDims(2));
	for (int i=0; i<oldMap.getDimsProd(); i++)
		oldMap_cplx.set(i, complex<double>(oldMap.get(i)) );

	// fft
	fft (oldMap_cplx, Foldmap);

	// reshape (need to handle both shrinking and growing in each dimension)
	Fnewmap.setAll(0);
	vec3<int> nyq( min(oldMap.getDims(0),newMap.getDims(0))/2,
	               min(oldMap.getDims(1),newMap.getDims(1))/2,
	               min(oldMap.getDims(2),newMap.getDims(2))/2 );
	vec3<int> nyqplus1_old( max(oldMap.getDims(0) - (min(oldMap.getDims(0),newMap.getDims(0))-nyq[0]) + 1 , nyq[0]+1) ,
	                        max(oldMap.getDims(1) - (min(oldMap.getDims(1),newMap.getDims(1))-nyq[1]) + 1 , nyq[1]+1) ,
	                        max(oldMap.getDims(2) - (min(oldMap.getDims(2),newMap.getDims(2))-nyq[2]) + 1 , nyq[2]+1) );
	vec3<int> nyqplus1_new( max(newMap.getDims(0) - (min(oldMap.getDims(0),newMap.getDims(0))-nyq[0]) + 1 , nyq[0]+1) ,
	                        max(newMap.getDims(1) - (min(oldMap.getDims(1),newMap.getDims(1))-nyq[1]) + 1 , nyq[1]+1) ,
	                        max(newMap.getDims(2) - (min(oldMap.getDims(2),newMap.getDims(2))-nyq[2]) + 1 , nyq[2]+1) );
	for (int i=0; i<newMap.getDims(0); i++)
		for (int j=0; j<newMap.getDims(1); j++)
			for (int k=0; k<newMap.getDims(2); k++) {
				if (i<=nyq[0]) {
					if (j<=nyq[1]) {
						if (k<=nyq[2])
							Fnewmap.set(i,j,k, Foldmap.get(i,j,k));
						else if (k>=nyqplus1_new[2])
							Fnewmap.set(i,j,k, Foldmap.get(i,j, k-nyqplus1_new[2]+nyqplus1_old[2]) );

					} else if (j>=nyqplus1_new[1]) {
						if (k<=nyq[2])
							Fnewmap.set(i,j,k, Foldmap.get(i,j-nyqplus1_new[1]+nyqplus1_old[1],k));
						else if (k>=nyqplus1_new[2])
							Fnewmap.set(i,j,k, Foldmap.get(i,j-nyqplus1_new[1]+nyqplus1_old[1], k-nyqplus1_new[2]+nyqplus1_old[2]) );

					}
				} else if (i>=nyqplus1_new[0]) {
					if (j<=nyq[1]) {
						if (k<=nyq[2])
							Fnewmap.set(i,j,k, Foldmap.get(i-nyqplus1_new[0]+nyqplus1_old[0],j,k));
						else if (k>=nyqplus1_new[2])
							Fnewmap.set(i,j,k, Foldmap.get(i-nyqplus1_new[0]+nyqplus1_old[0],j, k-nyqplus1_new[2]+nyqplus1_old[2]) );

					} else if (j>=nyqplus1_new[1]){
						if (k<=nyq[2])
							Fnewmap.set(i,j,k, Foldmap.get(i-nyqplus1_new[0]+nyqplus1_old[0],j-nyqplus1_new[1]+nyqplus1_old[1],k));
						else if (k>=nyqplus1_new[2])
							Fnewmap.set(i,j,k, Foldmap.get(i-nyqplus1_new[0]+nyqplus1_old[0],
							                               j-nyqplus1_new[1]+nyqplus1_old[1],
							                               k-nyqplus1_new[2]+nyqplus1_old[2]) );

					}
				}
			}

	// ifft
	ifft(Fnewmap, newMap_cplx);

	// recast to double
	for (int i=0; i<newMap.getDimsProd(); i++)
		newMap.set(i, newMap_cplx.get(i).real() );
	newMap.scale(1.0/(oldMap.getDimsProd()));
}



// resize a map (using splines)
void resize_spline(map_3d<double>& oldMap, map_3d<double>& newMap, int order) {
	if (newMap.getDims(0) == oldMap.getDims(0) && newMap.getDims(1) == oldMap.getDims(1) && newMap.getDims(2) == oldMap.getDims(2)) {
		memcpy(&newMap.get(0), &oldMap.get(0), oldMap.getDimsProd()*sizeof(double));
		return;
	}

	//cout << "Resizing " << oldMap.getDims(0) << "x" << oldMap.getDims(1) << "x" << oldMap.getDims(2) << " to "
	//                    << newMap.getDims(0) << "x" << newMap.getDims(1) << "x" << newMap.getDims(2) << endl;

	// reshape
	vec3<double> coord;
	oldMap.computeSplineCoeffs(order);

	for (int i=0; i<newMap.getDims(0); i++) {
		coord[0] = (double)i/newMap.getDims(0);
		for (int j=0; j<newMap.getDims(1); j++) {
			coord[1] = (double)j/newMap.getDims(1);
			for (int k=0; k<newMap.getDims(2); k++) {
				coord[2] = (double)k/newMap.getDims(2);
				newMap.set(i,j,k, oldMap.interp_spline(coord) );
			}
		}
	}
	
	oldMap.reset();  // delete spline coeffs
}


// resize a map (using nearest neighbor interpolation)
void resize_nn(map_3d<double>& oldMap, map_3d<double>& newMap) {
	if (newMap.getDims(0) == oldMap.getDims(0) && newMap.getDims(1) == oldMap.getDims(1) && newMap.getDims(2) == oldMap.getDims(2)) {
		memcpy(&newMap.get(0), &oldMap.get(0), oldMap.getDimsProd()*sizeof(double));
		return;
	}

	vec3<double> coord;

	//cout << "Resizing " << oldMap.getDims(0) << "x" << oldMap.getDims(1) << "x" << oldMap.getDims(2) << " to "
	//                    << newMap.getDims(0) << "x" << newMap.getDims(1) << "x" << newMap.getDims(2) << endl;

	// reshape
	for (int i=0; i<newMap.getDims(0); i++) {
		coord[0] = (double)i/newMap.getDims(0);
		for (int j=0; j<newMap.getDims(1); j++) {
			coord[1] = (double)j/newMap.getDims(1);
			for (int k=0; k<newMap.getDims(2); k++) {
				coord[2] = (double)k/newMap.getDims(2);
				newMap.set(i,j,k, oldMap.interp_nearest(coord) );
			}
		}
	}
}



//
// Reading a word from a big endian binary file
//
void ReadWord(ifstream &mapin, char *word, int byteOrder) {
	int* i = (int*) word;
	unsigned char byte_in[4];
	mapin.read((char*)byte_in, 4);
	if (byteOrder == 1)
		*i = (byte_in[0]) + (byte_in[1] << 8) + (byte_in[2] << 16) + (byte_in[3] << 24);  // intel
	else
		*i = (byte_in[3]) + (byte_in[2] << 8) + (byte_in[1] << 16) + (byte_in[0] << 24);  // others
}


////////////////////////////
//
// read an FS4 dens map
//
////////////////////////////
bool readDensityMap(map_3d<double>& map, string mapfile) {
	ifstream mapin(mapfile.c_str(), ios::in|ios::binary);

	if (!mapin.is_open())  {
		cerr << "ERROR: Unable to open match map '" << mapfile << "'\n";
		return false;
	}

	int readBuf, jump=4, ordering, byteOrder=1;
	vec3<int> dims;
	char map2fs_test[6];

	// was this map generated by map2fs??  if so, skip extra header
	mapin.seekg(27,ios_base::beg);
	mapin.read(map2fs_test, 6);
	if (!strncmp(map2fs_test, "map2fs", 6))
		jump += 44;
	mapin.seekg(0,ios_base::beg);  // back to beginning of file

	// read the first word in the file, which describes the size of
	//   the header block.  Determine Endianness. Skip ahead.
	unsigned char byte_in[4];
	mapin.read((char*)byte_in, 4);
	
	if ((byte_in[0]) + (byte_in[1] << 8) > (byte_in[3]) + (byte_in[2] << 8)) {
		byteOrder = 1;
		readBuf = (byte_in[0]) + (byte_in[1] << 8);
	} else {
		byteOrder = -1;
		readBuf = (byte_in[3]) + (byte_in[2] << 8);
	}
	mapin.seekg(readBuf+jump, ios_base::cur);

	// Next block is the map stats block ... read the first 5 words
	//     word 1    size of symmetry block
	//      "   2    nrow
	//      "   3    ncol
	//      "   4    nplane
	//      "   5    norn = ordering of axes: 0 (XZY), 1 (YZX), or 2 (XYZ)
	ReadWord(mapin, (char*)&readBuf , byteOrder);
	ReadWord(mapin, (char*)&dims[0] , byteOrder);
	ReadWord(mapin, (char*)&dims[1] , byteOrder);
	ReadWord(mapin, (char*)&dims[2] , byteOrder);
	mapin.seekg(4, ios_base::cur);
	ReadWord(mapin, (char*)&ordering , byteOrder);

	// read and store in >XYZ< ordering
	int _x, _y, _z;
	switch(ordering) {
		case 0:  _x = 2; _y = 0; _z = 1; break;
		case 1:  _x = 0; _y = 1; _z = 2; break;
		default: _x = 1; _y = 0; _z = 2;
	}

	map.alloc(dims[_y],dims[_x],dims[_z]);

	//cout << "Read match map '" << mapfile << "' " << map.getDims().to_string() << "\n";

	// Now begin reading the data
	// First skip ahead to the data block
	mapin.seekg(readBuf-16, ios_base::cur);

	vec3<int> X;
	for (X[2]=0; X[2]<dims[2]; X[2]++) {
		for (X[1]=0; X[1]<dims[1]; X[1]++) {
			mapin.seekg(4, ios_base::cur);
			for (X[0]=0; X[0]<dims[0]; X[0]++) {
				ReadWord(mapin, (char*)&readBuf, byteOrder);
				map.set(X[_y],X[_x],X[_z],(double)readBuf);
			}
			mapin.seekg(4, ios_base::cur);
		}
	}

	mapin.close();
	return true;
}
#endif

