// //////////////////////////////////////////////////////////////////////////
//
//     map_mask.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ////////////////////////////////////////////////////////////////////////////


#include "protein.h"
#include "crystal.h"
#include "map3d.h"
#include <ext/hash_set>
#include <ext/hash_map>
#include <map>

using __gnu_cxx::hash_set;
using __gnu_cxx::hash_map;
using __gnu_cxx::hash_multimap;


// masknode (16 bytes)
class maskNode {
public:
	maskNode( vec3<float> fracX , int resId , string atmId , int elt ) {
		this->fracX = fracX;
		if (atmId == " C  " ||atmId == " N  ") 
			is_bb_CN = 1;
		else
			is_bb_CN = 0;
		
		this->resId = (short)resId;
		this->elt   = (unsigned char) elt;
	}

	vec3<float> fracX;
	//string atmId;
	short resId;
	unsigned char elt;
	unsigned char is_bb_CN;
};

class MapMask {
public:
	MapMask() {}
	
	void init(Crystal& crystData) {
		dims[0] = (int)floor( crystData.A/6.0  );
		dims[1] = (int)floor( crystData.B/6.0  );
		dims[2] = (int)floor( crystData.C/6.0  );
	}
	
	void updateMask( int resId, vector<Atom> &thisRes , Crystal& crystData ) {
		int N = thisRes.size();
		
		for (int i=0; i<N; ++i) {
			vec3<double> frac_i = thisRes[i].cartX.apply(crystData.c2f);
			vec3<int> idx_i ( pos_mod( (int)floor( dims[0] * frac_i[0]  ) , dims[0] ) , 
			                  pos_mod( (int)floor( dims[1] * frac_i[1]  ) , dims[1] ) , 
			                  pos_mod( (int)floor( dims[2] * frac_i[2]  ) , dims[2] ) );
							  
			int toInsert = idx_i[0]*dims[1]*dims[2] + idx_i[1]*dims[2] + idx_i[2];
			mask.insert( pair< int , maskNode >
			               ( toInsert , maskNode( vec3<float>(frac_i[0],frac_i[1],frac_i[2]) , resId , 
						                          thisRes[i].atmid , thisRes[i].elt ) ) );
		}
	}
	
	
	bool checkMask( int resId, vector<Atom> &thisRes , Crystal &crystData , bool verbose = false) {
		int N = thisRes.size();
		int nsymm = crystData.nSymm();
		
		for (int i=0; i<N; ++i) {
			vec3<double> frac_i = thisRes[i].cartX.apply(crystData.c2f), frac_Si;
			int thisElt = thisRes[i].elt;
			unsigned int is_bb_CN;
			
			if (thisRes[i].atmid == " C  " || thisRes[i].atmid == " N  ") 
				is_bb_CN = 1;
			else
				is_bb_CN = 0;
				
			for (int j=0; j<nsymm; ++j) {
				crystData.getSymmetric(frac_i, frac_Si, j);	
			
				// check in a 3x3 'window'
				vec3<int> idx_i_ul ( pos_mod( (int)floor( dims[0] * frac_Si[0]  )-1 , dims[0] ) , 
				                     pos_mod( (int)floor( dims[1] * frac_Si[1]  )-1 , dims[1] ) , 
				                     pos_mod( (int)floor( dims[2] * frac_Si[2]  )-1 , dims[2] ) );
				
				for (int x=0; x<=2; ++x) 
				for (int y=0; y<=2; ++y) 
				for (int z=0; z<=2; ++z) {
					vec3<int> idx_i ( idx_i_ul[0]+x, idx_i_ul[1]+y, idx_i_ul[2]+z );
					
					if (idx_i[0] >= dims[0]) idx_i[0] -= dims[0];
					if (idx_i[1] >= dims[1]) idx_i[1] -= dims[1];
					if (idx_i[2] >= dims[2]) idx_i[2] -= dims[2];
					
					int toTest = idx_i[0]*dims[1]*dims[2] + idx_i[1]*dims[2] + idx_i[2];
					
					if (mask.count(toTest) > 0) {
						// closer look ... 
						//pair< hash_multimap< int,maskNode >::const_iterator, 
						//      hash_multimap< int,maskNode >::const_iterator > p = mask.equal_range(toTest);
						pair< multimap< int,maskNode >::const_iterator, 
						      multimap< int,maskNode >::const_iterator > p = mask.equal_range(toTest);
						
						//for (hash_multimap< int,maskNode >::const_iterator k = p.first; k != p.second; ++k) {
						for (multimap< int,maskNode >::const_iterator k = p.first; k != p.second; ++k) {
							vec3<float> frac_j = (*k).second.fracX;
							
							// MIN_dist is (for now) 1/2 the sum of van der waals radii
							double MIN_DIST = 0.5 * (chem_db_VDW[(*k).second.elt-1] + chem_db_VDW[thisElt-1]);
							
							vec3<double> del(frac_Si[0]-frac_j[0], frac_Si[1]-frac_j[1], frac_Si[2]-frac_j[2]);
							del.minmod1();
							
							if ( del.apply(crystData.f2c).norm2() < MIN_DIST*MIN_DIST ) {
							
								// check if collision == bonded atoms
								if ( abs(resId - (*k).second.resId) == 1 &&
								     ( (*k).second.is_bb_CN == 1 ) &&
									 (  is_bb_CN == 1 ) ){	
									; // do nothing
								} else {
									return true;   // collision
								}
							}
						}
					}
				}
			}
		}
		
		return false;  // no collision
	}
	
	
	bool checkMask(vec3<double> thisAtm , Crystal &crystData , bool verbose = false) {
		vector<Atom> tmp(1);
		tmp[0].atmid = " CA ";
		tmp[0].elt   = 6;
		tmp[0].cartX = thisAtm;
		
		return checkMask( -1 , tmp , crystData, verbose);
	}

private:
	vec3<int> dims;
	//hash_multimap< int , maskNode > mask;
	multimap< int , maskNode > mask;
};
