// //////////////////////////////////////////////////////////////////////////
//
//     spharm.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////

#include "nmer_lookup.h"
#include "map3d.h"
#include "soft_wrapper.h"

extern int POSITIONAL_COST[];
extern int MAX_PAMDIST;

double reso;


//
// nmer lookup
//
void spharm_lookup( string                        mer,
                    int                           nexs,
                    double                        radius,
                    vec3<int>                     dims,
                    Crystal                      &cryst,
                    map_3d< cubicSplineCoeff3d > &sphPtrs,
                    vector< Nmer >               &frags,           // output
                    vector< map_3d<double> >     &sigCoefsR,       // output
                    vector< map_3d<double> >     &sigCoefsI        // output
                  )
{
	//frags.clear();
	//sigCoefsR.clear(); sigCoefsI.clear();

	int R = sphPtrs.getDims(0);

	// db lookup
	nmer_lookup(mer,nexs,frags,POSITIONAL_COST, MAX_PAMDIST, radius);
	int nfrags = frags.size();

	sigCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	sigCoefsI.resize( nfrags, map_3d<double>(R,B,B) );
	
	map_3d<double> sphFragMap (R,2*B,2*B);
	map_3d<double> cartFragMap(dims[0],dims[1],dims[2]);

	// for each fragment ...
	for (int i=0; i<nfrags; ++i) {
		// ... compute Fc ...
		frags[i].getFcSphere(cartFragMap, cryst, reso);
		
		// ... spherical resample ...
		spherical_resample_cubic_spline(sphFragMap, cartFragMap, sphPtrs, vec3<int>(0,0,0));

		// ... standardize ...
		sph_standardize(sphFragMap);

		// ... and SH transform
		SOFT_sharm_transform(sphFragMap, sigCoefsR[i], sigCoefsI[i]);
		
		double sumCoef2 = 0.0;
		for (int r_idx = 0; r_idx < R; ++r_idx) {
			double thisRWt =  ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );
			for (int ii=0; ii<B; ++ii) 
			for (int jj=0; jj<B; ++jj) {
				if (ii==0 && jj==0) continue;
				sumCoef2 += thisRWt * ( square(sigCoefsR[i].get(r_idx,ii,jj)) + square(sigCoefsI[i].get(r_idx,ii,jj)) );
			}
		}
				
		// cerr << sqrt(sumCoef2) << endl;
	}
}

//
// nmer lookup
//
void spharm_lookupAndMask (
                     string                        mer,
                     int                           nexs,
                     double                        radius,
                     vec3<int>                     dims,
                     Crystal                      &cryst,
                     map_3d< cubicSplineCoeff3d > &sphPtrs,
                     map_3d< linearCoeff3d >         &linPtrs,
                     vector< Nmer >               &frags,           // output
                     vector< map_3d<double> >     &sigCoefsR,       // output
                     vector< map_3d<double> >     &sigCoefsI,       // output
                     vector< map_3d<double> >     &mskCoefsR,       // output
                     vector< map_3d<double> >     &mskCoefsI,       // output
					 double                        mask_distance,
					 int                           radius_align=0
                   )
{
	//frags.clear();
	//sigCoefsR.clear(); sigCoefsI.clear();

	int R = sphPtrs.getDims(0);

	// db lookup
	nmer_lookup(mer,nexs,frags,POSITIONAL_COST, MAX_PAMDIST,radius);
	int nfrags = frags.size();

	sigCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	sigCoefsI.resize( nfrags, map_3d<double>(R,B,B) );
	mskCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	mskCoefsI.resize( nfrags, map_3d<double>(R,B,B) );
	
	map_3d<double> sphFragMap(R,2*B,2*B), sphMaskMap(R,2*B,2*B);
	map_3d<double> cartFragMap(dims[0],dims[1],dims[2]);
	map_3d<double> cartMaskMap(dims[0],dims[1],dims[2]);

	// for each fragment ...
	for (int i=0; i<nfrags; ++i) {
		// ... compute Fc and mask ...
		frags[i].getFcSphere(cartFragMap, cryst, reso);
		frags[i].getMask(cartMaskMap, cryst, mask_distance, radius);

		// ... spherical resample ...
		spherical_resample_cubic_spline(sphFragMap, cartFragMap, sphPtrs, vec3<int>(0,0,0));
		spherical_resample_linear      (sphMaskMap, cartMaskMap, linPtrs, vec3<int>(0,0,0));
		
		// ... mask _everything_ w/i radius_align
		for (int j=0;j<radius_align; ++j) 
		for (int k=0; k<4*B*B; ++k)
			sphMaskMap.set(j,0,k, 1.0);
		
		// SH transform mask 
		SOFT_sharm_transform(sphMaskMap, mskCoefsR[i], mskCoefsI[i]);

		// ... standardize over mask ...
		sph_standardize(sphFragMap, sphMaskMap);
		
		// ... APPLY mask ... 
		map_times(sphFragMap , sphFragMap , sphMaskMap);
		//writeMapToDisk(sphMaskMap, "sphFragMap");

		// ... and SH transform
		SOFT_sharm_transform(sphFragMap, sigCoefsR[i], sigCoefsI[i]);
	}
}



//
// nmer lookup
//
void spharm_lookupFromFilelist(
                       string                        mer,
                       vector< string >             &similar_files,
                       vector< CAtrace >            &similar_pdbs,
                       double                        radius,
                       vec3<int>                     dims,
                       Crystal                      &cryst,
                       map_3d< cubicSplineCoeff3d > &sphPtrs,
					   vector< Nmer >               &frags,           // output
					   vector< map_3d<double> >     &sigCoefsR,       // output
					   vector< map_3d<double> >     &sigCoefsI        // output
					 )
{
	//frags.clear();
	//sigCoefsR.clear(); sigCoefsI.clear();

	int R = sphPtrs.getDims(0);

	// file lookup
	int nSim = similar_files.size();
	for (int i=0; i<nSim; ++i) {
		nmer_lookupInFile(mer, similar_files[i], similar_pdbs[i], frags, radius);
	}

	int nfrags = frags.size();
	if (nfrags == 0) return;

	sigCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	sigCoefsI.resize( nfrags, map_3d<double>(R,B,B) );

	map_3d<double> sphFragMap (R,2*B,2*B);
	map_3d<double> cartFragMap(dims[0],dims[1],dims[2]);

	// for each fragment ...
	for (int i=0; i<nfrags; ++i) {
		// ... compute Fc ...
		frags[i].getFcSphere(cartFragMap, cryst, reso);
		
		// ... spherical resample ...
		spherical_resample_cubic_spline(sphFragMap, cartFragMap, sphPtrs, vec3<int>(0,0,0));

		// ... standardize ...
		sph_standardize(sphFragMap);

		// ... and SH transform
		SOFT_sharm_transform(sphFragMap, sigCoefsR[i], sigCoefsI[i]);
	}
}


//
// nmer lookup
//
void spharm_lookupAndMaskFromFilelist(
                       string                        mer,
                       vector< string >             &similar_files,
                       vector< CAtrace >            &similar_pdbs,
                       double                        radius,
                       vec3<int>                     dims,
                       Crystal                      &cryst,
                       map_3d< cubicSplineCoeff3d > &sphPtrs,
					   map_3d< linearCoeff3d >         &linPtrs,
					   vector< Nmer >               &frags,           // output
					   vector< map_3d<double> >     &sigCoefsR,       // output
					   vector< map_3d<double> >     &sigCoefsI,       // output
                       vector< map_3d<double> >     &mskCoefsR,       // output
                       vector< map_3d<double> >     &mskCoefsI,       // output
					   double                        mask_distance,
					   int                           radius_align=0
					 )
{
	//frags.clear();
	//sigCoefsR.clear(); sigCoefsI.clear();

	int R = sphPtrs.getDims(0);

	// file lookup
	int nSim = similar_files.size();
	for (int i=0; i<nSim; ++i) {
		nmer_lookupInFile(mer, similar_files[i], similar_pdbs[i], frags, radius);
	}

	int nfrags = frags.size();
	if (nfrags == 0) return;

	sigCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	sigCoefsI.resize( nfrags, map_3d<double>(R,B,B) );
	mskCoefsR.resize( nfrags, map_3d<double>(R,B,B) );
	mskCoefsI.resize( nfrags, map_3d<double>(R,B,B) );
	
	map_3d<double> sphFragMap(R,2*B,2*B), sphMaskMap(R,2*B,2*B);
	map_3d<double> cartFragMap(dims[0],dims[1],dims[2]);
	map_3d<double> cartMaskMap(dims[0],dims[1],dims[2]);

	// for each fragment ...
	for (int i=0; i<nfrags; ++i) {
		// ... compute Fc and mask ...
		frags[i].getFcSphere(cartFragMap, cryst, reso);
		frags[i].getMask(cartMaskMap, cryst,  mask_distance, radius);

		// ... spherical resample ...
		spherical_resample_cubic_spline(sphFragMap, cartFragMap, sphPtrs, vec3<int>(0,0,0));
		spherical_resample_linear      (sphMaskMap, cartMaskMap, linPtrs, vec3<int>(0,0,0));
		
		// ... mask _everything_ w/i radius_align
		for (int j=0;j<radius_align; ++j) 
		for (int k=0; k<4*B*B; ++k)
			sphMaskMap.set(j,0,k, 1.0);
		
		// SH transform mask (then inverse transform to BW limit)
		SOFT_sharm_transform(sphMaskMap, mskCoefsR[i], mskCoefsI[i]);

		// ... standardize over mask ...
		sph_standardize(sphFragMap, sphMaskMap);
		
		// ... APPLY mask ... 
		map_times(sphFragMap , sphFragMap , sphMaskMap);

		// ... and SH transform
		SOFT_sharm_transform(sphFragMap, sigCoefsR[i], sigCoefsI[i]);

	}

}


//
//  write a map + orientations to disk
//
void writeMapsToDisk(map_3d< orientation > &Mdir, string filestem) {
	vec3<int> dims = Mdir.getDims();
	int N = dims[0]*dims[1]*dims[2];

	// write a seperate map for each parameter in each conformation
	map_3d<float> betaF(dims), gammaF(dims), betaB(dims), gammaB(dims);

	orientation orient_i;

	// write match_0, orient_0's
	betaF.setAll(0); gammaF.setAll(0);
	betaB.setAll(0); gammaB.setAll(0);
	for (int i=0; i<N; i++) {
		orient_i = Mdir.get(i);

		betaF.set (i,orient_i.betaF);
		gammaF.set(i,orient_i.gammaF);
		betaB.set (i,orient_i.betaB);
		gammaB.set(i,orient_i.gammaB);
	}
	writeMapToDisk(betaF,  string(filestem + ".betaF"));
	writeMapToDisk(gammaF, string(filestem + ".gammaF"));
	writeMapToDisk(betaB,  string(filestem + ".betaB"));
	writeMapToDisk(gammaB, string(filestem + ".gammaB"));
}
