// //////////////////////////////////////////////////////////////////////////
//
//     vec3.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#ifndef _VEC3_H_
#define _VEC3_H_ 1

#ifndef INFINITY
#define INFINITY 1e30
#endif

#include <string>
#include <math.h>
#include <vector>
#include <sstream>
#include <complex>
#include <fstream>
#include <iostream>
#include <iomanip>

using namespace std;

inline double fastexp(double X, int N) {
	double retval = X;
	while (N > 1) {
		if (N%2 == 0) {
			N /= 2;
			retval *= retval;
		} else {
			N = (N-1)/2;
			retval *= X*retval;
		}
	}
	return retval;
}

// one-liners
inline double atof(const string& s) { return atof(s.c_str());}
inline int    atoi(const string& s) { return atoi(s.c_str());}

inline string itoc(int i)    { std::ostringstream oss; oss << i; return oss.str(); }
inline string ftoc(double f) { std::ostringstream oss; oss << f; return oss.str(); }

inline int    pos_mod(int a   , int b)    { int c = a%b; if (c<0) c+= b; return c; }
inline double pos_mod(double a, double b) { double c = fmod(a,b); if (c<0) c+= b; return c; }

// max(min) elt and index of a vector
template <class T>
T vectorMax(vector<T> X, int &minIdx) {
	typename vector<T>::iterator p = max_element(X.begin(), X.end());
	T minValue = *p;
	minIdx = distance(X.begin(), p);
	return minValue;
}
template <class T>
T vectorMin(vector<T> X, int &minIdx) {
	typename vector<T>::iterator p = min_element(X.begin(), X.end());
	T minValue = *p;
	minIdx = distance(X.begin(), p);
	return minValue;
}


// trim whitespace
void trim(string& str) {
	string::size_type pos = str.find_last_not_of(' ');
	if(pos != string::npos) {
		str.erase(pos + 1);
		pos = str.find_first_not_of(' ');
		if(pos != string::npos) str.erase(0, pos);
	}
	else str.erase(str.begin(), str.end());
}

// tokenize
void Tokenize(const string    &str,
              vector<string>  &tokens,
              const string    &delimiters=" ") {
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        lastPos = str.find_first_not_of(delimiters, pos);
        pos = str.find_first_of(delimiters, lastPos);
    }
}



// max of 3 elements
template <class T>
T max3(T x, T y, T z) {
	 if (x>y)
	 	if (x>z)
	 		return x;
	 	else
	 		return z;
	 else
	 	if (y>z)
	 		return y;
	 	else
	 		return z;
}

////////////////////////////
//
// struct orientation
//         represents an orientation as a C_alpha +1 orientation (beta, gamma)
//         and a C_alpha -i orientation (beta, gamma)
//
////////////////////////////
struct orientation {
	orientation() { betaF = gammaF = betaB = gammaB = 0; }
	orientation(float bB, float gB, float bF, float gF) {
		betaF = bF; gammaF = gF; betaB = bB; gammaB = gB;
	}

	float betaF, gammaF, betaB, gammaB;
};

ostream& operator << (ostream& os, orientation& s) {
	os << "{ (" << s.betaB << "," << s.gammaB << ") (" << s.betaF << "," << s.gammaF << ") }";
	return os;
}



////////////////////////////
//
// mat33 class
//
////////////////////////////
template <class T>
class mat33 {
	template <class S> friend mat33<S> prod(const mat33<S>&, const mat33<S>&);

public:
	mat33() {
		M[0][1]=M[0][2]=M[1][0]=M[1][2]=M[2][0]=M[2][1]=0;
		M[0][0]=M[1][1]=M[2][2]=1;
	}
	mat33(T x11, T x12, T x13, T x21, T x22, T x23, T x31, T x32, T x33) {
		M[0][0]=x11; M[0][1]=x12; M[0][2]=x13;
		M[1][0]=x21; M[1][1]=x22; M[1][2]=x23;
		M[2][0]=x31; M[2][1]=x32; M[2][2]=x33;
	}
	mat33(T P[][3]) {
		M[0][0]=P[0][0]; M[0][1]=P[0][1]; M[0][2]=P[0][2];
		M[1][0]=P[1][0]; M[1][1]=P[1][1]; M[1][2]=P[1][2];
		M[2][0]=P[2][0]; M[2][1]=P[2][1]; M[2][2]=P[2][2];
	}

	// construct from rotation
	mat33(double a, double b, double g) {
		M[0][0] = (T)(- sin(a)*cos(b)*sin(g) + cos(a)*cos(g));
		M[0][1] = (T)(cos(a)*cos(b)*sin(g) + sin(a)*cos(g));
		M[0][2] = (T)(sin(b)*sin(g));
		M[1][0] = (T)(-sin(a)*cos(b)*cos(g) - cos(a)*sin(g));
		M[1][1] = (T)(cos(a)*cos(b)*cos(g) - sin(a)*sin(g));
		M[1][2] = (T)(sin(b)*cos(g));
		M[2][0] = (T)(sin(a)*sin(b));
		M[2][1] = (T)(-cos(a)*sin(b));
		M[2][2] = (T)(cos(b));
	}

	// calculate matrix inverse
	void set(T x11, T x12, T x13, T x21, T x22, T x23, T x31, T x32, T x33) {
		M[0][0]=x11; M[0][1]=x12; M[0][2]=x13;
		M[1][0]=x21; M[1][1]=x22; M[1][2]=x23;
		M[2][0]=x31; M[2][1]=x32; M[2][2]=x33;
	}
	void set(T P[][3]) {
		M[0][0]=P[0][0]; M[0][1]=P[0][1]; M[0][2]=P[0][2];
		M[1][0]=P[1][0]; M[1][1]=P[1][1]; M[1][2]=P[1][2];
		M[2][0]=P[2][0]; M[2][1]=P[2][1]; M[2][2]=P[2][2];
	}

	mat33<T> inv() {
		T D = M[0][0]*(M[1][1]*M[2][2]-M[2][1]*M[1][2]) -
		      M[0][1]*(M[1][0]*M[2][2]-M[1][2]*M[2][0]) +
		      M[0][2]*(M[1][0]*M[2][1]-M[1][1]*M[2][0]);
		if (D == 0)  {
			cerr << "mat33 ERROR: Inversion of singular matrix!\n";
			mat33<T> retval;
			return (retval);
		}

		return (mat33<T>( (M[1][1]*M[2][2]-M[1][2]*M[2][1])/D,
		                 -(M[0][1]*M[2][2]-M[0][2]*M[2][1])/D,
		                  (M[0][1]*M[1][2]-M[0][2]*M[1][1])/D,
		                 -(M[1][0]*M[2][2]-M[2][0]*M[1][2])/D,
		                  (M[0][0]*M[2][2]-M[0][2]*M[2][0])/D,
		                 -(M[0][0]*M[1][2]-M[0][2]*M[1][0])/D,
		                  (M[1][0]*M[2][1]-M[2][0]*M[1][1])/D,
		                 -(M[0][0]*M[2][1]-M[0][1]*M[2][0])/D,
		                  (M[0][0]*M[1][1]-M[0][1]*M[1][0])/D ));
	}
	
	string to_string() {
		std::ostringstream oss;
		oss << setprecision(3) << setiosflags(ios::fixed);
		oss << "\n     /" << setw(10) << M[0][0] << "," << setw(10) << M[0][1] << "," << setw(10) << M[0][2] << "\\";
		oss << "\nmat33|" << setw(10) << M[1][0] << "," << setw(10) << M[1][1] << "," << setw(10) << M[1][2] << "|";
		oss << "\n     \\" << setw(10) << M[2][0] << "," << setw(10) << M[2][1] << "," << setw(10) << M[2][2] << "/";
		oss << "\n";
		return oss.str();
	}

	inline T& operator()(int i, int j) { return M[i][j]; }
	inline const T& operator()(int i, int j) const { return M[i][j]; };

private:
	T M[3][3];
};

// matrix multiplication
template <class S>
mat33<S> prod(const mat33<S>& A, const mat33<S>& B) {
	return mat33<S>(A.M[0][0]*B.M[0][0] + A.M[0][1]*B.M[1][0] + A.M[0][2]*B.M[2][0],
	                A.M[0][0]*B.M[0][1] + A.M[0][1]*B.M[1][1] + A.M[0][2]*B.M[2][1],
	                A.M[0][0]*B.M[0][2] + A.M[0][1]*B.M[1][2] + A.M[0][2]*B.M[2][2],
	                A.M[1][0]*B.M[0][0] + A.M[1][1]*B.M[1][0] + A.M[1][2]*B.M[2][0],
	                A.M[1][0]*B.M[0][1] + A.M[1][1]*B.M[1][1] + A.M[1][2]*B.M[2][1],
	                A.M[1][0]*B.M[0][2] + A.M[1][1]*B.M[1][2] + A.M[1][2]*B.M[2][2],
	                A.M[2][0]*B.M[0][0] + A.M[2][1]*B.M[1][0] + A.M[2][2]*B.M[2][0],
	                A.M[2][0]*B.M[0][1] + A.M[2][1]*B.M[1][1] + A.M[2][2]*B.M[2][1],
	                A.M[2][0]*B.M[0][2] + A.M[2][1]*B.M[1][2] + A.M[2][2]*B.M[2][2]);
}




////////////////////////////
//
// vec3 class
//
////////////////////////////
template <class T>
class vec3 {
public:
	vec3() { pt[0]=0; pt[1]=0; pt[2]=0; }
	vec3(T x, T y, T z) { pt[0]=x; pt[1]=y; pt[2]=z; }
	vec3(T P[]) { pt[0]=P[0]; pt[1]=P[1]; pt[2]=P[2]; }
	vec3(int idx, vec3<T> dims) {
		pt[0]=  idx / (dims[1]*dims[2]);
		pt[1]= (idx / dims[2]) % dims[1];
		pt[2]=  idx % dims[2];
	}

	string to_string() {
		std::ostringstream oss;
		oss << setprecision(3) << setiosflags(ios::fixed);
		//oss << "vec3(" <<  pt[0] << "," << pt[1] << ","  << pt[2] << ")";
		oss << setw(8) << pt[0] << " , " << setw(8) << pt[1] << " , "  << setw(8) << pt[2];
		return oss.str();
	}

	inline T& operator[](int i) { return pt[i]; }
	inline const T& operator[] (int i) const { return pt[i]; }

	vec3<T> operator- () { return vec3<T>(-pt[0], -pt[1], -pt[2]); }

	vec3<T> operator+ (vec3<T> A) { return vec3<T>(pt[0]+A[0], pt[1]+A[1], pt[2]+A[2]); }
	vec3<T> operator- (vec3<T> A) { return vec3<T>(pt[0]-A[0], pt[1]-A[1], pt[2]-A[2]); }
	vec3<T> operator* (vec3<T> A) { return vec3<T>(pt[0]*A[0], pt[1]*A[1], pt[2]*A[2]); }
	vec3<T> operator/ (vec3<T> A) { return vec3<T>(pt[0]/A[0], pt[1]/A[1], pt[2]/A[2]); }

	vec3<T> operator+ (T A) { return vec3<T>(pt[0]+A, pt[1]+A, pt[2]+A); }
	vec3<T> operator- (T A) { return vec3<T>(pt[0]-A, pt[1]-A, pt[2]-A); }
	vec3<T> operator* (T A) { return vec3<T>(pt[0]*A, pt[1]*A, pt[2]*A); }
	vec3<T> operator/ (T A) { return vec3<T>(pt[0]/A, pt[1]/A, pt[2]/A); }

	inline vec3<T> apply(const mat33<T> &rot) {
		return vec3<T>( rot(0,0)*pt[0] + rot(0,1)*pt[1] + rot(0,2)*pt[2],
		                rot(1,0)*pt[0] + rot(1,1)*pt[1] + rot(1,2)*pt[2],
		                rot(2,0)*pt[0] + rot(2,1)*pt[1] + rot(2,2)*pt[2] );
	}

	inline vec3<T> apply(T rot[3][3], T trans[3]) {
		return vec3<T>( rot[0][0]*pt[0] + rot[0][1]*pt[1] + rot[0][2]*pt[2] + trans[0],
		                rot[1][0]*pt[0] + rot[1][1]*pt[1] + rot[1][2]*pt[2] + trans[1],
		                rot[2][0]*pt[0] + rot[2][1]*pt[1] + rot[2][2]*pt[2] + trans[2] );
	}

	inline vec3<T> apply(T rot[3][3]) {
		return vec3<T>( rot[0][0]*pt[0] + rot[0][1]*pt[1] + rot[0][2]*pt[2],
		                rot[1][0]*pt[0] + rot[1][1]*pt[1] + rot[1][2]*pt[2],
		                rot[2][0]*pt[0] + rot[2][1]*pt[1] + rot[2][2]*pt[2] );
	}

	inline T norm2() {
		return (square(pt[0])+square(pt[1])+square(pt[2]));
	}

	inline void mod(T div) {
		pt[0] = pos_mod(pt[0],div);
		pt[1] = pos_mod(pt[1],div);
		pt[2] = pos_mod(pt[2],div);
	}
	inline void mod(vec3<T> div) {
		pt[0] = pos_mod(pt[0],div[0]);
		pt[1] = pos_mod(pt[1],div[1]);
		pt[2] = pos_mod(pt[2],div[2]);
	}
	inline void minmod(vec3<T> div) {
		pt[0] = pos_mod( pt[0]+div[0]/2 , div[0])-div[0]/2;
		pt[1] = pos_mod( pt[1]+div[1]/2 , div[1])-div[1]/2;
		pt[2] = pos_mod( pt[2]+div[2]/2 , div[2])-div[2]/2;
	}
	inline void minmod1() {
		pt[0] = pos_mod(pt[0]+.5,1.0)-.5;
		pt[1] = pos_mod(pt[1]+.5,1.0)-.5;
		pt[2] = pos_mod(pt[2]+.5,1.0)-.5;
	}

	inline size_t hash() const {
		return (107*pt[0] + 11*pt[1] + pt[2]);
	}

	bool operator==(const vec3<T> A) const {
		return (pt[0]==A.pt[0] && pt[1]==A.pt[1] && pt[2]==A.pt[2]);
	}
	bool operator!=(const vec3<T> A) const {
		return (pt[0]!=A.pt[0] || pt[1]!=A.pt[1] || pt[2]!=A.pt[2]);
	}
	
	T* c_ptr() { return pt; }

	T x() { return pt[0]; }
	T y() { return pt[1]; }
	T z() { return pt[2]; }

	// spherical coords
	double r()     { return sqrt((double)(pt[0]*pt[0] + pt[1]*pt[1] + pt[2]*pt[2])); }
	double gamma() { return atan2( (double)pt[1] , (double)pt[0] ); }
	double beta()  { return acos( pt[2] / r() ); }

private:
	T pt[3];
};

template <class S>
S norm2(vec3<S> X, vec3<S> Y) {
	return ( square(X[0]-Y[0]) + square(X[1]-Y[1]) + square(X[2]-Y[2]) );
}

template <class S>
S dot(vec3<S> X, vec3<S> Y) {
	return ( (X[0]*Y[0]) + (X[1]*Y[1]) + (X[2]*Y[2]) );
}

//
// convert rotation matrix to euler angles
//    (adapted from matrix & quaternion faq, 
//     http://mccammon.ucsd.edu/~adcock/matrixfaq.html#Q37)
// our euler-angle representation uses ZXZ
vec3<double> rot2euler(mat33<double> M) {
	double alpha, beta, gamma;
	double tempx, tempy;
	
	beta  = acos(M(2,2));   // Calculate Y-axis angle 
	double C = sin(beta);
	if (fabs(C) > 1e-4) { // no gimball lock
		tempx    = -M(2,1) / C;
		tempy    =  M(2,0) / C;
		alpha  = atan2( tempy, tempx );
		
		tempx    =  M(1,2) / C;
		tempy    =  M(0,2) / C;
		gamma  = atan2( tempy, tempx );
	} else { // gimball lock
		alpha = 0; // Set X-axis angle to zero 
		
		tempx = M(1,1); // And calculate Z-axis angle 
		tempy = M(0,1);
		gamma = atan2( tempy, tempx );
	}
		
	return vec3<double>(alpha, beta, gamma);
}

//
// align x1-y1 to x2-y2
// TODO:!

inline int to_index(vec3<int> &X, vec3<int> &dims) { return (X[0]*dims[1]*dims[2] + X[1]*dims[2] + X[2]); }


#endif
