// //////////////////////////////////////////////////////////////////////////
//
//     BP.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////
	
#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>
#include <dirent.h>

#include "BP.h"
#include "getopts.h"   // parse command line args


//
// main()
//
int main(int argc, char** argv) {
	///
	/// read input files
	Options bpOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	bpOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	bpOpts.addOption("c",    "cryst",        "[REQUIRED] Input crystal file",      true);
	bpOpts.addOption("s",    "seq",          "[REQUIRED] Input sequence file",      true);
	bpOpts.addOption("o",    "sigma",        "Average squared rotation error (default 1.0)",      true);
	bpOpts.addOption("m",    "m-est",        "Probability of incorrect rotational alignment (default 0.1)",      true);
	bpOpts.addOption("i",    "iters",        "Iterations of BP (default 40)",      true);
	bpOpts.addOption("e",    "seleniums",    "PDB file containing locations of selenium atoms",      true);
	bpOpts.addOption("p",    "se-prior",     "If Se file specified, wgt on specified Se locations (default 0.9)",      true);
	bpOpts.addOption("x",    "memory",       "Maximum memory usage before paging (in MB)",   true);
	bpOpts.addOption("d",    "scratch",      "Scratch directory for paging (will use a unique subdir)",      true);
		
	string workingdir, crystfile, seqfile, seleniumfile;
	double seleniumprior = 0.9;
	
	SCRATCH_DIR = DEFAULT_SCRATCH_DIR;
	MEM_LIMIT   = DEFAULT_MEM_LIMIT;
	SIGMA        = 1.0;
	M_EST_ORIENT = 0.1;
	MAX_ITERS    = 40;
	
	bpOpts.parse(argc, argv);
	while ((switchInt = bpOpts.cycle()) >= 0) {
		vector<string> pargs;
		
		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  bpOpts.getArgs(switchInt); break;
		case 1:   // Option 'cryst'
			crystfile =  bpOpts.getArgs(switchInt); break;
		case 2:   // Option 'seq'
			seqfile =  bpOpts.getArgs(switchInt); break;
		case 3:   // Option 'sigma'
			SIGMA = atof( bpOpts.getArgs(switchInt) ); break;
		case 4:   // Option 'm-est'
			M_EST_ORIENT = atof( bpOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'iters'
			MAX_ITERS = atoi( bpOpts.getArgs(switchInt) ); break;
		case 6:  // Option 'seleniums'
			seleniumfile =  bpOpts.getArgs(switchInt); break;
		case 7:  // Option 'se-prior'
			seleniumprior =  atof( bpOpts.getArgs(switchInt)); break;
		case 8:   // Option 'memory'
			MEM_LIMIT = atof( bpOpts.getArgs(switchInt) ); break;
		case 9:   // Option 'scratchdir'
			SCRATCH_DIR = bpOpts.getArgs(switchInt); break;
		}
	}
	
	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || crystfile.length() == 0 || seqfile.length() == 0) {
		cerr << "Error: Must specify a working directory, crystal file, and sequence file.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}
	
	////////////////////
	//
	// read input files
	//
	Crystal crystData(crystfile);
	Sequence  seqData(seqfile);
	int nResSum = seqData.nres();
	vec3<int> dims = vec3<int>(0,0,0);
	
	string filestem = seqfile;
		filestem = filestem.substr(filestem.find_last_of('/')+1);
		filestem = filestem.substr(0,filestem.find_last_of('.'));
	string outpdbdir = workingdir + "/backbone_models/";
	string outmapdir = workingdir + "/posteriors/";
	string matchFormatString = workingdir + "/priors/";
	
	mkdir(outpdbdir.c_str(), 0755);
	mkdir(outmapdir.c_str(), 0755);
	
	//////////////////////
	///
	///  pretty-print
	cerr << "Running with inputs: " << endl;
	cerr << "    working directory = \'" << workingdir << "\'" << endl;
	cerr << "    crystal dims      = " << crystData.to_string() << endl;
	cerr << "            spacegp   = " << crystData.symmGpId << endl;
	cerr << "    protein sequence  = " << seqData.to_string() << endl;
	if (seleniumfile.length() > 0) {
		cerr <<  "    SE locations in   = " << seleniumfile << endl;
		cerr <<  "    SE prior prob.    = " << seleniumprior << endl;
	}
	
	cerr << endl;
	cerr << "Search parameters " << endl;
	cerr << "    sigma             = " << SIGMA << endl;
	cerr << "    M-estimate        = " << M_EST_ORIENT << endl;
	cerr << "    niters            = " << MAX_ITERS << endl;
	
	cerr << endl;
			
	////////////////////
	///
	///  if Se locations are provided, load them
	SElist SEs;
	if (seleniumfile.length() > 0) {
		SEs.load(seleniumfile);
		cerr << "Read " << SEs.nSe() << " seleniums\n";
	}
	
	
	////////////////////////////////////
	///
	/// Read match files
	///
	map< string, Belief > Prob0; // prior probabilities
	string mer;
	
	//
	// first get dims 
	for (int i=0; i<nResSum; i++) {
		if (seqData.serialGet(i) == '_') continue;  // gap
		
		mer = seqData.serialGet5mer(i);
		dims = Belief::getDims(matchFormatString, mer);
		break;
	}
	cerr << "    BP dims           = " << dims.to_string() << endl;
	crystData.buildMaps(dims);
	
	
	///////////////////////////////////
	///
	/// check if we need to page maps to disk
	///    --> if we DO, page that first map to disk
	double approxMemUsage = ( (nResSum*36.0) * crystData.getPtrMapInv().size() ) / (1024.0*1024.0) + 100.0;
	cerr << "Approximate memory needed: " << approxMemUsage << "MB\n";
	if (approxMemUsage > MEM_LIMIT) {
		PAGE_MAPS = true;
			
		// write to a unique sub-dir in SCRATCH_DIR
		srand(time(NULL));
		string uniq=".c"+itoc(rand());
		mkdir(SCRATCH_DIR.c_str(), 0755);   // ensure dirs exists
		SCRATCH_DIR = SCRATCH_DIR + "/" + uniq + "/";
		mkdir(SCRATCH_DIR.c_str(), 0755);
		
		cerr << "    page maps to \'" << SCRATCH_DIR << "\'\n";
	} else {
		cerr << "   keep maps in memory\n";
		PAGE_MAPS = false;
	}
	srand(1);
	
	////////////////////
	///
	/// cache LOCAL clique potentials psi_i->(i+1)
	///      and pairwise distances over the surface of a 3.8A sphere
	///
	getAdjPotentialList(dims,crystData);
	int n_adj_phis = phi_adj_idxs.size();
	
	
	////////////////////
	///
	/// build cost matrix for each orientation-neighbor pair
	///      if no angular info is available, assign uniform probability across rotations
	///
	vector<float> uniform_cost(n_adj_phis * crystData.nSymm()), 
	              uniform_inv_cost(n_adj_phis * crystData.nSymm());
	for (int j=0; j<n_adj_phis; j++) {
		uniform_cost[j] =    (1-M_EST_ORIENT) * phi_adj_phis[j] + (M_EST_ORIENT/n_adj_phis);
		uniform_inv_cost[j] = log(   uniform_cost[j]   );
		
		for (int symm_idx=1; symm_idx < crystData.nSymm(); ++symm_idx) {
			uniform_cost[ symm_idx*n_adj_phis + j]     = uniform_cost[j];
			uniform_inv_cost[ symm_idx*n_adj_phis + j] = uniform_inv_cost[j];
		}
	}
	phi_cost_cache[ 0 ] = uniform_cost;
	phi_inv_cost_cache[ 0 ] = uniform_inv_cost;	
	
	//
	// otherwise, build an orientation specific cost function
	//    precompute for each possible orientation (on this grid)
	for (int i=0; i<n_adj_phis; i++) {
		vector<float> cost_xi(n_adj_phis * crystData.nSymm()), 
		              inv_cost_xi(n_adj_phis * crystData.nSymm());
		
		double r_i = phi_adj_xs[i].r(), r_j;
		vec3<double> x_i, x_j;
		float dist_ij,cost_ij,dot_prod;

		x_i = phi_adj_xs[i]/r_i;
		
		for (int symm_idx=0; symm_idx < crystData.nSymm(); ++symm_idx) {
		
			float cost_sum=0.0, inv_cost_sum=0.0;
			
			for (int j=0; j<n_adj_phis; j++) {
				r_j = phi_adj_xs[j].r();
				x_j = (phi_adj_xs[j]/r_j).apply( crystData.getSymmInvRot(symm_idx) );
				
				dot_prod = dot( x_i , x_j );
				if (dot_prod > 1)  dot_prod =  1;
				if (dot_prod < -1) dot_prod = -1;
	
				// forward
				dist_ij = CA_CA_DIST * acos( dot_prod );  // distance on sphere surface
				cost_ij = phi_adj_phis[j] * exp( -square(dist_ij) / (2*square(SIGMA)) );
				cost_xi[ symm_idx*n_adj_phis + j ] = cost_ij ;
				cost_sum += cost_ij;
				
				// inverse
				dist_ij = CA_CA_DIST * acos( -dot_prod );  // distance on sphere surface
				cost_ij = phi_adj_phis[j] * exp( -square(dist_ij) / (2*square(SIGMA)) );
				inv_cost_xi[ symm_idx*n_adj_phis + j ] = cost_ij ;
				inv_cost_sum += cost_ij;
			}
			
			// normalize, log-trasform inv_*
			for (int j=0; j<n_adj_phis; j++) {
				cost_xi[ symm_idx*n_adj_phis + j]     =       (1-M_EST_ORIENT) *     cost_xi[symm_idx*n_adj_phis +j]/    cost_sum 
				                                            +    M_EST_ORIENT  * uniform_cost[j];
				inv_cost_xi[ symm_idx*n_adj_phis + j] = log ( (1-M_EST_ORIENT) * inv_cost_xi[symm_idx*n_adj_phis +j]/inv_cost_sum 
				                                            +    M_EST_ORIENT  * uniform_cost[j] );
															
			}
		}

		int idx = to_index(phi_adj_idxs[i] , dims);
		phi_cost_cache[idx]     = cost_xi;
		phi_inv_cost_cache[idx] = inv_cost_xi;
	}

	///
	/// Finally, read marginals
	///
	for (int i=0; i<nResSum; i++) {
		if (seqData.serialGet(i) == '_') continue;  // gap
		mer = seqData.serialGet5mer(i);

		// If we haven't seen this 5mer before, load it from disk
		if (Prob0.count(mer) == 0) {
			Prob0[mer] = Belief(matchFormatString, SEs, seleniumprior, mer, crystData);
			Prob0[mer].M.log_transform();
						
			if (PAGE_MAPS) Prob0[mer].dumpToDisk(SCRATCH_DIR+"/p0_"+mer);
		}
	}
	

	////////////////////
	//
	// Initialize model
	//      allocate space for belief, messages
	//      set initial belief to prior
	//      precompute backward & global msgs
	//
	cerr << "Initializing: \n";
	
	vector< compact_map_3d<float> >  b(nResSum);       // belief (projected)
	
	vector< AdjMessage >     msg_f(nResSum);   // forward  messages (from res i->(i+1))
	vector< AdjMessage >     msg_b(nResSum);   // backward messages (         i->(i-1))
	vector< compact_map_3d<float> >  msg_glob(nResSum); // global messages FROM res i(collision avoidance)

	compact_map_3d<double> msg_glob_prod;     // product of global messages
	msg_glob_prod.alloc(crystData, dims);
	msg_glob_prod.setAll(0.0);

	// temporary map variables
	compact_map_3d< float > b_i, b_j;

	// for each residue in each chain
	for (int i=0; i<nResSum; i++) {
		mer = seqData.serialGet5mer(i);
		cerr << mer[2]; cerr.flush();
		if (mer[2] == '_') continue;  // gap

		Belief *thisProb;
		thisProb = &Prob0[mer];

		// set belief to prior
		b[i].alloc(crystData, dims);
		
		if (PAGE_MAPS) thisProb->loadFromDisk(SCRATCH_DIR+"/p0_"+mer);
		
		for (int x=0; x<b[i].getNPts(); ++x) {
			b[i].d_set( x , (double) exp( thisProb->M.d_get(x) ) );
		}
		

		// pre-compute backwards messages
		/*
		*/
		if (mer[1] != '_') {
			compute_adj_message(msg_b[i], b[i], thisProb->phi_b, crystData);

			msg_b[i].M.normalize( M_EST_POSITIONS ) ;
			msg_b[i].M.log_transform();
			//msg_b[i].M.testIfContainsNaN("m_b_0");
			
			if (PAGE_MAPS) msg_b[i].dumpToDisk(SCRATCH_DIR+"/msgb"+itoc(i));
		}
		if (PAGE_MAPS) thisProb->dumpNoWrite();
		
		// allocate space for global messages
		msg_glob[i].alloc(crystData , dims);
		msg_glob[i].setAll(0);

		// log transform belief
		b[i].log_transform();
		//b[i].testIfContainsNaN("b_0");
			
		
		if (PAGE_MAPS) {
			msg_glob[i].dumpDataToDisk(SCRATCH_DIR+"/msgglob"+itoc(i)+".dat");
			b[i].dumpDataToDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
		}
	}

	// early trace??
	doTrace(string(outpdbdir+filestem+"_iter0.pdb"), b, seqData, crystData);

	//////////////////////////
	//
	// main loop!
	//
	double delta = INFINITY, RMSd;
	int iter=0, j_start, j_stop, j_inc;
	cerr << endl;
	while (delta > 1e-4*nResSum && ++iter <= MAX_ITERS) {   // limit # iterations
		cerr << "Iteration " << iter << ":\n";
		delta = 0;

		// alternate forward and backward passes
		if (iter%2) {
			j_start = 0; j_stop = nResSum; j_inc = 1;
		} else {
			j_start = nResSum-1; j_stop = -1; j_inc = -1;
		}

		//
		// for j=1..nres, compute b[j] as product (actually sum-of-logs) of:
		//       M[j]        --  prior belief
		//       msg_f[j-1]  --  msg from neighbor j-1
		//       msg_b[j+1]  --  msg from neighbor j+1
		//       msg_glob_prod/(msg_glob[j-1]*msg_glob[j]*msg_glob[j+1])
		//                   --  combined msg from all non-adjacent neighbors
		//
		for (int j=j_start; j != j_stop; j += j_inc) {
			mer = seqData.serialGet5mer(j);
			cerr << mer[2]; cerr.flush();
			if (mer[2] == '_') continue;  // gap
					
			// update global msg
			if (PAGE_MAPS) msg_glob[j].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
			map_minus(msg_glob_prod, msg_glob_prod, msg_glob[j]);
						
			// set b(j) to global msg * prior
			Belief& prob_j = Prob0[ mer ];
			
			if (PAGE_MAPS) prob_j.loadFromDisk(SCRATCH_DIR+"/p0_"+mer);
			
			map_plus( b_j, prob_j.M, msg_glob_prod );

			//  only one adjacent node has been updated since last iteration
			//  compute the updated message from this residue (i->j)
			int i=j-j_inc;
			
			if (mer[2-j_inc] != '_') {
				string mer_i = seqData.serialGet5mer(i);
				
				if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
				b_i = b[i];
				if (PAGE_MAPS) b[i].dumpNoWrite();
				
				Belief* prob_i = &Prob0[ mer_i ];
				if (PAGE_MAPS && mer_i != mer) {
					prob_i->loadFromDisk(SCRATCH_DIR+"/p0_"+mer_i);
				}
				
				
				if (j_inc > 0) {  
				// on a forward pass ...
					// page in
					if (PAGE_MAPS) {
						msg_b[j].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j));
						if (iter>1)
							msg_f[i].realloc();   // don't care what it is, just reallocate space
					}
					
					belief_subtract_and_proj(b_i, prob_i->phi_f, msg_b[j]);
					b_i.invlog_transform();
					
					compute_adj_message(msg_f[i], b_i, prob_i->phi_f, crystData);
					msg_f[i].M.normalize(  M_EST_POSITIONS  );
					msg_f[i].M.log_transform();
					//msg_f[i].M.testIfContainsNaN("m_f");
					
					// page out
					if (PAGE_MAPS) {
						msg_b[j].dumpNoWrite();
						msg_f[i].dumpToDisk(SCRATCH_DIR+"/msgf"+itoc(i));
					}
				} else {
				// on a backward pass ...
					// page in
					if (PAGE_MAPS) {
						msg_f[j].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j));
						msg_b[i].realloc();   // don't care what it is, just reallocate space
					}
					
					belief_subtract_and_proj(b_i, prob_i->phi_b, msg_f[j]);
					b_i.invlog_transform();
					compute_adj_message(msg_b[i], b_i, prob_i->phi_b, crystData);
					msg_b[i].M.normalize(  M_EST_POSITIONS  );
					msg_b[i].M.log_transform();
					//msg_b[i].M.testIfContainsNaN("m_b");
					
					// page out
					if (PAGE_MAPS) {
						msg_f[j].dumpNoWrite();
						msg_b[i].dumpToDisk(SCRATCH_DIR+"/msgb"+itoc(i));
					}
				}
				
				if (PAGE_MAPS && mer_i != mer) {
					prob_i->dumpNoWrite();
				}
			}

			// aggregate incoming messages from adjacent residues
			if (mer[1] != '_') {
				// page in
				if (PAGE_MAPS) {
					msg_glob[j-1].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j-1)+".dat");
					msg_f[j-1].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j-1));
				}
				
				map_minus( b_j, b_j, msg_glob[j-1] );
				belief_add_and_proj ( b_j, prob_j.phi_b, msg_f[j-1] );
				
				// page out
				if (PAGE_MAPS) {
					msg_glob[j-1].dumpNoWrite();
					msg_f[j-1].dumpNoWrite();
				}
			}
			if (mer[3] != '_') {
				// page in 
				if (PAGE_MAPS) {
					msg_glob[j+1].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j+1)+".dat");
					msg_b[j+1].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j+1));
				}
					
				map_minus( b_j, b_j, msg_glob[j+1] );
				belief_add_and_proj ( b_j, prob_j.phi_f, msg_b[j+1] );
				
				// page out
				if (PAGE_MAPS) {
					msg_glob[j+1].dumpNoWrite();
					msg_b[j+1].dumpNoWrite();
				}
			}
			if (PAGE_MAPS) prob_j.dumpNoWrite();

			// normalize belief
			b_j.logNormalize( M_EST_POSITIONS );
			//b_j.testIfContainsNaN("b_j");
			
			// compute change since last iteration
			if (PAGE_MAPS) b[j].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
			double this_delta = map_delta(b_j,b[j]);
			
			if (isnan(this_delta)) {
				cerr << "isNaN(delta[" << j << "]) == true!\n";
				cerr << "     " << b_j.sum() << " , " << b[j].sum() << " , " 
				     << msg_glob[j-1].sum() <<  " , " << msg_glob[j].sum() <<  " , " << msg_glob[j+1].sum() << " , " 
					 << msg_f[j-1].M.sum() << " , " << msg_b[j+1].M.sum() << endl;
				exit(1);
			}
			
			delta += this_delta;
			
			b[j] = b_j;
			if (PAGE_MAPS) b[j].dumpDataToDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
			
			// occupancy message 
			b_j.invlog_transform();
			compute_occ_message(iter, msg_glob[j], b_j, crystData);  // normalizes + log transforms msg
			//msg_glob[j].testIfContainsNaN("glob_j");

			map_plus ( msg_glob_prod, msg_glob_prod, msg_glob[j] );
			if (PAGE_MAPS) msg_glob[j].dumpDataToDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");		
			
		}
		cout << endl;
		cout << "Sum change in belief = " << delta << endl << endl;
		
		doTrace (string(outpdbdir+filestem+"_iter"+itoc(iter)+".pdb"), b, seqData, crystData);
	}

	////////////////////
	//
	// calculate final trace, error
	//    - oversample the maps, and ensure adjacent CA's are 3.8A apart
	//    - also write final marginals to disk
	//
	doTrace (string(outpdbdir+filestem+"_final.pdb"), b, seqData, crystData);
	
	int resId;
	char resType, chnId;
	for (int i=0; i < nResSum; i++) {
		if (b[i].getNPts() > 0) {
			seqData.serialGet(i, resId, resType, chnId);
			
			if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
			writeMapToDisk(b[i], string(outmapdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal"));
			if (PAGE_MAPS) b[i].dumpNoWrite();
		}
	}
	
	// clean up 
	if (PAGE_MAPS) {
		DIR *pdir;
		struct dirent *pent;
		
		pdir=opendir(SCRATCH_DIR.c_str()); //"." refers to the current dir
		if (!pdir){
			printf ("opendir() failure; terminating");
			exit(1);
		}
		while ( pent = readdir(pdir) ){
			//cerr << "  deleting " << pent->d_name << endl;
			unlink( (SCRATCH_DIR+pent->d_name).c_str() );
		}
  		closedir(pdir);
		
  		rmdir(SCRATCH_DIR.c_str());
	}
}
