// //////////////////////////////////////////////////////////////////////////
//
//     shfear.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////

#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

extern "C" {
#include "fftw3.h"
#include "complex.h" 
#include "csecond.h"
#include "so3_correlate_fftw.h"
#include "soft_fftw_pc.h"
#include "soft_fftw.h"
#include "makeWigner.h"
#include "makeweights.h"
#include "FST_semi_memo.h"
#include "cospmls.h"
#include "primitive_FST.h"
#include "seminaive.h"
}

#include "spharm.h"
#include "getopts.h"   // parse command line args

int POSITIONAL_COST[] = {1,1,1,1,1};  // allow ANY position to change
int MAX_PAMDIST = 20;


///
/// main()
///
int main(int argc, char** argv) {
	srand(1);

	/////////////////////
	/////////////////////
	///
	/// read input files
	Options shfearOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	shfearOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	shfearOpts.addOption("m",    "map",          "[REQUIRED] Input fs4 map file",      true);
	shfearOpts.addOption("c",    "cryst",        "[REQUIRED] Input crystal file",      true);
	shfearOpts.addOption("s",    "seq",          "[REQUIRED] Input sequence file",      true);
	shfearOpts.addOption("r",    "reso",          "Density map resolution (default 2.0)",      true);
	shfearOpts.addOption("B",    "bandwidth",     "Spherical harmonic bandwidth (default 12)",      true);
	shfearOpts.addOption("R",    "radius",        "Fragment radius (default 5)",      true);
	shfearOpts.addOption("N",    "nfrags",        "Number of fragments considered per amino acid (default 50)",      true);
	shfearOpts.addOption("x",    "grid",          "Approximate map grid spacing (default 1A)",      true);
	shfearOpts.addOption("o",    "omit",          "Using density values, omit fraction of map (default 0.8)",     true);
	shfearOpts.addOption("p",    "parallel",      "Parallelize <job#> of <Njobs>",      true);
	shfearOpts.addOption("a",    "analogues",     "List of structural analogues for fragment generation",      true);
	shfearOpts.addOption("f",    "force",         "Force priors to be overwritten",     false);
		
	string workingdir, mapfile, crystfile, seqfile, seleniumfile;
	vector< string >  similar_files;
	vector< CAtrace > similar_pdbs;
	bool forceOverwrite = false;
	
	// defaults
	double grid=1.0, CC_CUTOFF=1.0, percentile=0.8, seleniumprior=0.9;
	int beginAtRes=0, skipByRes=1, Rmax=5, nfrags=50, nSimilar=0;
	Bout = B = 12;
	reso = 2.0;
	
	shfearOpts.parse(argc, argv);
	while ((switchInt = shfearOpts.cycle()) >= 0) {
		vector<string> pargs;
		Crystal nullCryst;
		
		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  shfearOpts.getArgs(switchInt); break;
		case 1:   // Option 'map'
			mapfile =  shfearOpts.getArgs(switchInt); break;
		case 2:   // Option 'cryst'
			crystfile =  shfearOpts.getArgs(switchInt); break;
		case 3:   // Option 'seq'
			seqfile =  shfearOpts.getArgs(switchInt); break;
		case 4:   // Option 'reso'
			reso = atof( shfearOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'bandwidth'
			B = Bout = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 6:   // Option 'radius'
			Rmax = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 7:   // Option 'nfrags'
			nfrags = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 8:   // Option 'grid'
			grid = atof( shfearOpts.getArgs(switchInt) ); break;
		case 9:   // Option 'omit'
			percentile = atof( shfearOpts.getArgs(switchInt) ); break;
		case 10:   // Option 'parallel'
			Tokenize( shfearOpts.getArgs(switchInt) , pargs );
			if (pargs.size() == 2) {
				skipByRes = atoi(pargs[1]);
				beginAtRes = pos_mod( atoi(pargs[0])-1 , skipByRes );
			}
			break;
		case 11:  // Option 'analogues'
			Tokenize( shfearOpts.getArgs(switchInt) , similar_files );
			nSimilar = similar_files.size();
			for (int i=0; i<nSimilar; ++i)
				similar_pdbs.push_back( CAtrace( similar_files[i] , nullCryst ) );
			break;
		case 12:  // Option 'force'
			forceOverwrite = true;
		}
	}
	
	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || mapfile.length() == 0 || crystfile.length() == 0 || seqfile.length() == 0) {
		cerr << "Error: Must specify a working directory, fs4 map file, crystal file, and sequence file.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}
	
	map_3d<double> origMap;
	readDensityMap(origMap, mapfile);
	Crystal crystData(crystfile);
	Sequence seq(seqfile);

	// density map grid
	Rstep = 1.0;
	vec3<int> dims = crystData.getDims(grid);
	int dimsProduct = dims[0]*dims[1]*dims[2];
	crystData.buildMaps(dims);

	// resample map
	map_3d<double> map(dims);
	resize_fft(origMap, map);
	
	// output directories
	string outdir = workingdir+"/priors/";
	//string outdir = workingdir+"/priors.shfear"+itoc(B)+"/";
	mkdir(workingdir.c_str(), 0755);   // ensure dirs exists
	mkdir(outdir.c_str(),     0755);

	//////////////////////
	///
	///  pretty-print
	if (skipByRes > 1) {
		cerr << "*** PARALLEL RUN #" << beginAtRes+1 << " of " << skipByRes << " *** " << endl;
		cerr << endl;
	}
	cerr << "Running with inputs: " << endl;
	cerr << "    working directory = \'" << workingdir << "\'" << endl;
	cerr << "    fs4-format map    = <" << mapfile << ">" << endl;
	cerr << "    crystal dims      = " << crystData.to_string() << endl;
	cerr << "            spacegp   = " << crystData.symmGpId << endl;
	cerr << "    protein sequence  = " << seq.to_string() << endl;
	cerr << "    resolution        = " << reso << endl;
	
	cerr << endl;
	cerr << "Search parameters " << endl;
	cerr << "    B                 = " << B << endl;
	cerr << "    radius            = " << Rmax << endl;
	cerr << "    nfrags            = " << nfrags << endl;
	
	cerr << endl;
	if (nSimilar > 0) {
		cerr << "Using similar maps: " << endl;
		for (int i=0; i<nSimilar; ++i) 
			cerr << "    " << similar_files[i] << endl;
	}
		
	// only consider the top 10% density locations
	percentile = min(1.0,percentile);
	percentile = max(percentile,0.0);
	
	vector<double> projection = map.project(crystData.getAsymmUnit());
	sort(projection.begin(), projection.end());
	double thresh = projection[(int)floor(percentile*projection.size()) - 1];
	
	cerr << "Unit cell grid <" << dims.to_string() << ">" << endl;
	cerr << "    In asymm unit : " << projection.size() << endl;
	cerr << "    With density > " << thresh << "(=" << floor(100*percentile+0.5) << " %ile): " 
	     << ceil((1-percentile)*projection.size()) << endl;
	cerr << endl;
	
	// map spherical harmonics
	map_3d<double> sigR(Rmax,2*B,2*B);
	map_3d<double> sigCoefR(Rmax,B,B);
	map_3d<double> sigCoefI(Rmax,B,B); 
	
	map_3d< cubicSplineCoeff3d > sphPtrsMap(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsMap, dims, crystData);
	
	// build tiny unit cells
	//    for constructing maps from fragments
	//    (we don't want to build over the entire crystal)
	double radiusSmall = Rmax * Rstep + 1.0;
	double cellEdgeSmall = 2.0 * radiusSmall + 12.0;
	int nRadsteps;
	double deltaX = min(1.0, reso/3.0);

	// small -- for DB lookup
	nRadsteps = (int)ceil( (cellEdgeSmall) / deltaX );
	Crystal crystSmall(cellEdgeSmall,cellEdgeSmall,cellEdgeSmall, 90.00,90.00,90.00);
	vec3<int> dimsSmall(nRadsteps, nRadsteps, nRadsteps);
	map_3d< cubicSplineCoeff3d > sphPtrsSmall(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsSmall, dimsSmall, crystSmall);
	
	SOFT_initialize();
	//
	///////////////////

	///////////////////
	//
	//  for each residue, consider region centered on that residue
	//      cluster all instances of that 5mer, saving centroids
	//      extract spherical harmonics
	vector< Nmer >               centroids;
	vector< map_3d<double> >     coeffR_centroids;
	vector< map_3d<double> >     coeffI_centroids;
	vector< double >             vars;
	
	for (int i_res = beginAtRes; i_res < seq.nres(); i_res+=skipByRes) {
		if (seq.serialGet(i_res) == '_') continue;  // gap

		string mer = seq.serialGet5mer(i_res);
		
		string outfile = outdir + "match_" + mer;
		if (exists( string(outfile+".cc.mat") )) {
			if (!forceOverwrite) {
				cerr << outfile << " exists, continuing (use '-f' to force overwrite)\n";
				continue;
			} else {
				cerr << outfile << " exists, will be overwritten\n";
				continue;
			}
		}

		//
		// do DB lookup and clustering to find nmer
		centroids.clear();		
		
		// first, lookup in file
		if (nSimilar > 0) {
			spharm_lookupFromFilelist(mer, similar_files, similar_pdbs, radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, 
			                          centroids, coeffR_centroids, coeffI_centroids);
		}
		// then, ONLY IF no nmers are found in the PROVIDED pdb file
		//    should we go to the database
		if (centroids.size() == 0) {
			spharm_lookup(mer, nfrags, radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, 
			              centroids, coeffR_centroids, coeffI_centroids);
		}
		
		int nfragments = centroids.size();
		if (nfragments == 0) { cerr << "WARNING! No clusters found for \'" << mer << "\'.\n"; continue; }
		cerr << "Searching density map for " << nfragments
		     << " templates corresponding to sequence \'" << mer << "\'" << endl;
		
		//
		// compute SphHarmonics for each centroid fragment
		map_3d< orientation > matchOrient( dims , orientation(-10,-10,-10,-10) );  // outside normal range of beta, gamma
		map_3d< double >      matchMax( dims, 0 );
		
		vec3<int> X;
		int sumAsymm = (int)crystData.getAsymmUnit().sum(), counter = 0;
		
		double CCsum=0.0, CCsum2=0.0;
		int nCC=0;
		
		for (X[0]=0; X[0]<dims[0]; ++X[0]) { 
			for (X[1]=0; X[1]<dims[1]; ++X[1]) {
				for (X[2]=0; X[2]<dims[2]; ++X[2]) {
					if (crystData.getAsymmUnit().get(X[0],X[1],X[2]) == 0) continue;
					if (map.get(X[0],X[1],X[2]) < thresh) continue;
					
					// 1) extract
					spherical_resample_cubic_spline(sigR, map, sphPtrsMap, X);
					
					// 2) standardize 
					sph_standardize(sigR);
					
					// 3) spharm
					SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);
					
					// 4) compare to each template fragment
					for (int i_frag=0; i_frag < nfragments; i_frag++) {
						Nmer &frag = centroids[i_frag];
			
						vec3<double> CAminus1 = (mer[1] == '_' ? vec3<double>( 3.8,0.0,0.0) : frag.getCA(1));
						vec3<double> CAplus1  = (mer[3] == '_' ? vec3<double>(-3.8,0.0,0.0) : frag.getCA(3));			
						
						SOFT_so3_correlate(sigCoefR,sigCoefI, coeffR_centroids[i_frag],coeffI_centroids[i_frag]);
		
						// find best rotation, (correlation) for THIS cluster
						double maxval = -INFINITY, thisVal;
						int maxloc = -1;
						for (int i=0 ; i<8*Bout*Bout*Bout ; ++i) {
							thisVal = so3SigSum.get(i).real();
							CCsum += thisVal;
							CCsum2 += thisVal*thisVal;
							nCC++;
							if (thisVal >= maxval) {
								maxval = thisVal; maxloc = i ;
							}
						}
						
						// now compare to other clusters
						if ( maxval > matchMax.get(X[0],X[1],X[2]) ) {
							matchMax.set(X[0],X[1],X[2],  maxval);
							
							mat33<double> R_abg;
							SOFT_idx_to_rot(maxloc,R_abg);
							
							vec3<double> R_CAm1 = CAminus1.apply(R_abg);
							vec3<double> R_CAp1 = CAplus1.apply(R_abg);
							
							if (    isnan(R_CAm1.beta()) || isnan(R_CAm1.gamma()) 
							     || isnan(R_CAp1.beta()) || isnan(R_CAp1.gamma()) ) {
								cerr << "ERROR! NaN in orientation\n";
								cerr << "     " << CAminus1.to_string() << " --> " << R_CAm1.to_string() 
								       << " --> (" << R_CAm1.beta() << "," << R_CAm1.gamma() << ")" << endl;
								cerr << "     " << CAplus1.to_string() << " --> " << R_CAp1.to_string() 
								       << " --> (" << R_CAp1.beta() << "," << R_CAp1.gamma() << ")" << endl;
								exit(1);
							}
							
							matchOrient.set(X[0],X[1],X[2],
							    orientation(R_CAm1.beta(),R_CAm1.gamma(),R_CAp1.beta(),R_CAp1.gamma()));
						}
					} // end loop over clusters
				}
			}
		}  // end loop over map  
		
		
		//
		// symmetric transfer
		map_3d<int> ptrMap(dims), symMap(dims);
		crystData.maxSymmetric(matchMax, ptrMap, symMap);
		for (int j=0; j<dimsProduct; j++) {
			int jptr = ptrMap.get(j);
			if (jptr == j) continue;
			orientation &j_orient = matchOrient.get(j);
			orientation &jptr_orient = matchOrient.get(jptr);
			
			// forward
			vec3<double> x_F ( cos(jptr_orient.gammaF)*sin(jptr_orient.betaF),
			                   sin(jptr_orient.gammaF)*sin(jptr_orient.betaF),
			                   cos(jptr_orient.betaF) );
			x_F = x_F.apply(crystData.c2f);     // convert to fractional
			x_F = x_F.apply(crystData.getSymmInvRot( symMap.get(j)));  // rotate
			x_F = x_F.apply(crystData.f2c);     // convert back to cartesian
			
			// backward
			vec3<double> x_B ( cos(jptr_orient.gammaB)*sin(jptr_orient.betaB),
			                   sin(jptr_orient.gammaB)*sin(jptr_orient.betaB),
			                   cos(jptr_orient.betaB) );
			x_B = x_B.apply(crystData.c2f);     // convert to fractional
			x_B = x_B.apply(crystData.getSymmInvRot( symMap.get(j)));  // rotate
			x_B = x_B.apply(crystData.f2c);     // convert back to cartesian
			
			j_orient.betaF = x_F.beta();
			j_orient.gammaF = x_F.gamma();
			j_orient.betaB = x_B.beta();
			j_orient.gammaB = x_B.gamma();
		}

		//
		// write to disk
		string outfilestem = outdir + "match_" + mer;
		writeMapToDisk (matchMax   , string(outfilestem + ".cc"));
		writeMapsToDisk(matchOrient, outfilestem);
	}  // end loop over residues
}
