// //////////////////////////////////////////////////////////////////////////
//
//     sidechain_naive.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////

#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>

#include "spharm.h"
#include "getopts.h"   // parse command line args

int POSITIONAL_COST[] = {1,1,9999,1,1};  // don't let middle position change
int MAX_PAMDIST = 20;


static map_3d< double >* getAdjPotential(vec3<int> dims , Crystal &crystData) {
	static map_3d< double > map;
	srand(1);
	
	if (map.getDimsProd() == 0) {
		vec3<double> fX, fXovr;  // frac coords
		double pR;        // polar 'R'
	
		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(crystData.f2c).norm2() )
		);
	
		// use random sampling
		int NUM_SAMPLES = 100000;
		
		map.alloc(dims);
		map.setAll(0);
		
		for (int i=0; i<NUM_SAMPLES; ++i) {
			vec3<double> fracX(
						((double)rand()/RAND_MAX-0.5)/dims[0],
						((double)rand()/RAND_MAX-0.5)/dims[1],
						((double)rand()/RAND_MAX-0.5)/dims[2]); // start
			
			double distXY = 3.802525 + 0.05084*randnorm();   // dist
			
			// direction
			vec3<double> dirXY;
			double del;
			do {
				dirXY[0] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[1] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[2] = 2.0*(double)rand()/RAND_MAX-1.0;
				del = dirXY.norm2();
			} while (del >= 1.0 || del == 0);
			
			del = distXY / sqrt(del);
			dirXY = dirXY * del;
			
			vec3<double> fracY = fracX + dirXY.apply(crystData.c2f);
			vec3<int> idxY( pos_mod( (int)floor(fracY[0]*dims[0]+0.5) , dims[0]),
							pos_mod( (int)floor(fracY[1]*dims[1]+0.5) , dims[1]),
							pos_mod( (int)floor(fracY[2]*dims[2]+0.5) , dims[2]) );
							
			map.inc(idxY[0],idxY[1],idxY[2], 1.0);	
		}
		
		// ensure symmetry
		for (int i=0; i<dims[0]; i++)
		for (int j=0; j<dims[1]; j++)
		for (int k=0; k<dims[2]; k++) {
			fX[0] = ((double)i)/dims[0];  if (fX[0]>.5) fX[0] -=1;
			fX[1] = ((double)j)/dims[1];  if (fX[1]>.5) fX[1] -=1;
			fX[2] = ((double)k)/dims[2];  if (fX[2]>.5) fX[2] -=1;
	
			vec3<double> sX;
			vec3<int> sIdx;
			
			double map_ijk = map.get(i,j,k);
			if (map_ijk == 0) continue;   // ???
			
			for (int s=1; s<crystData.nSymm(); ++s) {
				sX = fX.apply( crystData.getSymmRot(s) );
				sIdx[0] = pos_mod( (int)floor(sX[0]*dims[0]+0.5) , dims[0]);
				sIdx[1] = pos_mod( (int)floor(sX[1]*dims[1]+0.5) , dims[1]);
				sIdx[2] = pos_mod( (int)floor(sX[2]*dims[2]+0.5) , dims[2]);
				
				map_ijk = max(map_ijk, map.get(sIdx[0],sIdx[1],sIdx[2]));
			}
	
			map.set(i,j,k, map_ijk);
		}
		
		map.normalize();  // sum to 1
		
		//writeMapToDisk(map, "phiAdj");
	}
	
	return &map;
}


///
/// main()
///
int main(int argc, char** argv) {
	srand(1);

	/////////////////////
	///
	/// read input files
	Options shfearOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	shfearOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	shfearOpts.addOption("m",    "map",          "[REQUIRED] Input fs4 map file",      true);
	shfearOpts.addOption("p",    "pred",          "[REQUIRED] Predicted CA-trace",      true);
	shfearOpts.addOption("r",    "reso",          "Density map resolution",      true);
	shfearOpts.addOption("B",    "bandwidth",     "Spherical harmonic bandwidth (default 16)",      true);
	shfearOpts.addOption("g",    "rad-align",     "Frag radius for alignment (default 5)",      true);
	shfearOpts.addOption("k",    "rad-mask",      "Frag radius for computing CC (default 8)",      true);
	shfearOpts.addOption("d",    "mask-dist",     "Mask dist for computing CC (default 3)",      true);
	shfearOpts.addOption("N",    "nfrags",        "#fragments per amino acid (default 50)",      true);
	shfearOpts.addOption("a",    "analogues",     "Structural analogue list for fragment generation",      true);
	
	string workingdir, mapfile, crystfile, predCAfile;
	vector< string >  similar_files;
	vector< CAtrace > similar_pdbs;
	
	// defaults
	double MASK_DIST=3.0;
	int Ralign=5, Rmax=8, NEXAMPLES=50, nSimilar=0;
	Bout = B = 16;
	reso = 2.0;
	
	shfearOpts.parse(argc, argv);
	while ((switchInt = shfearOpts.cycle()) >= 0) {
		vector<string> pargs;
		Crystal nullCryst;
		
		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  shfearOpts.getArgs(switchInt); break;
		case 1:   // Option 'map'
			mapfile =  shfearOpts.getArgs(switchInt); break;
		case 2:   // Option 'pred-CA'
			predCAfile =  shfearOpts.getArgs(switchInt); break;
		case 3:   // Option 'reso'
			reso = atof( shfearOpts.getArgs(switchInt) ); break;
		case 4:   // Option 'bandwidth'
			B = Bout = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'rad-align'
			Ralign = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 6:   // Option 'rad-mask'
			Rmax = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 7:   // Option 'mask-dist'
			MASK_DIST = atof( shfearOpts.getArgs(switchInt) ); break;
		case 8:   // Option 'nfrags'
			NEXAMPLES = atoi( shfearOpts.getArgs(switchInt) ); break;
		case 9:  // Option 'analogues'
			Tokenize( shfearOpts.getArgs(switchInt) , similar_files );
			nSimilar = similar_files.size();
			for (int i=0; i<nSimilar; ++i)
				similar_pdbs.push_back( CAtrace( similar_files[i] , nullCryst) );
			break;
		}
	}

	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || mapfile.length() == 0 || predCAfile.length() == 0) {
		cerr << "Error: Must specify a working directory, fs4 map file, and a CA trace.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}
	
	// read input files
	map_3d<double> origMap;
	readDensityMap(origMap, mapfile);
	Crystal crystData;
	CAtrace prot(predCAfile,crystData);
	int nResSum = prot.nres(), NFRAGS;
	
	if (Ralign > Rmax) {
		cerr << "WARNING: Ralign > Rmax\n";
		Ralign=Rmax;
	}
	Rstep = 1.0;
	
	chem_db_initialize();
	
	// alignment dist
	double workingMAX_CA_DEV = 3.8 * sqrt(2.0) * M_PI / B;   // the furthest a CA can deviate
	bool FORCE_ALIGN_CA = true;
	
	map_3d<double> map;
	vec3<int> dims = crystData.getDims(1.0);
	int dimsProduct = dims[0]*dims[1]*dims[2];
	crystData.buildMaps(dims);
	map.alloc(dims[0],dims[1],dims[2]);
	resize_fft(origMap, map); // resample map
		
	// map spherical harmonics
	map_3d<double> sigR(Rmax,2*B,2*B), sigCoefR(Rmax,B,B), sigCoefI(Rmax,B,B); 
	map_3d<double> sig2R(Rmax,2*B,2*B), sig2CoefR(Rmax,B,B), sig2CoefI(Rmax,B,B); 
	
	map_3d< cubicSplineCoeff3d > sphPtrsMap(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsMap, dims, crystData);
	map_3d< linearCoeff3d > linPtrsMap(Rmax,2*B,2*B);
	cache_spherical_resampling_linear(linPtrsMap, dims, crystData);
	
	// build tiny unit cell
	//    for constructing maps from fragments
	//    (we don't want to build over the entire crystal)
	double radiusSmall = Rmax * Rstep + 1.0;
	double cellEdgeSmall = 2.0 * radiusSmall + 12.0;
	int nRadsteps;
	double deltaX = min(1.0, reso/3.0);

	nRadsteps = (int)ceil( (cellEdgeSmall) / deltaX );
	Crystal crystSmall(cellEdgeSmall,cellEdgeSmall,cellEdgeSmall, 90.00,90.00,90.00);
	vec3<int> dimsSmall(nRadsteps, nRadsteps, nRadsteps);
	
	map_3d< cubicSplineCoeff3d > sphPtrsSmall(Rmax,2*B,2*B);
	cache_spherical_resampling_cubic_spline(sphPtrsSmall, dimsSmall, crystSmall);
	map_3d< linearCoeff3d > linPtrsSmall(Rmax,2*B,2*B);
	cache_spherical_resampling_linear(linPtrsSmall, dimsSmall, crystSmall);

	SOFT_initialize();

	vector< Nmer >               frags;
	vector< map_3d<double> >     tmpCoefsR, tmpCoefsI;
	vector< map_3d<double> >     mskCoefsR, mskCoefsI;

	int atmCounter = 0;
	
	// open outfile
	string outfile = workingdir + "/naive_allatom_model.pdb";
	ofstream out( outfile.c_str() );
	out << crystData.crystline << endl;;
	
	for (int i_res = 0; i_res < prot.nres(); i_res++) {
		string mer = prot.serialGet5mer(i_res);
		if (mer[2] == '_') continue;  // gap
		Residue thisRes;
		thisRes.restype = AA2ID[ mer[2] ];

		// do database lookup for 'mer' instances
		// first, lookup in file (if provided)
		frags.clear();
		if (nSimilar > 0) {
			if (MASK_DIST > 0) {
				spharm_lookupAndMaskFromFilelist(mer, 
										similar_files, similar_pdbs, 
										radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, linPtrsSmall,
										frags, 
										tmpCoefsR, tmpCoefsI, 
										mskCoefsR, mskCoefsI,
										MASK_DIST, Ralign);
			} else {
				spharm_lookupFromFilelist(mer, 
				                    similar_files, similar_pdbs, 
							        radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, 
							        frags, 
							        tmpCoefsR, tmpCoefsI);
			}
		}
								
		if (MASK_DIST > 0) {
			spharm_lookupAndMask(mer, 
						NEXAMPLES, 
						radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, linPtrsSmall,
						frags, 
						tmpCoefsR, tmpCoefsI, 
						mskCoefsR, mskCoefsI,
						MASK_DIST, Ralign);
		} else {
			spharm_lookup(mer, 
			       NEXAMPLES,
			       radiusSmall, dimsSmall, crystSmall, sphPtrsSmall, 
			       frags, 
			       tmpCoefsR, tmpCoefsI);
		}				
							
		int nfragments = frags.size();
		if (nfragments == 0) { cerr << "ERROR! No fragments found.\n"; continue; }

		// find the predicted CA location
		vec3<double> trueLocfX = prot.serialGet(i_res).fracX;
		vec3<int> trueLocIdx;
		trueLocIdx[0] = pos_mod(int(floor(trueLocfX[0]*dims[0] + .5)),dims[0]);
		trueLocIdx[1] = pos_mod(int(floor(trueLocfX[1]*dims[1] + .5)),dims[1]);
		trueLocIdx[2] = pos_mod(int(floor(trueLocfX[2]*dims[2] + .5)),dims[2]);

		// aligh each frag to predicted backbone
		vec3<double> pred_CAminus1(0.0,0.0,0.0), pred_CAplus1(0.0,0.0,0.0);
		if (mer[1]!='_') pred_CAminus1 = prot.serialGet(i_res-1).cartX - prot.serialGet(i_res).cartX;
		if (mer[3]!='_') pred_CAplus1 = prot.serialGet(i_res+1).cartX - prot.serialGet(i_res).cartX;
	
		vec3<int>    delIdx_p1, delIdx_m1;
		bool align_p1=false, align_m1=false;
		
		if (mer[1]!='_') {				
			delIdx_m1[0] = pos_mod( (int)(floor(prot.serialGet(i_res-1).fracX[0]*dims[0]+0.5) 
			                           - floor( prot.serialGet(i_res).fracX[0]*dims[0]+0.5)) , dims[0]);
			delIdx_m1[1] = pos_mod( (int)(floor(prot.serialGet(i_res-1).fracX[1]*dims[1]+0.5) 
			                           - floor( prot.serialGet(i_res).fracX[1]*dims[1]+0.5)) , dims[1]);
			delIdx_m1[2] = pos_mod( (int)(floor(prot.serialGet(i_res-1).fracX[2]*dims[2]+0.5)
			                           - floor( prot.serialGet(i_res).fracX[2]*dims[2]+0.5)) , dims[2]);
								
			if ( getAdjPotential(dims , crystData)->get(delIdx_m1[0],delIdx_m1[1],delIdx_m1[2]) > 1e-6 ) 
				align_m1 = true;
		}
		if (mer[3]!='_') {				
			delIdx_p1[0] = pos_mod( (int)(floor(prot.serialGet(i_res+1).fracX[0]*dims[0]+0.5) 
			                           - floor( prot.serialGet(i_res).fracX[0]*dims[0]+0.5)) , dims[0]);
			delIdx_p1[1] = pos_mod( (int)(floor(prot.serialGet(i_res+1).fracX[1]*dims[1]+0.5) 
			                           - floor( prot.serialGet(i_res).fracX[1]*dims[1]+0.5)) , dims[1]);
			delIdx_p1[2] = pos_mod( (int)(floor(prot.serialGet(i_res+1).fracX[2]*dims[2]+0.5)
			                           - floor( prot.serialGet(i_res).fracX[2]*dims[2]+0.5)) , dims[2]);
								
			if ( getAdjPotential(dims , crystData)->get(delIdx_p1[0],delIdx_p1[1],delIdx_p1[2]) > 1e-6 ) 
				align_p1 = true;
		}
		
		//
		// check for an unsatisfiable angle between the two
		vec3<double> x_i, x_j;
		double p_angle, dot_prod;
		if ( mer[1]!='_' && mer[3]!='_' ) {
			x_i = pred_CAminus1 / pred_CAminus1.r();
			x_j = pred_CAplus1 / pred_CAplus1.r();
			dot_prod = dot( x_i , x_j ); if (dot_prod > 1) dot_prod =  1; if (dot_prod < -1) dot_prod = -1;
			p_angle = (180.0/M_PI) * acos( dot_prod );  // distance on sphere surface
			
			if (p_angle < 75.0) {  // unsatisfiable
				align_m1 = align_p1 = false;
			}
		}
		
		// resample concentric spheres around 'trueLocIdx'
		// then do spherical harmonic transform	
		spherical_resample_cubic_spline(sigR, map, sphPtrsMap, trueLocIdx, Rmax);
		sph_standardize(sigR);
		map_times(sig2R, sigR, sigR);
				
		SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);
		SOFT_sharm_transform(sig2R, sig2CoefR, sig2CoefI);
					
		double bestMatch = -INFINITY, thisVal, bestAlignMatch;
		int bestFragIdx;
		mat33<double> thisRot, bestRot, bestAlignRot;
		vec3<double> thisEulerAngles, bestAngles, bestAlignAngles;
		
		// rotated spharms
		map_3d<double> tmpRotCoefsR(Rmax,B,B), 
		               tmpRotCoefsI(Rmax,B,B), 
					   mskRotCoefsR(Rmax,B,B), 
					   mskRotCoefsI(Rmax,B,B);
		
		for (int frag_idx=0; frag_idx < nfragments; frag_idx++) {
			Nmer &frag = frags[frag_idx];
			vec3<double> tmpl_CAminus1 = frag.getCA(1),
						 tmpl_CAplus1  = frag.getCA(3);
			
			if ( FORCE_ALIGN_CA && align_m1 && align_p1 ) {
			
				vector< vec3<double> > CAs(5 , vec3<double>(0,0,0) );
				CAs[1] = pred_CAminus1;
				CAs[3] = pred_CAplus1;
				
				align(CAs, frag, bestAlignRot, thisVal);
				bestAlignAngles = rot2euler(bestAlignRot);
				
				SOFT_so3_rotate( tmpRotCoefsR, tmpRotCoefsI, tmpCoefsR[frag_idx], tmpCoefsI[frag_idx],
								bestAlignAngles[2],bestAlignAngles[1],bestAlignAngles[0]);
								
				if (MASK_DIST > 0) {
					// compute masked CC
					SOFT_so3_rotate( mskRotCoefsR, mskRotCoefsI, mskCoefsR[frag_idx], mskCoefsI[frag_idx],
									bestAlignAngles[2],bestAlignAngles[1],bestAlignAngles[0]);
					
					thisVal = SOFT_S2_correlate( sigCoefR, sigCoefI, tmpRotCoefsR, tmpRotCoefsI, mskRotCoefsR, mskRotCoefsI );
				} else {
					thisVal = SOFT_S2_correlate( sigCoefR, sigCoefI, tmpRotCoefsR, tmpRotCoefsI );
				}
				
				if (thisVal > bestMatch) {
					bestMatch = thisVal;
					bestRot = bestAlignRot;
					bestAngles = bestAlignAngles;
					bestFragIdx = frag_idx;
				}

			} else {
			
				SOFT_so3_correlate(
							sigCoefR,             sigCoefI, 
							tmpCoefsR[frag_idx], tmpCoefsI[frag_idx],
							Ralign);
				
				// search for best rotation			  
				bestAlignMatch = -INFINITY;
				
				for (int i=0 ; i<8*Bout*Bout*Bout ; ++i) {
					thisVal = so3SigSum.get(i).real();
					
					if ( thisVal > bestAlignMatch ) {
						SOFT_idx_to_euler(i,thisEulerAngles);
						thisRot = mat33<double>(thisEulerAngles[0],thisEulerAngles[1],thisEulerAngles[2]);
						
						vec3<double> tmpl_CAminus1 = (frag.getCA(1)).apply(thisRot);
						vec3<double> tmpl_CAplus1  = (frag.getCA(3)).apply(thisRot);
						
						double delCAp1=0.0, delCAm1=0.0;
						if (align_m1) delCAm1 = (pred_CAminus1 - tmpl_CAminus1).norm2();
						if (align_p1) delCAp1 = (pred_CAplus1 - tmpl_CAplus1).norm2();
		
						if (  delCAm1 <= workingMAX_CA_DEV*workingMAX_CA_DEV && 
							delCAp1 <= workingMAX_CA_DEV*workingMAX_CA_DEV    ) {
							bestAlignMatch = thisVal;
							bestAlignRot = thisRot;
							bestAlignAngles = thisEulerAngles;
						}
					}
				}
				
				if (MASK_DIST > 0) {
					// compute masked CC
					SOFT_so3_rotate( tmpRotCoefsR, tmpRotCoefsI, tmpCoefsR[frag_idx], tmpCoefsI[frag_idx],
									bestAlignAngles[2],bestAlignAngles[1],bestAlignAngles[0]);
					SOFT_so3_rotate( mskRotCoefsR, mskRotCoefsI, mskCoefsR[frag_idx], mskCoefsI[frag_idx],
									bestAlignAngles[2],bestAlignAngles[1],bestAlignAngles[0]);
					
					thisVal = SOFT_S2_correlate( sigCoefR, sigCoefI, tmpRotCoefsR, tmpRotCoefsI, mskRotCoefsR, mskRotCoefsI );
					if (thisVal > bestMatch) {
						bestMatch = thisVal;
						bestRot = bestAlignRot;
						bestAngles = bestAlignAngles;
						bestFragIdx = frag_idx;
					}
				} else {
					if (bestAlignMatch > bestMatch) {
						bestMatch = bestAlignMatch;
						bestRot = bestAlignRot;
						bestAngles = bestAlignAngles;
						bestFragIdx = frag_idx;
					}
				}				
			}
			
			
		} // end loop over fragments
		
		//
		// dump pdb
		Nmer &frag = frags[bestFragIdx];
		frag.rotate(bestRot);
		char resType, chnId;
		int  resId;
		 
		frag.getCenterRes(thisRes);
		PDBResidue res = prot.serialGet(i_res);
		prot.serialGet(i_res, resId, resType, chnId);

		cerr << "RES_" << i_res << ".cc(" << bestFragIdx << "/" << nfragments << ") , " 
		     << bestMatch << " , {" << bestAngles.to_string() << "}" << endl;

		out << setiosflags(ios_base::right|ios::fixed) << setprecision(3);
		for (int i_atm=0; i_atm<thisRes.atoms.size(); i_atm++) {
			out << "ATOM  " << setw(5) << atmCounter++ << " " << thisRes.atoms[i_atm].atmid << " "
			     << threeLetterCode(thisRes.restype)
                 << " " << chnId << setw(4) << resId+1 << "    ";
			out << setw(8) << res.cartX[0] + thisRes.atoms[i_atm].cartX[0] 
 			     << setw(8) << res.cartX[1] + thisRes.atoms[i_atm].cartX[1]
			     << setw(8) << res.cartX[2] + thisRes.atoms[i_atm].cartX[2];
			out << "  1.00 " << thisRes.b << endl;
		}
	}
	out << "TER   " << endl;
}
