// //////////////////////////////////////////////////////////////////////////
//
//     sidechain_pf.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>

extern "C" {
#include "fftw3.h"
#include "complex.h" 
#include "csecond.h"
#include "so3_correlate_fftw.h"
#include "soft_fftw_pc.h"
#include "soft_fftw.h"
#include "makeWigner.h"
#include "makeweights.h"
#include "FST_semi_memo.h"
#include "cospmls.h"
#include "primitive_FST.h"
#include "seminaive.h"
}

#include "map_mask.h"
#include "spharm.h"
#include "compact_map3d.h"

int POSITIONAL_COST[] = {1,1,9999,1,1};  // don't let middle position change
int MAX_PAMDIST = 20;

const double DIST_mu     = 3.802525 ;
const double DIST_sigma  = 0.05084 ;

double ENTROPY_CUTOFF;  //? what should this be?  tune??
double B_MAX_CUTOFF;

int FIRST_BUCKET, LAST_BUCKET;


///
/// Particle
///
class Particle {
public:
	Particle() {
		weight = 1.0;
	}
	
	Particle(int nres, Crystal &crystData) {
		trace.resize(nres);
		b_cover.resize(nres, false);
		mask.init(crystData);
		weight = 1.0;
	}

	vector< Residue > trace;
	MapMask           mask;
	vector< bool >    b_cover;
	
	double weight;
};


///
/// class Posterior
///    wrapper for posterior distribution
class Posterior {
public:
	Posterior() { }

	Posterior( string marginalFileName, 
			   string mer, 
			   Crystal& crystData, vec3<int> &dims)
	{
		map_3d<double> b_full;

		// read POSTERIOR (marginal) maps
		if (!b_full.readFromFile(marginalFileName)) {
			cerr << "WARNING: " << marginalFileName << " not found!\n";
			return;
		}
		dims = b_full.getDims();
		int dimsProduct = dims[0]*dims[1]*dims[2];
		crystData.buildMaps(dims);
		
		// compact
		M.alloc( crystData, dims );
		for (int x=0; x<dimsProduct; ++x) 
			if (crystData.getAsymmUnit().get(x) > 0) 
				M.i_set(x, (float)b_full.get(x));
		
		// normalize
		M.invlog_transform();
		M.normalize();
	}
	
	compact_map_3d<float>  M;
};


///
/// fragment_lookup
///
void fragment_lookup( string fragmentDir, 
                      string mer,
					  int Rmax, int Ralign, int NEXAMPLES, double MASK_DIST,
                      vector< map_3d<double> >  &coeffsR, 
                      vector< map_3d<double> >  &coeffsI, 
                      vector< map_3d<double> >  &maskCoeffsR, 
                      vector< map_3d<double> >  &maskCoeffsI,
                      vector< Nmer >          &fragments,
					  vector< string >  &similar_files,
					  vector< CAtrace > &similar_pdbs ) {
	coeffsR.clear(); coeffsI.clear();
	maskCoeffsR.clear(); maskCoeffsI.clear();
	fragments.clear();
	
	// build tiny unit cell for constructing maps from fragments
	//    (we don't want to build over the entire crystal)
	static double radiusSpHarm = Rmax * Rstep + 1.0;
	static double cellEdgeSpHarm = 2.0 * radiusSpHarm + 12.0;
	static double deltaX = min(1.0, reso/3.0);
	static int nRadsteps = (int)ceil( (cellEdgeSpHarm) / deltaX );
	static Crystal crystSpHarm(cellEdgeSpHarm,cellEdgeSpHarm,cellEdgeSpHarm, 90.00,90.00,90.00);
	static vec3<int> dimsSpHarm(nRadsteps, nRadsteps, nRadsteps);
	
	// cache interpolation coefficients
	// cubic spline interpolation for density
	static map_3d< cubicSplineCoeff3d > sphPtrsSpHarm;
	if (sphPtrsSpHarm.getDimsProd() == 0) {
		sphPtrsSpHarm.alloc(Rmax,2*B,2*B);
		cache_spherical_resampling_cubic_spline(sphPtrsSpHarm, dimsSpHarm, crystSpHarm);
	}
		
	// linear interpolation for mask (to keep mask non-negative)
	static map_3d< linearCoeff3d > linPtrsSpHarm;
	if (linPtrsSpHarm.getDimsProd() == 0) {
		linPtrsSpHarm.alloc(Rmax,2*B,2*B);
		cache_spherical_resampling_linear(linPtrsSpHarm, dimsSpHarm, crystSpHarm);
	}
	int nSimilar = similar_files.size();
	
	if (mer[2] == '_') return;  // gap

	string fragmentFile = fragmentDir+"fragments_"+mer+"_r"+itoc(Rmax)+"_b"+itoc(B)+"_N"+itoc(NEXAMPLES)+"_m"+ftoc(MASK_DIST)+".dat";
	if (exists(fragmentFile)) {
		ifstream inSph( fragmentFile.c_str() );
		if (!inSph.is_open())  {
			cerr << "ERROR: Unable to open fragments file " << fragmentFile << "\n";
			exit(1);
		}
			
		int nTemplates;
		inSph.read( (char*)&nTemplates , sizeof(int) );
			
		fragments.resize( nTemplates );
		coeffsR.resize( nTemplates );
		coeffsI.resize( nTemplates );
		if (MASK_DIST > 0) {
			maskCoeffsR.resize( nTemplates );
			maskCoeffsI.resize( nTemplates );
		}
			
		for (int i_frag=0; i_frag<nTemplates; ++i_frag) {
			// write info necessary to load fragment
			int filenamelen, tgtResid;
			string pdbfile;
			char tgtChain;
			char stringbuff[1024];
				
			inSph.read( (char*)&filenamelen , sizeof(int) );
			inSph.read( (char*)stringbuff , filenamelen*sizeof(char) );
			inSph.read( (char*)&tgtChain , sizeof(char) );
			inSph.read( (char*)&tgtResid , sizeof(int) );
			pdbfile = string(stringbuff);
				
			fragments[i_frag].readFile(pdbfile, tgtChain, tgtResid);
				
			coeffsR[i_frag].alloc(Rmax,B,B);
			inSph.read( (char*)&coeffsR[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			coeffsI[i_frag].alloc(Rmax,B,B);
			inSph.read( (char*)&coeffsI[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			if (MASK_DIST > 0) {
				maskCoeffsR[i_frag].alloc(Rmax,B,B);
				inSph.read( (char*)&maskCoeffsR[i_frag].get(0) , Rmax*B*B*sizeof(double) );
				maskCoeffsI[i_frag].alloc(Rmax,B,B);
				inSph.read( (char*)&maskCoeffsI[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			}
		} 
	} else {
		// fragment lookup
		if (nSimilar > 0) {
			if (MASK_DIST > 0) {
				spharm_lookupAndMaskFromFilelist(mer, similar_files, similar_pdbs, 
							radiusSpHarm, dimsSpHarm, crystSpHarm, sphPtrsSpHarm, linPtrsSpHarm,
							fragments, coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
							MASK_DIST, Ralign);
			} else {
				spharm_lookupFromFilelist(mer, similar_files, similar_pdbs, 
							radiusSpHarm, dimsSpHarm, crystSpHarm, sphPtrsSpHarm, 
							fragments, coeffsR, coeffsI);
			}
		}
		if (MASK_DIST > 0) {
			spharm_lookupAndMask(mer, NEXAMPLES, 
								radiusSpHarm, dimsSpHarm, crystSpHarm, sphPtrsSpHarm, linPtrsSpHarm,
								fragments, coeffsR, coeffsI, maskCoeffsR, maskCoeffsI,
								MASK_DIST, Ralign);
		} else {
			spharm_lookup(mer, NEXAMPLES, radiusSpHarm, dimsSpHarm, crystSpHarm, sphPtrsSpHarm, 
							fragments, coeffsR, coeffsI);
		}

		int nTemplates = fragments.size();
		if (nTemplates == 0) { cerr << "(2) FATAL ERROR! No \'" << mer << "\' clusters found.\n"; exit(1); }
		
		// write to file for fast lookup in the future
		ofstream outSph( fragmentFile.c_str() );
		outSph.write( (char*)&nTemplates , sizeof(int) );
		for (int i_frag=0; i_frag<nTemplates; ++i_frag) {
			Nmer &frag = fragments[i_frag];
			
			// write info necessary to load fragment
			int filenamelen = frag.pdbfile.size()+1;
			outSph.write( (char*)&filenamelen , sizeof(int) );
			outSph.write( (char*)frag.pdbfile.c_str() , filenamelen*sizeof(char) );
			outSph.write( (char*)&frag.tgtChain , sizeof(char) );
			outSph.write( (char*)&frag.tgtResid , sizeof(int) );
			
			
			outSph.write( (char*)&coeffsR[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			outSph.write( (char*)&coeffsI[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			if (MASK_DIST > 0) {
				outSph.write( (char*)&maskCoeffsR[i_frag].get(0) , Rmax*B*B*sizeof(double) );
				outSph.write( (char*)&maskCoeffsR[i_frag].get(0) , Rmax*B*B*sizeof(double) );
			}
		}
	}
}


///
/// forward_sampling
///
int forward_sample_CA(
              int                               dir,             // +1:forward / -1:backward
			  vector< compact_map_3d<float>* >  bs,            // beliefs to consider
              Crystal                          &crystData, 
              Residue                          &prevRes,         // N-CA-C (backwards) of CURRENT res
              Residue                          &thisRes,         // N-CA-C of forward/backward selected res
			  double                           &weightSum,       // output: chosen b_i sum
			  double                           &weight           // output: chosen b_i
        ) 
{
	static int NUM_SAMPLES = 10000;

	vec3<int> dims = bs[0]->getDims();
	vector< vec3< double > > cartCAs(NUM_SAMPLES);
	vector< double > thisResProb(NUM_SAMPLES);
	vector< short >  copyTaken(NUM_SAMPLES);
	
	double probSum = 0.0, psiDevSum = 0.0;
	double omega, phi, psi, temp1, temp2;
	
	vec3<double> fracCA, prevprevCA;
	vec3<int> idxCA;
	
	// copy over old atoms
	if (dir < 0) {
		thisRes.CAp1 = prevRes.CA;
		prevprevCA = prevRes.CAp1;
	} else {
		thisRes.CAm1 = prevRes.CA;
		prevprevCA = prevRes.CAm1;
	}
	// perp to prevprev
	vec3<double> perpPPCA = prevprevCA - prevRes.CA;
	if (perpPPCA[0] == 0.0 && perpPPCA[1] == 0.0)
		perpPPCA = prevprevCA + vec3<double>(0.0, perpPPCA[2], -perpPPCA[1]);
	else
		perpPPCA = prevprevCA + vec3<double>(perpPPCA[1], -perpPPCA[0], 0.0);
	
	/////////////////////////////	
	/// do NUM_SAMPLES iterations of forward(backward) kinematics 
	///    to get an approximate forward(backward) distribution 
	/////////////////////////////
	int nBs = bs.size();
	for (int i=0; i<NUM_SAMPLES; ++i) {
		double dist = DIST_mu + randnorm() * DIST_sigma;
		double theta = sample_CACACA_pseudoangle();
		double torsion = 360.0 * ((double)rand())/RAND_MAX;
		get_coord_from_angle(  perpPPCA , prevprevCA,  prevRes.CA,  cartCAs[i], torsion, theta, dist);
		
		// find the corresponding grid point
		fracCA = cartCAs[i].apply( crystData.c2f );
		idxCA[0] = pos_mod( (int)floor(fracCA[0]*dims[0]+0.5) , dims[0] );
		idxCA[1] = pos_mod( (int)floor(fracCA[1]*dims[1]+0.5) , dims[1] );
		idxCA[2] = pos_mod( (int)floor(fracCA[2]*dims[2]+0.5) , dims[2] );
		
		double thisB = 0.0;
			
		// check all posteriors
		for (int i_cpy=0; i_cpy<nBs; ++i_cpy) {
			if (thisB < bs[i_cpy]->i_get( idxCA[0] , idxCA[1] , idxCA[2])) {
				thisB = bs[i_cpy]->i_get( idxCA[0] , idxCA[1] , idxCA[2] ) ;
				copyTaken[i] = i_cpy;
			}
		}
		
		thisResProb[i] = thisB;  // sampling wrt psi_fk
		probSum   += thisResProb[i];
	}
	
	/////////////////////////////
	/// pick a sample at random, with prob(x) ~ thisResProb
	/// TODO: chunk thisResProb so random selection is faster (?)
	/////////////////////////////
	double samplePicker = probSum*((double)rand())/RAND_MAX;
	for (int i=0; i<NUM_SAMPLES; ++i) {
		samplePicker -= thisResProb[i];
		if (samplePicker < 0.0) {
			// recall stored bb-config
			thisRes.CA = cartCAs[i]; 
			
			// weight is sum of b_i's considered
			weightSum = probSum / NUM_SAMPLES;
			weight = thisResProb[i];
			
			// our work here is done
			return copyTaken[i];
		}
	}
}
	

///
/// start a trace
///
void startTrace(     Residue                  &thisRes,
                     compact_map_3d<float>*   b,   // beliefs of any identical/near-identical chains
                     map_3d<double>           &map , 
				     vector< Nmer >           &fragments, 
					 vector< map_3d<double> > &tmplCoefsR, vector< map_3d<double> > &tmplCoefsI,
					 int                       Ralign,
				     Crystal                  &crystData,
				     double                   &weight,
					 MapMask                  &mask,
					 int                       resid
               ) {
	vec3<double> fracX_i, cartX_i;
	vec3<int> idxX_i; 
	int NFRAGS = fragments.size();
	int Rmax = tmplCoefsR[0].getDims(0);
	vec3<int> dims = map.getDims();
				
	double b_taken;
	map_3d<double> sigR(Rmax,2*B,2*B);
	map_3d<double> sigCoefR(Rmax,B,B);
	map_3d<double> sigCoefI(Rmax,B,B); 
	
	do {
		// pick a location with probability given by b_i
		b_taken = b->randomLocation( idxX_i );
		
		//  fracX_i uniformly distributed inside voxel idxX_i
		fracX_i[0] = ((double)idxX_i[0] + (((double)rand())/RAND_MAX - 0.5)) / dims[0];
		fracX_i[1] = ((double)idxX_i[1] + (((double)rand())/RAND_MAX - 0.5)) / dims[1];
		fracX_i[2] = ((double)idxX_i[2] + (((double)rand())/RAND_MAX - 0.5)) / dims[2];
		
		cartX_i = fracX_i.apply(crystData.f2c);
	} while ( mask.checkMask( cartX_i , crystData) );
		
	spherical_resample_cubic_spline( sigR, map, fracX_i, crystData);
	sph_standardize(sigR);
	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);
	
	///////////////////
	/// consider all fragments at all rotations for seed
	///////////////////
	vector< vector<double> > allCCs(NFRAGS, vector<double>(8*Bout*Bout*Bout, 0.0) );
	vector< vector<double> > allProbs(NFRAGS, vector<double>(8*Bout*Bout*Bout, 0.0) );
	double probSum=0.0, thisCC, pNull;
	double CCsum=0.0, CCsum2=0.0;
	int nCCs = 0;
	
	for (int i_frag = 0; i_frag < NFRAGS; ++i_frag) {
		Nmer &tmplFrag = fragments[ i_frag ];
		
		SOFT_so3_correlate(sigCoefR, sigCoefI, tmplCoefsR[i_frag], tmplCoefsI[i_frag], Ralign);
		
		for (int k=0; k<8*Bout*Bout*Bout; ++k) {
			thisCC = so3SigSum.get(k).real() / ( M_PI*sqrt(2.0) );  
			     // denominator handles const. scaling between so3_correlate and S2_correlate
			
			
			if (!isnan(thisCC)) {
				allCCs[i_frag][k] = thisCC;
				CCsum  += thisCC;
				CCsum2 += thisCC*thisCC;
				nCCs++;
			} else {
				cerr << "WARNING! in starttrace(...)" << endl;
				cerr << "     allCCs[" << i_frag << "][" << k << "] = NaN\n";
			}
		}
	}
	CCsum /= nCCs;
	CCsum2 = sqrt(CCsum2/nCCs - CCsum*CCsum);
	
	if (CCsum2 == 0) CCsum2 = 1.0;
	
	// convert CCs to Z-scores then log-probs
	vector<double> templSums(NFRAGS,0.0);
	probSum=0.0;
	for (int i_frag = 0; i_frag < NFRAGS; ++i_frag) {
		for (int k=0; k<8*Bout*Bout*Bout; ++k) {
			pNull = normCDF( (CCsum - allCCs[i_frag][k] ) / CCsum2 );
			
			allProbs[i_frag][k] = (1.0 - pNull) / pNull;
	
			templSums[i_frag] += allProbs[i_frag][k];
		}
		probSum += templSums[i_frag];
	}

	///////////////
	/// pick a fragment+orientation at random
	///////////////
	double fragPicker = probSum*((double)rand())/RAND_MAX;
	
	mat33<double> rot_taken;
	double prob_taken, cc_taken;
	int fragidx_taken = -1;
	
	for (int i_frag = 0; i_frag < NFRAGS; ++i_frag) {
		if (fragPicker <= templSums[i_frag]) {
			fragidx_taken = i_frag;
			
			for (int k=0; k<8*Bout*Bout*Bout; ++k) {
				fragPicker -= allProbs[i_frag][k];
				if (fragPicker <= 0) {
					// choose this orientation
					SOFT_idx_to_rot(k,rot_taken);
					prob_taken = allProbs[i_frag][k]; // /probSum;
					cc_taken = allCCs[i_frag][k];  
					goto chosen;  // break both loops
				}
			}
		}
		
		fragPicker -= templSums[i_frag];
	}
	
	///////////////////
	/// update weight
	///    prob_taken is the probability of this fragment+orientation _given_ location+image
	///    prob_match ESTIMATES the prob of fragmemt+orientation+location _given_ image
	///////////////////
chosen:
	// rotate, save fragment
	Nmer frag_taken = fragments[fragidx_taken];
	frag_taken.rotate(rot_taken);
	frag_taken.translate(cartX_i);
	frag_taken.getCenterRes(thisRes);
	
	// initial BB "seed"
	thisRes.CAm1 = frag_taken.getAtom(1," CA ");
	thisRes.N    = frag_taken.getAtom(2," N  "); 
	thisRes.CA   = frag_taken.getAtom(2," CA "); 
	thisRes.C    = frag_taken.getAtom(2," C  "); 
	thisRes.O    = frag_taken.getAtom(2," O  "); 
	thisRes.CAp1 = frag_taken.getAtom(3," CA "); 
	
	weight *= 1.0; 
	
	// check ALL atoms for collisions
	if ( mask.checkMask( resid, thisRes.atoms , crystData , true ) ) {
		weight = 0.0;
		thisRes.atoms.clear();
		thisRes.N = thisRes.CA = thisRes.C = thisRes.O = vec3<double>(0.0,0.0,0.0);
	} else {
		// update mask
		mask.updateMask( resid , thisRes.atoms , crystData);
	}
}


///
/// extend a trace
///
void  matchSidechains( vec3<double>              CAm1, vec3<double> CA, vec3<double> CAp1,
                       map_3d<double>           &map , 
			           vector< Nmer >           &fragments, 
					   vector< map_3d<double> > &tmplCoefsR, vector< map_3d<double> > &tmplCoefsI,
					   vector< map_3d<double> > &maskCoefsR, vector< map_3d<double> > &maskCoefsI,
                       Crystal                  &crystData,
				       vector<double>           &CCs,
					   vector< mat33<double> >  &rots
                     ) {
	vec3<int> dims = map.getDims();
	int Rmax = tmplCoefsR[0].getDims(0);
	int NFRAGS = fragments.size();
	
	map_3d<double> sigR(Rmax,2*B,2*B), sigCoefR(Rmax,B,B), sigCoefI(Rmax,B,B); 
	map_3d<double> tmplRotCoefsR(Rmax,B,B), tmplRotCoefsI(Rmax,B,B);   // rotated spharms
	map_3d<double> maskRotCoefsR(Rmax,B,B), maskRotCoefsI(Rmax,B,B);   // rotated spharms
	
	////////////////
	///
	/// extract spherical harmonics around 'CA'
	///
	vec3<double> bwCA=CAm1-CA, fwCA=CAp1-CA, fracX = CA.apply( crystData.c2f );
	spherical_resample_cubic_spline( sigR, map, fracX, crystData );
	sph_standardize(sigR);
	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);

	double rmsd, U[3][3];
	double xcoords[2][3], ycoords[2][3];
	
	for (int i_frag=0; i_frag<NFRAGS; ++i_frag) {
		Nmer &frag = fragments[i_frag];

		// align frag to thisRes.(C/CA/N)
		// TODO: closed form version of two-vector rotational alignment 
		//       (i.e. don't call calculate_rotation_rmsd_notranslate)
		vec3<double> fragCAm1=frag.getAtom(1," CA "), fragCAp1=frag.getAtom(3," CA ");
		for (int x=0;x<3;++x) {
			xcoords[0][x]=bwCA[x];
			xcoords[1][x]=fwCA[x];
			ycoords[0][x]=fragCAm1[x];
			ycoords[1][x]=fragCAp1[x];
		}
		calculate_rotation_rmsd_notranslate(xcoords, ycoords, 2, U, &rmsd);
		
		// rotate spharm
		rots[i_frag].set(U);
		vec3<double> thisEulerAngles = rot2euler( rots[i_frag] );
		
		SOFT_so3_rotate( tmplRotCoefsR, tmplRotCoefsI, tmplCoefsR[i_frag], tmplCoefsI[i_frag],
						 thisEulerAngles[2], thisEulerAngles[1], thisEulerAngles[0]);
		
		if (maskCoefsR.size() > 0) {  // is a mask defined?
			SOFT_so3_rotate( maskRotCoefsR, maskRotCoefsI, maskCoefsR[i_frag], maskCoefsI[i_frag],
							 thisEulerAngles[2], thisEulerAngles[1], thisEulerAngles[0]);
					
			CCs[i_frag] = SOFT_S2_correlate( sigCoefR, sigCoefI, tmplRotCoefsR, tmplRotCoefsI, maskRotCoefsR, maskRotCoefsI );
		} else {
			CCs[i_frag] = SOFT_S2_correlate( sigCoefR, sigCoefI, tmplRotCoefsR, tmplRotCoefsI ); // compute correlation
		}
	}
}


////
//// match chain terminus
////
void matchTerminus( int                       dir,
                    Residue                  &thisRes, vec3<double> CA1, vec3<double> CA2,
                    map_3d<double>           &map , 
			        vector< Nmer >           &fragments, 
					vector< map_3d<double> > &tmplCoefsR, vector< map_3d<double> > &tmplCoefsI,
					vector< map_3d<double> > &maskCoefsR, vector< map_3d<double> > &maskCoefsI,
                    Crystal                  &crystData,
				     double                   &weight,
					 MapMask                  &mask,
					 int                       resid
                  ) {
	vec3<int> dims = map.getDims();
	int Rmax = tmplCoefsR[0].getDims(0);
	int NFRAGS = fragments.size();
	
	map_3d<double> sigR(Rmax,2*B,2*B), sigCoefR(Rmax,B,B), sigCoefI(Rmax,B,B); 
	map_3d<double> tmplRotCoefsR(Rmax,B,B), tmplRotCoefsI(Rmax,B,B);   // rotated spharms
	map_3d<double> maskRotCoefsR(Rmax,B,B), maskRotCoefsI(Rmax,B,B);   // rotated spharms
	
	
	///
	/// extract spherical harmonics around 'CA'
	///
	vec3<double> CAd1=CA1-thisRes.CA, 
	             CAd2=CA2-thisRes.CA, 
				 fracX = thisRes.CA.apply( crystData.c2f );
	spherical_resample_cubic_spline( sigR, map, fracX, crystData );
	sph_standardize(sigR);
	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);

	double rmsd, U[3][3];
	double xcoords[2][3], ycoords[2][3];
	double thisCC, bestCC=-INFINITY;
	mat33<double> thisRot, bestRot;
	int bestFragIdx;
	
	for (int i_frag=0; i_frag<NFRAGS; ++i_frag) {
		Nmer &frag = fragments[i_frag];

		// align frag to thisRes.(C/CA/N)
		// TODO: closed form version of two-vector rotational alignment 
		//       (i.e. don't call calculate_rotation_rmsd_notranslate)
		vec3<double> fragCA1, fragCA2;
		
		if (dir>0) {
			fragCA1=frag.getAtom(1," CA ");
			fragCA2=frag.getAtom(0," CA ");
		} else {
			fragCA1=frag.getAtom(3," CA ");
			fragCA2=frag.getAtom(4," CA ");
		}			
		
		for (int x=0;x<3;++x) {
			xcoords[0][x]=CAd1[x];
			xcoords[1][x]=CAd2[x];
			ycoords[0][x]=fragCA1[x];
			ycoords[1][x]=fragCA2[x];
		}
		calculate_rotation_rmsd_notranslate(xcoords, ycoords, 2, U, &rmsd);
		
		// rotate spharm
		thisRot.set(U);
		vec3<double> thisEulerAngles = rot2euler( thisRot );
		
		SOFT_so3_rotate( tmplRotCoefsR, tmplRotCoefsI, tmplCoefsR[i_frag], tmplCoefsI[i_frag],
						 thisEulerAngles[2], thisEulerAngles[1], thisEulerAngles[0]);
		
		if (maskCoefsR.size() > 0) {  // is a mask defined?
			SOFT_so3_rotate( maskRotCoefsR, maskRotCoefsI, maskCoefsR[i_frag], maskCoefsI[i_frag],
							 thisEulerAngles[2], thisEulerAngles[1], thisEulerAngles[0]);
					
			thisCC = SOFT_S2_correlate( sigCoefR, sigCoefI, tmplRotCoefsR, tmplRotCoefsI, maskRotCoefsR, maskRotCoefsI );
		} else {
			thisCC = SOFT_S2_correlate( sigCoefR, sigCoefI, tmplRotCoefsR, tmplRotCoefsI ); // compute correlation
		}
		
		if (thisCC > bestCC) {
			bestCC = thisCC;
			bestRot = thisRot;
			bestFragIdx = i_frag;
		}
	}
	
	Nmer frag_taken = fragments[bestFragIdx];
	frag_taken.rotate(bestRot);
	frag_taken.translate(thisRes.CA);
	frag_taken.getCenterRes(thisRes);

	thisRes.CAm1 = frag_taken.getAtom(1," CA ");
	thisRes.N    = frag_taken.getAtom(2," N  "); 
	thisRes.C    = frag_taken.getAtom(2," C  "); 
	thisRes.O    = frag_taken.getAtom(2," O  "); 
	thisRes.CAp1 = frag_taken.getAtom(3," CA "); 
	
	// check  for collisions
	if (mask.checkMask( resid, thisRes.atoms , crystData)) {
		weight = 0;
		thisRes.atoms.clear();
		thisRes.N = thisRes.CA = thisRes.C = thisRes.O = vec3<double>(0.0,0.0,0.0);		
	} else {
		// update mask
		mask.updateMask( resid , thisRes.atoms , crystData);
	}		
}

///
/// resample traces
///
void resample_traces( vector< Particle > &parts , vector< int > &bCover, bool force = false ) {
	int NPARTICLES = parts.size();  
					   
	vector< Particle > partsCopy=parts;
	
	// sum weights, find best trace
	double weightSum2=0.0, weightSum=0.0, thisWeight;
	for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
		thisWeight  = parts[i_prt].weight;
		weightSum  += thisWeight;
		weightSum2 += thisWeight*thisWeight;
	}
	
	// if (wtSum2 < cutoff??) return
	double stdevWt = sqrt( weightSum2/NPARTICLES - weightSum*weightSum/(NPARTICLES*NPARTICLES) );
	double degen = (weightSum/NPARTICLES) / stdevWt;
 	
	if (!force && degen > 0.7071) // sqrt(2)/2
		return;
		
	// full resample
	for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
		double selectPrt = weightSum * ((double)rand()) / RAND_MAX;
			
		for (int j_prt = 0; j_prt < NPARTICLES; ++j_prt) {
			selectPrt -= partsCopy[j_prt].weight;
			if (selectPrt <= 0) {
				parts[i_prt] = partsCopy[j_prt];
				break;
			}
		}
	}
	
	int nres = bCover.size();
	for (int i_res = 0; i_res < nres; ++i_res)
		bCover[ i_res ] = 0;	

	for (int i_prt = 0; i_prt < NPARTICLES; ++i_prt) {
		// reset weights (since we resampled)
		parts[i_prt].weight = 1.0;
		
		// recompute bCover
		for (int i_res = 0; i_res < nres; ++i_res) {
			if ( parts[i_prt].b_cover[ i_res ] == true )
				bCover[ i_res ]++;	
		}
	}
		
	return;
}


/////////////////////////
///
/// similar to BP.h's doTrace, try to get the particle into a single asymm unit
///
class traceNode {
public:
	vec3<double> fracX, cartX;
	vec3<int>    idxX;
	float        prob;
	int          resPtr;
};

class trace {
public:
	vec3<double>      cCoM, fCoM;
	vector<traceNode> res;
	int          symCpy;
};

void findSingleAsymmUnit(Particle &p, int nRes, Crystal &crystData) {
	// traces
	traceNode       thisRes, prevRes;
	trace           thisChain;	
	vector< trace > protModel;

	///////////////////
	///
	/// A) extract CA's for each res, determine connectivity
	vector< vec3<double> > fsymm;
	for (int i=0; i < nRes; i++) {
		if (  p.trace[i].atoms.size() != 0  )  {
			thisRes.fracX = p.trace[i].CA.apply( crystData.c2f );
			thisRes.resPtr = i;

			crystData.getSymmetric(thisRes.fracX, fsymm);
			
			if ( thisChain.res.size() != 0 ) {
				bool connectToPrev = false;
				
				//for (int j=0; j<fsymm.size(); ++j) {
					int j=0;
					vec3<double> delfXY = fsymm[j] - thisChain.res.back().fracX;
					delfXY.minmod1();
					
					// check if connected to previous
					if ( delfXY.apply( crystData.f2c ).norm2() < 4.3*4.3 ) {
						// CONNECTED!
						connectToPrev = true;
						thisRes.fracX = thisChain.res.back().fracX + delfXY;
						thisRes.cartX = thisChain.res.back().cartX + delfXY.apply( crystData.f2c );
						//cerr << "Connect " << thisChain.res.back().cartX.to_string() << " to " << thisRes.cartX.to_string() << endl;
				//		break;
						
					}
				//}
				
				if ( !connectToPrev ) {
					// NOT CONNECTED !
					thisChain.cCoM = thisChain.cCoM / thisChain.res.size();   // compute CoM
					thisChain.fCoM = thisChain.cCoM.apply( crystData.c2f ); //thisChain.fCoM.mod(1);
					
					if (thisChain.res.size() >= 3) {
						protModel.push_back(thisChain);
					} else {
						// delete this 'chain'
						for (int j=0; j < thisChain.res.size(); ++j) {
							int iPtr = thisChain.res[j].resPtr;
							p.trace[iPtr].atoms.clear();
							p.trace[iPtr].N = p.trace[iPtr].CA = p.trace[iPtr].C = p.trace[iPtr].O = vec3<double>(0.0,0.0,0.0);		
						}
					}
					
					thisChain.res.clear();
					thisChain.cCoM = vec3<double>(0,0,0);
					
					thisRes.cartX = thisRes.fracX.apply( crystData.f2c );
				}
			} else {
				thisChain.cCoM = vec3<double>(0,0,0);
				thisRes.cartX = thisRes.fracX.apply( crystData.f2c );
			}
			
			thisChain.res.push_back( thisRes );
			thisChain.cCoM = thisChain.cCoM + thisRes.cartX;
		}	
	}
	
	// add last chain
	thisChain.cCoM = thisChain.cCoM / thisChain.res.size();   // compute CoM
	thisChain.fCoM = thisChain.cCoM.apply( crystData.c2f ); //thisChain.fCoM.mod(1);
	protModel.push_back(thisChain);

	///////////////////
	///
	/// B) move all to correct asymm unit
	int nchains = protModel.size();
	int movedCount=1;
	vector<bool> moved(nchains, false);
	int maxChainLen=protModel[0].res.size(), idxMaxChain=0;
	vec3<double> combinedCoM, sumCoM;
	
	// first fix longest chain
	for (int i=1; i<nchains; ++i) {
		if ( protModel[i].res.size() > maxChainLen ) {
			maxChainLen = protModel[i].res.size();
			idxMaxChain = i;
		}
	}
	moved[idxMaxChain] = true;
	combinedCoM = sumCoM = protModel[idxMaxChain].fCoM;
	protModel[idxMaxChain].symCpy = 0;
	
	while (movedCount < nchains) {
		// for each chain, find distance closest to combinedCoM
		int idxMinDist, copyMinDist=0, idxResMinDist;
		double thisDist, minDist=INFINITY;
		vec3<double> minCoM;
		
		for (int i=0; i<nchains; ++i) {
			if (moved[i]) continue; 
			
			crystData.getSymmetric(protModel[i].fCoM, fsymm);
			for (int j=0; j<fsymm.size(); ++j) {
				vec3<double> delfXY = fsymm[j] - combinedCoM;
				delfXY.minmod1();
				
				thisDist = delfXY.apply(crystData.f2c).norm2();
				
				if (thisDist < minDist) {
					minDist = thisDist;
					idxMinDist = i;
					copyMinDist = j;
					minCoM = combinedCoM + delfXY;
				}
			}
		}
		
		// update CoM
		moved[idxMinDist] = true;
		movedCount++;
		sumCoM = sumCoM + minCoM;
		combinedCoM = sumCoM / movedCount;
 		protModel[idxMinDist].fCoM = minCoM;
		protModel[idxMinDist].symCpy = copyMinDist;
		
		// move chain 'idxMinDist'
		int nres = protModel[idxMinDist].res.size();
		
		// find+set res closest to com
		minDist = INFINITY;
		for (int i=0; i<nres; ++i) {
			crystData.getSymmetric( protModel[idxMinDist].res[i].fracX, 
			                        protModel[idxMinDist].res[i].fracX , 
									copyMinDist);
			
			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - minCoM;
			delfXY.minmod1();
			
			protModel[idxMinDist].res[i].fracX = minCoM + delfXY;
			
			thisDist = delfXY.apply(crystData.f2c).norm2();
			
			if (thisDist < minDist) {
				minDist = thisDist;
				idxResMinDist = i;
			}
		}
		
		protModel[idxMinDist].res[idxResMinDist].cartX = 
		                  protModel[idxMinDist].res[idxResMinDist].fracX.apply( crystData.f2c );
		
		// move BW ... 
		for (int i=idxResMinDist-1; i>=0; --i) {
			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - protModel[idxMinDist].res[i+1].fracX;
			delfXY.minmod1();
			
			//cerr << "interresdist  = " << sqrt( delfXY.apply( crystData.f2c ).norm2() ) << endl;

			protModel[idxMinDist].res[i].fracX = protModel[idxMinDist].res[i+1].fracX + delfXY;
			protModel[idxMinDist].res[i].cartX = protModel[idxMinDist].res[i].fracX.apply( crystData.f2c );
		}
		
		// and FW ...
		for (int i=idxResMinDist+1; i<nres; ++i) {
			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - protModel[idxMinDist].res[i-1].fracX;
			delfXY.minmod1();
			
			//cerr << "interresdist  = " << sqrt( delfXY.apply( crystData.f2c ).norm2() ) << endl;

			protModel[idxMinDist].res[i].fracX = protModel[idxMinDist].res[i-1].fracX + delfXY;
			protModel[idxMinDist].res[i].cartX = protModel[idxMinDist].res[i].fracX.apply( crystData.f2c );
		}
	}
	
	///////////////////
	///
	/// C) now move ALL atoms
	for (int i=0; i<protModel.size(); ++i) {
		int copyMinDist = protModel[i].symCpy;
		
		for (int j=0; j<protModel[i].res.size(); ++j) {
			int iPtr = protModel[i].res[j].resPtr;
		
			// fix CA
			p.trace[iPtr].CA = protModel[i].res[j].cartX;
			
			// fix other sidechain atoms
			vec3<double> delXY, fracCA=protModel[i].res[j].fracX, symmFracX;
			
			for (int k=0; k<p.trace[iPtr].atoms.size(); ++k) {
				crystData.getSymmetric(  p.trace[iPtr].atoms[k].cartX.apply( crystData.c2f ) ,
				                         symmFracX , copyMinDist );
				delXY = symmFracX - fracCA;	delXY.minmod1();
				p.trace[iPtr].atoms[k].cartX = p.trace[iPtr].CA + delXY.apply( crystData.f2c );
			}
				
			// and bb atoms (if they EXIST!)
			if (p.trace[iPtr].C.norm2() > 1e-6) {
				crystData.getSymmetric(  p.trace[iPtr].C.apply( crystData.c2f ) ,
										symmFracX , copyMinDist );
				delXY = symmFracX - fracCA;	delXY.minmod1();
				p.trace[iPtr].C = p.trace[iPtr].CA + delXY.apply( crystData.f2c );
			}
			
			if (p.trace[iPtr].N.norm2() > 1e-6) {
				crystData.getSymmetric(  p.trace[iPtr].N.apply( crystData.c2f ) ,
				                         symmFracX , copyMinDist );
				delXY = symmFracX - fracCA;	delXY.minmod1();
				p.trace[iPtr].N = p.trace[iPtr].CA + delXY.apply( crystData.f2c );
			}
			
			if (p.trace[iPtr].O.norm2() > 1e-6) {
				crystData.getSymmetric(  p.trace[iPtr].O.apply( crystData.c2f ) ,
				                         symmFracX , copyMinDist );
				delXY = symmFracX - fracCA;	delXY.minmod1();
				p.trace[iPtr].O = p.trace[iPtr].CA + delXY.apply( crystData.f2c );
			}
		}
	}
}


void findSingleAsymmUnit(Particle &p, Sequence &seq, Crystal &crystData) {
	findSingleAsymmUnit(p, seq.nres(), crystData);
}


///
/// write a particle to PDB file
///
void writeParticleToFile(   string    filename, 
                            Particle &part, 
                            Sequence &seqData, 
                            Crystal  &crystData) {
	// open file 
	ofstream out( filename.c_str() );
	int atmCounter = 1;
	int nResSum = seqData.nres();
	
	out << "REMARK  Generated automatically by sidechain_pf.cpp\n";
	out << crystData.crystline << endl;;
	out << setiosflags(ios_base::right|ios::fixed) << setprecision(3);
		
	for (int i_res = 0; i_res < nResSum; ++i_res) {
		string mer = seqData.serialGet5mer(i_res);
		if (mer[2] == '_') continue;  // gap
		
		Residue &thisRes = part.trace[i_res];
		if (thisRes.atoms.size()  == 0 ) continue;  // skipped
		
		string resCode = threeLetterCode(mer[2]); //threeLetterCode(thisRes.restype);
		
		// backbone
		if (thisRes.N.norm2() > 1e-6) {
			out << "ATOM  " << setw(5) << atmCounter++ << " " << " N  " << " " << resCode << " " << (char)('A') << setw(4) << i_res+1 << "    ";
			out << setw(8) << thisRes.N[0] << setw(8) << thisRes.N[1] << setw(8) << thisRes.N[2] << "  1.00 " << thisRes.b << endl;
		}
		out << "ATOM  " << setw(5) << atmCounter++ << " " << " CA " << " " << resCode << " " << (char)('A') << setw(4) << i_res+1 << "    ";
		out << setw(8) << thisRes.CA[0] << setw(8) << thisRes.CA[1] << setw(8) << thisRes.CA[2] << "  1.00 " << thisRes.b << endl;
		if (thisRes.C.norm2() > 1e-6) {
			out << "ATOM  " << setw(5) << atmCounter++ << " " << " C  " << " " << resCode << " " << (char)('A') << setw(4) << i_res+1 << "    ";
			out << setw(8) << thisRes.C[0] << setw(8) << thisRes.C[1] << setw(8) << thisRes.C[2] << "  1.00 " << thisRes.b << endl;
		}
		if (thisRes.O.norm2() > 1e-6) {
			out << "ATOM  " << setw(5) << atmCounter++ << " " << " O  " << " " << resCode << " " << (char)('A') << setw(4) << i_res+1 << "    ";
			out << setw(8) << thisRes.O[0] << setw(8) << thisRes.O[1] << setw(8) << thisRes.O[2] << "  1.00 " << thisRes.b << endl;
		}
		
		// sidechain
		for (int i_atm=0; i_atm<thisRes.atoms.size(); i_atm++) {
			if (  thisRes.atoms[i_atm].atmid == " N  " || 
				thisRes.atoms[i_atm].atmid == " CA " || 
				thisRes.atoms[i_atm].atmid == " C  " || 
				thisRes.atoms[i_atm].atmid == " O  ") continue;
				
			// mse fix
			if (thisRes.atoms[i_atm].atmid == "SE  " || thisRes.atoms[i_atm].atmid == "Se  ") 
				thisRes.atoms[i_atm].atmid = " SD ";
				
			out << "ATOM  " << setw(5) << atmCounter++ << " " << thisRes.atoms[i_atm].atmid << " "
				<< resCode << " " << (char)('A') << setw(4) << i_res+1 << "    ";
			out << setw(8) << thisRes.atoms[i_atm].cartX[0] 
				<< setw(8) << thisRes.atoms[i_atm].cartX[1]
				<< setw(8) << thisRes.atoms[i_atm].cartX[2];
			out << "  1.00 " << thisRes.b << endl;
		}
	}
	out << "TER   \n";	
}
