// //////////////////////////////////////////////////////////////////////////
//
//     get_RMSd.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>
extern "C" {
#include "rmsd.h"
}
#include "crystal.h"
#include "chem.h"
#include "map3d.h"
#include "getopts.h"   // parse command line args

#define CA_CA_DIST 3.802525

using namespace std;

// equivalence classes of sidechain atoms
// assume they are correctly assigned in postprocessing
//     NH1/NH2 in ARG
//     OD1/OD2 in ASP
//     OD1/ND2 in ASN
//     OE1/NE1 in GLN
//     OE1/OE2 in GLU
//     CD1/CD2 and CE1/CE2 in TYR/PHE
bool equivAtom(string resId, string atm1Id, string &atm2Id) {
	static map<string, string> equiv;
	if (equiv.empty()) {
		equiv["ARG. NH1"] = " NH2"; equiv["ARG. NH2"] = " NH1";
		equiv["ASP. OD1"] = " OD2"; equiv["ASP. OD2"] = " OD1";
		equiv["ASN. OD1"] = " ND2"; equiv["ASN. ND2"] = " OD1";
		equiv["GLN. OE1"] = " NE2"; equiv["GLN. NE2"] = " OE1";
		equiv["GLU. OE1"] = " OE2"; equiv["GLU. OE2"] = " OE1";
		equiv["TYR. CD1"] = " CD2"; equiv["TYR. CD2"] = " CD1";
		equiv["TYR. CE1"] = " CE2"; equiv["TYR. CE2"] = " CE1";
		equiv["PHE. CD1"] = " CD2"; equiv["PHE. CD2"] = " CD1";
		equiv["PHE. CE1"] = " CE2"; equiv["PHE. CE2"] = " CE1";
	}

	map<string,string>::iterator p = equiv.find(resId+"."+atm1Id);
	if (p != equiv.end()) {
		atm2Id = p->second;
		return true;
	}
	atm2Id = "";
	return false;
}


static map_3d< double >* getAdjPotential(vec3<int> dims , Crystal &crystData) {
	static map_3d< double > map;
	srand(1);
	
	if (map.getDimsProd() == 0) {
		vec3<double> fX, fXovr;  // frac coords
		double pR;        // polar 'R'
	
		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(crystData.f2c).norm2() )
		);
	
		// use random sampling
		int NUM_SAMPLES = 100000;
		
		map.alloc(dims);
		map.setAll(0);
		
		for (int i=0; i<NUM_SAMPLES; ++i) {
			vec3<double> fracX(
						((double)rand()/RAND_MAX-0.5)/dims[0],
						((double)rand()/RAND_MAX-0.5)/dims[1],
						((double)rand()/RAND_MAX-0.5)/dims[2]); // start
			
			double distXY = 3.802525 + 0.05084*randnorm();   // dist
			
			// direction
			vec3<double> dirXY;
			double del;
			do {
				dirXY[0] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[1] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[2] = 2.0*(double)rand()/RAND_MAX-1.0;
				del = dirXY.norm2();
			} while (del >= 1.0 || del == 0);
			
			del = distXY / sqrt(del);
			dirXY = dirXY * del;
			
			vec3<double> fracY = fracX + dirXY.apply(crystData.c2f);
			vec3<int> idxY( pos_mod( (int)floor(fracY[0]*dims[0]+0.5) , dims[0]),
							pos_mod( (int)floor(fracY[1]*dims[1]+0.5) , dims[1]),
							pos_mod( (int)floor(fracY[2]*dims[2]+0.5) , dims[2]) );
							
			map.inc(idxY[0],idxY[1],idxY[2], 1.0);	
		}
		
		// ensure symmetry
		for (int i=0; i<dims[0]; i++)
		for (int j=0; j<dims[1]; j++)
		for (int k=0; k<dims[2]; k++) {
			fX[0] = ((double)i)/dims[0];  if (fX[0]>.5) fX[0] -=1;
			fX[1] = ((double)j)/dims[1];  if (fX[1]>.5) fX[1] -=1;
			fX[2] = ((double)k)/dims[2];  if (fX[2]>.5) fX[2] -=1;
	
			vec3<double> sX;
			vec3<int> sIdx;
			
			double map_ijk = map.get(i,j,k);
			if (map_ijk == 0) continue;   // ???
			
			for (int s=1; s<crystData.nSymm(); ++s) {
				sX = fX.apply( crystData.getSymmRot(s) );
				sIdx[0] = pos_mod( (int)floor(sX[0]*dims[0]+0.5) , dims[0]);
				sIdx[1] = pos_mod( (int)floor(sX[1]*dims[1]+0.5) , dims[1]);
				sIdx[2] = pos_mod( (int)floor(sX[2]*dims[2]+0.5) , dims[2]);
				
				map_ijk = max(map_ijk, map.get(sIdx[0],sIdx[1],sIdx[2]));
			}
	
			map.set(i,j,k, map_ijk);
		}
		
		map.normalize();  // sum to 1
		
		//writeMapToDisk(map, "phiAdj");
	}
	
	return &map;
}



//
struct Atom {
	string atmId;
	vec3<double> fracX, cartX;
};

struct Residue {
	int id;
	string typeLong;
	char typeShort;
	double B;

	vec3<double> fracX, cartX;  // location
	vector<Atom> sidechainAtoms;
	vector<Atom>       allAtoms;
};


class Chain{
	friend bool operator==(Chain& X, Chain& Y);
public:
	Chain() {
		idChar = ' ';
	}

	Chain(char chnId) {
		idChar = chnId;
	}

	void clear() {
		reses.clear();
	}

	void setId(char id) {
		idChar = id;
	}

	void addResidue(Residue R) {
		reses.push_back(R);
	}

	double likelihoodGain(double prior) {
		double gain = 0.0;
		for (int i=0; i<reses.size(); ++i) {
			gain += log(reses[i].B) - prior;
		}
		return gain;
	}

	int nres() { return reses.size(); }
	char id()  { return idChar; }

	int getCoords(double coords[][3], int startAt=0, int nRes=-1) {
		if (nRes == -1) nRes = reses.size();
		int endAt = min(startAt+nRes, (int)reses.size());
		for (int i=startAt; i<endAt; ++i) {
			coords[i-startAt][0] = reses[i].cartX[0];
			coords[i-startAt][1] = reses[i].cartX[1];
			coords[i-startAt][2] = reses[i].cartX[2];
		}

		return (endAt-startAt);
	}

	int getSubchain(Chain& sub, int startAt=0, int nRes=-1) {
		if (nRes == -1) nRes = reses.size();
		int endAt = min(startAt+nRes, (int)reses.size());
		sub.clear();
		for (int i=startAt; i<endAt; ++i) {
			sub.addResidue(reses[i]);
		}

		return (endAt-startAt);
	}

	int firstResId() { return reses.front().id; }
	int lastResId() { return reses.back().id; }

	inline Residue& back() { return reses.back(); }
	inline Residue& operator[](int i) { return reses[i]; }
	inline const Residue& operator[] (int i) const { return reses[i]; }

private:
	vector<Residue> reses;
	char idChar;
};


class Protein {
public:
	Protein(string pdbfile, Crystal &crystData, double MINCHAINDIST, double BFACTORCUTOFF, vec3<int> dims=vec3<int>(0,0,0) ) {
		ifstream pdbin(pdbfile.c_str());
		if (!pdbin.is_open()) {
			cerr << "ERROR: Unable to open PDB file '" << pdbfile << "'\n";
			throw 0;
		}

		string line;
		bool inserted;
		char thisChainId='\0', lastChainId;
		Chain thisChain;
		Residue thisRes;
		Atom thisAtom;
		
		// counters
		int nRes=0, nAtoms=0;

		thisRes.id = -999;
		while(getline(pdbin,line)) {
			if ( (line.substr(0,3) == "TER" ) ) {
				if (thisRes.id != -999) {  // add last residue
					if (thisRes.B >= BFACTORCUTOFF) {
						thisChain.addResidue(thisRes);
						nRes++;
						nAtoms+=thisRes.allAtoms.size();
					}
					thisRes.allAtoms.clear();
					thisRes.sidechainAtoms.clear();
					thisRes.id = -999;
				}
				
				// new chain
				chains.push_back(thisChain);
				srcPtr.push_back( chains.size()-1 );
				thisChain.clear();
			}
			
			if ( (line.substr(0,6) == "CRYST1" ) ) {
				crystData.init(line);
			}
			

			if ((line.substr(0,6) == "ATOM  " || line.substr(0,6) == "HETATM")) {
				if (thisChain.nres() == 0) {
					thisChain.setId(line[21]);  // new chain
				}
					
				int resId  = atoi(line.substr(22,4).c_str());
				if (thisRes.id != -999 && resId != thisRes.id) {

					if (thisRes.B >= BFACTORCUTOFF) {
						thisChain.addResidue(thisRes);
						nRes++;
						nAtoms+=thisRes.allAtoms.size();	
					}
					thisRes.allAtoms.clear();
					thisRes.sidechainAtoms.clear();
					thisRes.id = -999;
				}

				vec3<double> thisCartX(atof(line.substr(30,8)),atof(line.substr(38,8)),atof(line.substr(46,8)));
				vec3<double> thisFracX = thisCartX.apply(crystData.c2f); thisFracX.mod(1);

				if (line.substr(13,2) == "CA") {
					// add a new residue to the current chain
					thisRes.cartX = thisCartX;
					thisRes.fracX = thisFracX;
					thisRes.typeLong = line.substr(17,3);
					thisRes.typeShort = oneLetterCode(thisRes.typeLong);

					thisRes.id = atoi(line.substr(22,4).c_str());  // only set resId when we see a CA

					thisRes.B = atof(line.substr(61,6));   // just store CA's B-factor
					                                       //   (since we often use it to store "confidence")


					// is this a new chain?
					if (thisChain.nres() > 0) {
						if (dims[0]*dims[1]*dims[2] == 0)  {
							if ( (thisRes.cartX - thisChain.back().cartX).norm2() > MINCHAINDIST*MINCHAINDIST ) {
								chains.push_back(thisChain);
								srcPtr.push_back( chains.size()-1 );
								thisChain.clear();
							}
						} else {
							vec3<int>  delIdxXY;
							
							delIdxXY[0] = pos_mod( (int)(floor(thisChain.back().fracX[0]*dims[0]+0.5) 
							                           - floor(thisRes.fracX[0]*dims[0]+0.5)) , dims[0]);
							delIdxXY[1] = pos_mod( (int)(floor(thisChain.back().fracX[1]*dims[1]+0.5) 
							                           - floor(thisRes.fracX[1]*dims[1]+0.5)) , dims[1]);
							delIdxXY[2] = pos_mod( (int)(floor(thisChain.back().fracX[2]*dims[2]+0.5)
							                           - floor(thisRes.fracX[2]*dims[2]+0.5)) , dims[2]);
							
							if ( getAdjPotential(dims , crystData)->get(delIdxXY[0],delIdxXY[1],delIdxXY[2]) < 1e-6 ) {
								chains.push_back(thisChain);
								srcPtr.push_back( chains.size()-1 );
								thisChain.clear();
							}
						}
					}
				}

				thisAtom.cartX = thisCartX;
				thisAtom.fracX = thisFracX;
				thisAtom.atmId = line.substr(12,4);
				thisRes.allAtoms.push_back(thisAtom);
				if (    thisAtom.atmId != " CA " && thisAtom.atmId != " C  "
				     && thisAtom.atmId != " N  " && thisAtom.atmId != " O  "  )
					thisRes.sidechainAtoms.push_back(thisAtom);

			}
		}

		// add last residue ... 
		if (thisRes.id != -999) { 
			if (thisRes.B > BFACTORCUTOFF) {
				thisChain.addResidue(thisRes);
				nRes++;
				nAtoms+=thisRes.allAtoms.size();	
			}
			thisRes.allAtoms.clear();
			thisRes.sidechainAtoms.clear();
			thisRes.id = -999;
		}
		
		// ... and add last chain
		if (thisChain.nres() > 0) {
			chains.push_back( thisChain );
			srcPtr.push_back( chains.size()-1 );
		}
		
		//cerr << "Read " << nAtoms << " atoms in " << nRes << " residues in " << chains.size() << " chains\n";
	}

	void generateSymmetric(Crystal crystData) {
		//
		// make symm copies of chain
		Residue symmRes, *prevSymmRes;
		Chain thisChain;

		int nsymm = crystData.nSymm();
		int nchains = chains.size();
		int natoms;

		//cout << "Generating " << nsymm-1 << " symmetric copies of " << nchains << " chains.\n";
		for (int i=1; i<nsymm; ++i) {   // symmgp1 == identity
			for (int j=0; j<nchains; ++j) {
				thisChain.clear();
				thisChain.setId(chains[j].id());
				for (int k=0; k<chains[j].nres(); ++k) {
					Residue &thisRes = chains[j][k];
					symmRes = thisRes;
					crystData.getSymmetric(thisRes.fracX, symmRes.fracX, i);

					if (k == 0) {
						symmRes.cartX = symmRes.fracX.apply(crystData.f2c);
					} else {
						vec3<double> delta = symmRes.fracX - prevSymmRes->fracX; delta.minmod1();
						symmRes.cartX = prevSymmRes->cartX + delta.apply(crystData.f2c);
					}

					// other atoms
					natoms = thisRes.allAtoms.size();
					for (int l=0; l<natoms; ++l) {
						crystData.getSymmetric(thisRes.allAtoms[l].fracX, symmRes.allAtoms[l].fracX, i);
						vec3<double> delta = symmRes.allAtoms[l].fracX - symmRes.fracX; delta.minmod1();
						symmRes.allAtoms[l].cartX = symmRes.cartX + delta.apply(crystData.f2c);
					}

					// sidechain atoms
					natoms = thisRes.sidechainAtoms.size();
					for (int l=0; l<natoms; ++l) {
						crystData.getSymmetric(thisRes.sidechainAtoms[l].fracX, symmRes.sidechainAtoms[l].fracX, i);
						vec3<double> delta = symmRes.sidechainAtoms[l].fracX - symmRes.fracX; delta.minmod1();
						symmRes.sidechainAtoms[l].cartX = symmRes.cartX + delta.apply(crystData.f2c);
					}

					thisChain.addResidue(symmRes);
					prevSymmRes = &thisChain.back();
				}
				chains.push_back( thisChain );
				srcPtr.push_back( srcPtr[j] );
			}
		}

		//cout << "Read " << chains.size() << " chains\n";
	}

	void generateOriginShifts(Crystal crystData) {
		
	}
	int nChains() { return chains.size(); };

	/////////////////////
	//
	// serialize all copies of all UNIQUE chains
	int nres() {
		int reses = 0;
		for (int i=0; i<chains.size(); i++)
			reses += chains[i].nres();
		return reses;
	}
	//
	/////////////////////

	inline int getSrc(int i) { return srcPtr[i]; }

	inline Chain& operator[](int i) { return chains[i]; }
	inline const Chain& operator[](int i) const { return chains[i]; }

private:
	vector< Chain >  chains;
	vector< int >    srcPtr;
};


//
inline double calcSquaredError(vec3<double> &PredFracX, vec3<double> &ActFracX, Crystal &crystData) {
	vec3<double> delF = (PredFracX - ActFracX); delF.minmod1();
	return (delF.apply(crystData.f2c)).norm2();
}

inline int calcSquaredError(double & sqError, string resId, vector<Atom> &Pred, vector<Atom> &Act, Crystal &crystData) {
	int Npred = Pred.size();
	int Nact = Act.size();
	int nAtoms = 0; // number of aligned atoms
	double thisError;
	sqError = 0;
	
	for (int i=0; i<Npred; ++i) {
		Atom &pred_i = Pred[i];
		string altName;
				
		thisError = INFINITY;
		if ( equivAtom(resId, pred_i.atmId, altName) ) {
			for (int j=0; j<Nact; ++j) {
				if ( Act[j].atmId == pred_i.atmId || Act[j].atmId == altName )
					thisError = min(thisError, calcSquaredError(pred_i.fracX, Act[j].fracX, crystData));
			}
		} else {
			for (int j=0; j<Nact; ++j) {
				if ( Act[j].atmId == pred_i.atmId ) {
					thisError = min(thisError, calcSquaredError(pred_i.fracX, Act[j].fracX, crystData));
					//cerr << "Error " << resId << " " 
					//     << pred_i.fracX.to_string() << " , " << Act[j].fracX.to_string() << " = " 
					//     << calcSquaredError(pred_i.fracX, Act[j].fracX, crystData) << endl;
				}
			}
		}
		
		if (thisError == INFINITY) {
			;//cerr << "WARNING! " << resId << "." <<  pred_i.atmId << " not found!\n";
		} else {
			nAtoms++;
			sqError += thisError;
		}
	}
	
	return nAtoms;
}

//
// aligns Pred to the same cryst copy as Act

vector<Atom> alignTo(Residue &Pred, vec3<double> cartX, Crystal &crystData) {
	vector<Atom> newXs;
	vector< vec3<double> > symm;
	vec3<double> symm_i;
	vec3<double> fracX = cartX.apply( crystData.c2f );
	
	double mindist = INFINITY, dist;
	int bestSymmCopy;
	
	// find best symmcopy (using CA)
	crystData.getSymmetric(Pred.fracX, symm);
	for (int k=0; k<symm.size(); ++k) {
		vec3<double> delf = (symm[k]-fracX); delf.minmod1();
		dist = sqrt(delf.apply(crystData.f2c).norm2());
		if (dist < mindist) {
			mindist = dist;
			bestSymmCopy = k;
		}
	}


	for (int i=0; i< Pred.allAtoms.size(); ++i) {
		Atom pred_i = Pred.allAtoms[i];
		pred_i.fracX.mod(1);
		crystData.getSymmetric(pred_i.fracX, symm_i, bestSymmCopy);
				
		vec3<double> delf = (symm_i-fracX); delf.minmod1();
		pred_i.cartX = cartX + delf.apply(crystData.f2c);
				
		newXs.push_back(pred_i);
	}
		
	return newXs;
}
vector<Atom> alignTo(Residue &Pred, Residue &Act, Crystal &crystData) {
	return alignTo(Pred, Act.cartX, crystData);
}

double calcCA_RMSd(Chain &A, Chain &B, Crystal &crystData) {
	double error = 0.0;
	for (int i = 0; i<A.nres(); i++)
		error += calcSquaredError(A[i].fracX,B[i].fracX,crystData);
	return sqrt(error/A.nres());
}

double calcCA_cover(Chain &A, Chain &B, Crystal &crystData, double maxDist, int &hitCount) {
	hitCount = 0;
	double nearMiss = INFINITY;
	for (int i = 0; i<A.nres(); i++) {
		double thiserror = calcSquaredError(A[i].fracX,B[i].fracX,crystData);
		if (thiserror < maxDist*maxDist) {
			hitCount++;
		} else {
			nearMiss = min(nearMiss , thiserror);
		}
	}
	
	return ( sqrt(nearMiss) );
}


//
//  info about actual ("ground truth") to predicted mapping
class a2p_mapnode {
public:
	a2p_mapnode() {
		CA_sqdist=INFINITY;
		allAtom_sqdist=INFINITY;
		sidechain_sqdist=INFINITY;
		
		nAtoms = nSidechainAtoms = 0;
		predChn=predRes=-1;
		correctRes = false;
	}
	
	double CA_sqdist, allAtom_sqdist, sidechain_sqdist;
	int nAtoms, nSidechainAtoms;
	bool correctRes;
	
	// mapping
	int predChn, predRes;
};

//
//  info about predicted to actual mapping
class p2a_mapnode {
public:
	p2a_mapnode() {
		isMapped = isDuped = false;
		
		actChn = actRes = -1;
	}
	
	bool isMapped, isDuped;
	
	// mapping
	int actChn, actRes;
};

////////////////////////////////
////////////////////////////////
////
//// main
////
int main(int argc, char** argv) {
	double mov_com[3], mov_to_ref[3], U[3][3];

	/////////////////////
	/////////////////////
	///
	/// read input files
	Options bpOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	bpOpts.addOption("t",    "true",      "[REQUIRED] True PDB",      true);
	bpOpts.addOption("p",    "pred",      "[REQUIRED] Predicted PDB",      true);
	bpOpts.addOption("c",    "correct",   "Consider AA correct if CA is < this distance of true (default=2A)",      true);
	bpOpts.addOption("o",    "outpdb",      "Write PDB in correct asymm unit to STDOUT",      false);
	bpOpts.addOption("s",    "stats",      "Write matching statistics to STDOUT",      false);
	bpOpts.addOption("n",    "nobreak",      "Don't try to break chains",      false);
	bpOpts.addOption("v",    "verbose",      "Use verbose output",      false);
		
	string truePDB, predPDB;
	double minChainDist = 4.3, grid = 0.0, CORRECT_WITHIN = 2.0;
	
	vec3<int> dims(0,0,0);
	bool verbose = false, write_pdb = false, DONT_BREAK = false, PRINTSTATS = false;
	
	bpOpts.parse(argc, argv);
	while ((switchInt = bpOpts.cycle()) >= 0) {
		vector<string> pargs;
		
		switch(switchInt) {
		case 0:   // Option 'true'
			truePDB =  bpOpts.getArgs(switchInt); break;
		case 1:   // Option 'pred'
			predPDB =  bpOpts.getArgs(switchInt); break;
		case 2:   // Option 'correct'
			CORRECT_WITHIN =  atof( bpOpts.getArgs(switchInt) ); break;
		case 3:   // Option 'out-pdb'
			write_pdb = true; break;
		case 4:   // Option 'stats'
			PRINTSTATS = true; break;
		case 5:   // Option 'nobreak'
			DONT_BREAK = true; break;
		case 6:   // Option 'verbose'
			verbose = true; break;
		}
	}

	Crystal crystData;
	Protein actual(truePDB,crystData,4.3,-INFINITY);
	Protein pred  (predPDB,crystData,4.3,-INFINITY);
		
	int realNumChns = actual.nChains();
	
	//  GENERATE symm copies of actual
	actual.generateSymmetric(crystData);
	
	//
	// MAPPING pred --> actual
	vector< vector<p2a_mapnode> > p2a_nobreak, p2a_break;
	for (int i_chn = 0; i_chn < pred.nChains(); ++i_chn) {
		p2a_nobreak.push_back (  vector<p2a_mapnode>(pred[i_chn].nres()) );
		p2a_break.push_back   (  vector<p2a_mapnode>(pred[i_chn].nres()) );
	}
	//
	// MAPPING actual --> pred
	vector< vector<a2p_mapnode> > a2p_nobreak, a2p_break;
	for (int j_chn = 0; j_chn < actual.nChains(); ++j_chn) {
		a2p_nobreak.push_back (  vector<a2p_mapnode>(actual[j_chn].nres()) );
		a2p_break.push_back   (  vector<a2p_mapnode>(actual[j_chn].nres()) );
	}
	

	// align each predicted CHAIN to best subseq in original protein
	int atomidx=0, nchains=0, nResCtr=0, nAtmCtr=0;
	Chain actual_j, pred_i;
	bool needChainBreak = false;
	for (int i_chn = 0; i_chn < pred.nChains(); ++i_chn) {
		int nPredReses = pred[i_chn].nres();
		
		if (nPredReses == 0) continue; /// bug in parser .. will fix
		
		// align this PREDICTED chain with every possible subsequence
		//     compute only CA RMS to do this quicker
		int best_cover=0, this_cover;
		double tiebreaker_rmsd, best_tiebreaker_rmsd = INFINITY;
		
		int bestChn=-1, bestActStart=-1, bestPredStart=-1,
		                bestActEnd=-1,   bestPredEnd=-1, bestCommonReses;
						
		int actStart,actEnd,predStart,predEnd;
		
		for (int j_chn = 0; j_chn < actual.nChains(); ++j_chn) {
			int nActualReses, nCommonReses;
			nActualReses = actual[j_chn].nres();
			
			// loop over predicted starting locations
			for (int j_res = -nPredReses+1; j_res < actual[j_chn].nres(); ++j_res) {
			
				// compute overlapping fragments
				actStart = (j_res < 0) ? 0:j_res;
				actEnd   = (j_res+nPredReses >= nActualReses) ? nActualReses-1 : j_res+nPredReses-1;
				
				nCommonReses = actEnd-actStart+1;
				
				predStart = (j_res < 0) ? -j_res : 0 ;
				predEnd   = predStart + nCommonReses - 1;
				
				pred[i_chn].getSubchain(pred_i, predStart, nCommonReses);
				actual[j_chn].getSubchain(actual_j, actStart, nCommonReses);
				
				if ( actual_j.nres() == pred_i.nres() ) {  // make certain all residues were legally defined
				
					tiebreaker_rmsd = calcCA_cover(pred_i, actual_j, crystData, CORRECT_WITHIN, this_cover);
					//cerr << tiebreaker_rmsd << " , " << this_cover << ".";
					
					if (    best_cover < this_cover 
					     || ( best_cover == 0 && tiebreaker_rmsd < best_tiebreaker_rmsd ) ) {
						best_tiebreaker_rmsd = tiebreaker_rmsd;
						best_cover = this_cover; bestChn  = j_chn;
						
						bestActStart  = actStart;  bestActEnd = actEnd;
						bestPredStart = predStart; bestPredEnd = predEnd;
						bestCommonReses = nCommonReses;
					}
				} else {
					cerr << "ERROR: predicted and actual chain mismatch!\n";
				}
			}
		} 

		if (verbose) {
			cerr << "ALIGNING predicted[" << pred[i_chn].id() << "]("     << pred[i_chn][bestPredStart ].id << ":" 
			                                                              << pred[i_chn][bestPredEnd].id << ") to "
			                 << "actual[" << actual[bestChn].id() << "](" << actual[bestChn][bestActStart].id << ":" 
							                                              << actual[bestChn][bestActEnd].id << ")"
			                 << " with matched-residues = " << best_cover ;
			if (best_cover == 0)
				cerr << "  (closest = " << best_tiebreaker_rmsd << ")";
			cerr << endl;
		}
		
		//
		//  compute res-by-res squared error
		//     compute CA error, allAtom error, sidechain error
		int predChn = i_chn, actChn = bestChn;  // consistent naming
		for (int j_res=0; j_res < bestCommonReses; ++j_res) {
			int actChnSrc = actual.getSrc(bestChn);
			int actRes  = bestActStart+j_res;
			int predRes = bestPredStart+j_res;
			
			double thisDist = calcSquaredError(actual[actChn][actRes].fracX, pred[predChn][predRes].fracX, crystData);
			
			if (PRINTSTATS)
				cout << pred[predChn][predRes].B << "," << sqrt(thisDist) << endl;
			
			if (thisDist <= CORRECT_WITHIN*CORRECT_WITHIN) {
				double allAtomDist, sidechainDist;
				
				// resolve "collisions" -- that is, two residues in the same location -- 
				//    by throwing out the shorter chain
				if (a2p_nobreak[actChnSrc][actRes].predChn >= 0) {
					// collision!! there is already something mapped to this
					//    find the collision
					int collidePredChn = a2p_nobreak[actChnSrc][actRes].predChn;
					int collidePredRes = a2p_nobreak[actChnSrc][actRes].predRes;
					
					// keep the longer chain
					if ( pred[collidePredChn].nres() < pred[predChn].nres() ) {
						// overwrite old mapping
						string &resId  = pred[predChn][predRes].typeLong;
						int nAtom      = calcSquaredError(allAtomDist, resId,
						                                  actual[actChn][actRes].allAtoms, 
						                                  pred[predChn][predRes].allAtoms, crystData);
						int nSidechain = calcSquaredError(sidechainDist, resId,
						                                  actual[actChn][actRes].sidechainAtoms, 
						                                  pred[predChn][predRes].sidechainAtoms, crystData);
								  
						a2p_nobreak[actChnSrc][actRes].CA_sqdist        = thisDist;
						a2p_nobreak[actChnSrc][actRes].allAtom_sqdist   = allAtomDist;
						a2p_nobreak[actChnSrc][actRes].sidechain_sqdist = sidechainDist;
						a2p_nobreak[actChnSrc][actRes].nAtoms            = nAtom;
						a2p_nobreak[actChnSrc][actRes].nSidechainAtoms   = nSidechain;
						
						if (actual[actChnSrc][actRes].typeShort == pred[predChn][predRes].typeShort)
							a2p_nobreak[actChnSrc][actRes].correctRes = true;
						else
							a2p_nobreak[actChnSrc][actRes].correctRes = false;
						
						// update mapping, mark collision
						a2p_nobreak[actChnSrc][actRes].predChn = predChn;
						a2p_nobreak[actChnSrc][actRes].predRes = predRes;
						
						p2a_nobreak[predChn][predRes].actChn = actChn;
						p2a_nobreak[predChn][predRes].actRes = actRes;
						p2a_nobreak[predChn][predRes].isMapped = true;
						p2a_nobreak[collidePredChn][collidePredRes].isDuped = true;
						
					} else {
						// keep old map, mark collision
						p2a_nobreak[predChn][predRes].actChn = actChn;
						p2a_nobreak[predChn][predRes].actRes = actRes;
						p2a_nobreak[predChn][predRes].isMapped = true;
						p2a_nobreak[predChn][predRes].isDuped = true;
					}
					
					if (verbose) {
						cerr << "COLLISION! predicted[" 
						<< pred[predChn].id()        << "]("  << pred[predChn][predRes].id 
						                  <<   ") and [" << pred[collidePredChn].id() << "]("  
										  << pred[collidePredChn][collidePredRes].id 
			                              << ") map to "
			                              <<  "actual[" << actual[actChn].id()       << "](" 
										  << actual[actChn][actRes].id << ")\n";
					}
				} else {
					string &resId  = pred[predChn][predRes].typeLong;
					int nAtom      = calcSquaredError(allAtomDist, resId,
														actual[actChn][actRes].allAtoms, 
														pred[predChn][predRes].allAtoms, crystData);
					int nSidechain = calcSquaredError(sidechainDist, resId,
														actual[actChn][actRes].sidechainAtoms, 
														pred[predChn][predRes].sidechainAtoms, crystData);
								
					a2p_nobreak[actChnSrc][actRes].CA_sqdist        = thisDist;
					a2p_nobreak[actChnSrc][actRes].allAtom_sqdist   = allAtomDist;
					a2p_nobreak[actChnSrc][actRes].sidechain_sqdist = sidechainDist;
					a2p_nobreak[actChnSrc][actRes].nAtoms            = nAtom;
					a2p_nobreak[actChnSrc][actRes].nSidechainAtoms   = nSidechain;
					
					if (actual[actChnSrc][actRes].typeShort == pred[predChn][predRes].typeShort)
						a2p_nobreak[actChnSrc][actRes].correctRes = true;
					else
						a2p_nobreak[actChnSrc][actRes].correctRes = false;
					
					// update mapping, mark collision
					a2p_nobreak[actChnSrc][actRes].predChn = predChn;
					a2p_nobreak[actChnSrc][actRes].predRes = predRes;
					
					p2a_nobreak[predChn][predRes].actChn = actChn;
					p2a_nobreak[predChn][predRes].actRes = actRes;
					p2a_nobreak[predChn][predRes].isMapped = true;
				}
			} 
		}
		nchains++;
		
		//
		//  compute res-by-res squared error WITH chain breaks 
		for (int i_res=0; i_res < pred[i_chn].nres(); ++i_res) {
			
			double minBreakDist = INFINITY;
			int minBreakChn, minBreakRes;
			// compare this res to EVERY pred res in every chain
			for (int j_chn = 0; j_chn < actual.nChains(); ++j_chn) {

				for (int j_res = 0; j_res < actual[j_chn].nres(); ++j_res) {
					double thisDist = calcSquaredError(actual[j_chn][j_res].fracX,
					                                   pred[i_chn][i_res].fracX, crystData);
					
					if (thisDist < minBreakDist) {
						minBreakDist = thisDist;
						minBreakChn = j_chn;
						minBreakRes = j_res;
					}
				}
			}
			
			if (minBreakDist <= CORRECT_WITHIN*CORRECT_WITHIN) {
				double allAtomDist, sidechainDist;
				int j_chnSrc = actual.getSrc(minBreakChn);
				
				string &resId  = pred[i_chn][i_res].typeLong;
				int nAtom      = calcSquaredError(allAtomDist, resId,
													actual[minBreakChn][minBreakRes].allAtoms, 
													pred[i_chn][i_res].allAtoms, crystData);
				int nSidechain = calcSquaredError(sidechainDist, resId,
													actual[minBreakChn][minBreakRes].sidechainAtoms, 
													pred[i_chn][i_res].sidechainAtoms, crystData);
							
				a2p_break[j_chnSrc][minBreakRes].CA_sqdist        = minBreakDist;
				a2p_break[j_chnSrc][minBreakRes].allAtom_sqdist   = allAtomDist;
				a2p_break[j_chnSrc][minBreakRes].sidechain_sqdist = sidechainDist;
				a2p_break[j_chnSrc][minBreakRes].nAtoms            = nAtom;
				a2p_break[j_chnSrc][minBreakRes].nSidechainAtoms   = nSidechain;
				
				if (actual[j_chnSrc][minBreakRes].typeShort == pred[i_chn][i_res].typeShort)
					a2p_break[j_chnSrc][minBreakRes].correctRes = true;
				else
					a2p_break[j_chnSrc][minBreakRes].correctRes = false;
				
				// update mapping, mark collision
				a2p_break[j_chnSrc][minBreakRes].predChn = i_chn;
				a2p_break[j_chnSrc][minBreakRes].predRes = i_res;
				
				p2a_break[i_chn][i_res].actChn = minBreakChn;
				p2a_break[i_chn][i_res].actRes = minBreakRes;
				p2a_break[i_chn][i_res].isMapped = true;
				
				if (    p2a_break[i_chn][i_res].actChn != p2a_nobreak[i_chn][i_res].actChn ||
						p2a_break[i_chn][i_res].actRes != p2a_nobreak[i_chn][i_res].actRes) 
						needChainBreak = true;
			}
		}

		
		//
		// now select predicted fragment copy that matches actual[0], print to stdout
		if (write_pdb) {
			cout << setiosflags(ios_base::right|ios::fixed) << setprecision(3);
			vector< vector<Atom> > allAtoms_aligned(nPredReses);
			for (int j_res=0; j_res < bestCommonReses; ++j_res) {
				allAtoms_aligned[bestPredStart+j_res] = 
						alignTo( pred[i_chn][bestPredStart+j_res] , 
									actual[actual.getSrc(bestChn)][bestActStart+j_res], 
									crystData);
			}
			// before
			for (int j_res=bestPredStart-1; j_res >= 0; --j_res) {
				allAtoms_aligned[j_res] = 
					alignTo( pred[i_chn][j_res] , allAtoms_aligned[j_res+1][0].cartX , crystData);
			}
			// after
			for (int j_res=bestPredEnd+1; j_res < nPredReses; ++j_res) {
				allAtoms_aligned[j_res] = 
					alignTo( pred[i_chn][j_res] , allAtoms_aligned[j_res-1][0].cartX , crystData);
			}
			
			for (int j_res=0; j_res < nPredReses; ++j_res) {
				++nResCtr;
				for (int k=0; k<allAtoms_aligned[j_res].size(); ++k) {
					cout << "ATOM  " << setw(5) << ++nAtmCtr << " " << allAtoms_aligned[j_res][k].atmId << " " << pred[i_chn][j_res].typeLong 
						<< " " << (char)('A'+i_chn%26)
						<< setw(4) << nResCtr << "    ";
					cout << setw(8) << allAtoms_aligned[j_res][k].cartX[0]
						 << setw(8) << allAtoms_aligned[j_res][k].cartX[1]
						 << setw(8) << allAtoms_aligned[j_res][k].cartX[2];
					cout << "  1.00 " << pred[i_chn][j_res].B << endl;
	
				}
			}
			cout << "TER   \n"; ++nResCtr;
		}
	}

	// final composite stats
	int nFound=0, nAtomSum=0, nCorrectId=0, nSidechainAtomSum=0, nRes = 0;
	int  nPredRes = 0;
	double CA_RMSd=0.0, allAtom_RMSd=0.0, sidechain_RMSd=0.0;

	// loop actual structure
	for (int j_chn = 0; j_chn < actual.nChains(); ++j_chn) {
		if (actual.getSrc(j_chn) != j_chn) continue;  // this chain is a copy of some other
		nRes += actual[j_chn].nres();

		for (int j_res = 0; j_res < actual[j_chn].nres(); ++j_res) {
			if ( a2p_nobreak[j_chn][j_res].predChn >= 0 ) {
				nFound++;

				CA_RMSd        +=  a2p_nobreak[j_chn][j_res].CA_sqdist;
				allAtom_RMSd   +=  a2p_nobreak[j_chn][j_res].allAtom_sqdist;
				sidechain_RMSd +=  a2p_nobreak[j_chn][j_res].sidechain_sqdist;
				
				nAtomSum          += a2p_nobreak[j_chn][j_res].nAtoms;
				nSidechainAtomSum += a2p_nobreak[j_chn][j_res].nSidechainAtoms;

				if (a2p_nobreak[j_chn][j_res].correctRes) nCorrectId++;
			}
		}
	}
	
		
	// loop through predicted structure
	for (int i_chn = 0; i_chn < pred.nChains(); ++i_chn) {
		nPredRes += pred[i_chn].nres();
	}

	if (nFound == 0) {
		cerr << "ALIGNMENT ERROR! " << endl;
		cerr << "   No acceptable fragments found" << endl;
		exit(1);
	}
	

	// compact
	if (!verbose) {
		cerr << setprecision(3);
		cerr << nchains << " , " << 100*((double)nFound)/nRes 
		     << " , " << 100*((double)nCorrectId)/nRes 
			 << " ,, " << 100*((double)nFound)/nPredRes 
		     << " , " << 100*((double)nCorrectId)/nPredRes;
			
			
		cerr << setprecision(4);
		cerr << " ,, " << sqrt(CA_RMSd/nFound)
		     << " , "  << sqrt(allAtom_RMSd/nAtomSum)
		     << " , ";
		if (nSidechainAtomSum > 0) 
			cerr << sqrt(sidechain_RMSd/nSidechainAtomSum);
		else
			cerr << " na";
	
	} else {	
		cerr << setprecision(3);
		cerr << endl;
		cerr << "Percent true structure located <" << CORRECT_WITHIN << "A: " << 100*((double)nFound)/nRes << endl;
		cerr << "Percent true structure correctly identified <" << CORRECT_WITHIN << "A: " << 100*((double)nCorrectId)/nRes << endl;
		
		cerr << endl ;
		cerr << "Percent predicted structure correct <" << CORRECT_WITHIN << "A: " << 100*((double)nFound)/nPredRes << endl;
		cerr << "Percent predicted structure correctly identified <" << CORRECT_WITHIN << "A: " << 100*((double)nCorrectId)/nPredRes << endl;
		
		cerr << setprecision(4);
		cerr << endl;
		cerr << "CA RMS Error: " << sqrt(CA_RMSd/nFound) << endl;
		cerr << "All-atom RMS Error: " << sqrt(allAtom_RMSd/nAtomSum) << endl;
		cerr << "Sidechain RMS Error: ";
		if (nSidechainAtomSum > 0) 
			cerr << sqrt(sidechain_RMSd/nSidechainAtomSum);
		else
			cerr << " na";
			
		
		cerr << endl;
		cerr << "Predicted chains: " << nchains << endl;
	}
	
	// if there are missconnected domains ...
	if (needChainBreak && !DONT_BREAK) {
		// final composite stats
		nFound = nAtomSum = nCorrectId = nSidechainAtomSum = nRes = 0;
		CA_RMSd = allAtom_RMSd = sidechain_RMSd = 0.0;

		if (verbose) {
			cerr << endl << endl;
			cerr << "*** ALLOWING CHAIN BREAKS ***" << endl;
		}
	
		// loop through actual structure
		for (int j_chn = 0; j_chn < actual.nChains(); ++j_chn) {
			if (actual.getSrc(j_chn) != j_chn) continue;  // this chain is a copy of some other
			nRes += actual[j_chn].nres();
	
			for (int j_res = 0; j_res < actual[j_chn].nres(); ++j_res) {
				if ( a2p_break[j_chn][j_res].predChn >= 0 ) {
					nFound++;
	
					CA_RMSd        +=  a2p_break[j_chn][j_res].CA_sqdist;
					allAtom_RMSd   +=  a2p_break[j_chn][j_res].allAtom_sqdist;
					sidechain_RMSd +=  a2p_break[j_chn][j_res].sidechain_sqdist;
					
					nAtomSum          += a2p_break[j_chn][j_res].nAtoms;
					nSidechainAtomSum += a2p_break[j_chn][j_res].nSidechainAtoms;
	
					if (a2p_break[j_chn][j_res].correctRes) nCorrectId++;
				}
			}
		}
		
		// loop through predicted structure
		if (verbose) {
			for (int i_chn = 0; i_chn < pred.nChains(); ++i_chn) {
				for (int i_res = 1; i_res < pred[i_chn].nres(); ++i_res) {
					if ( !p2a_break[i_chn][i_res].isDuped && p2a_break[i_chn][i_res].isMapped && 
					     !p2a_break[i_chn][i_res-1].isDuped && p2a_break[i_chn][i_res-1].isMapped &&
					      ( p2a_break[i_chn][i_res].actChn != p2a_break[i_chn][i_res-1].actChn ||
					        p2a_break[i_chn][i_res].actRes != p2a_break[i_chn][i_res-1].actRes+1 ) )
							cerr << "BREAKING false connection between " 
								<< "predicted[" << pred[i_chn].id() << "]("  << pred[i_chn][i_res-1].id 
								<<     ") and [" << pred[i_chn].id() << "]("  << pred[i_chn][i_res].id 
								<< ")\n";
				}
			}
		}
	
		if (nFound == 0) {
			cerr << "ALIGNMENT ERROR! " << endl;
			cerr << "   No acceptable fragments found" << endl;
			exit(1);
		}
		
	
		// compact
		if (!verbose) {
			cerr << setprecision(3);
			cerr << " ,,, " << 100*((double)nFound)/nRes 
			     << " , " << 100*((double)nCorrectId)/nRes 
				 << " ,, " << 100*((double)nFound)/nPredRes 
			     << " , " << 100*((double)nCorrectId)/nPredRes;
			
			cerr << " ,, " << sqrt(CA_RMSd/nFound)
				<< " , "  << sqrt(allAtom_RMSd/nAtomSum)
				<< " , ";
			if (nSidechainAtomSum > 0) 
				cerr << sqrt(sidechain_RMSd/nSidechainAtomSum);
			else
				cerr << " na";
		} else {	
			cerr << setprecision(3);
			cerr << endl;
			cerr << "Percent true structure located <" << CORRECT_WITHIN << "A: " << 100*((double)nFound)/nRes << endl;
			cerr << "Percent true structure correctly identified <" << CORRECT_WITHIN << "A: " << 100*((double)nCorrectId)/nRes << endl;
			
			cerr << endl ;
			cerr << "Percent predicted structure correct <" << CORRECT_WITHIN << "A: " << 100*((double)nFound)/nPredRes << endl;
			cerr << "Percent predicted structure correctly identified <" << CORRECT_WITHIN << "A: " << 100*((double)nCorrectId)/nPredRes << endl;
	
			cerr << setprecision(4);
			cerr << endl;
			cerr << "CA RMS Error: " << sqrt(CA_RMSd/nFound) << endl;
			cerr << "All-atom RMS Error: " << sqrt(allAtom_RMSd/nAtomSum) << endl;
			cerr << "Sidechain RMS Error: ";
			if (nSidechainAtomSum > 0) 
				cerr << sqrt(sidechain_RMSd/nSidechainAtomSum);
			else
				cerr << " na";
		}
	}
	cerr << endl;
}
