// //////////////////////////////////////////////////////////////////////////
//
//     protein_fk.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//   portions based on the program 'angles2xyz' in the Yakusa package
//   see http://bioserv.rpbs.jussieu.fr/Yakusa/download/index.html
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
// 
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
// 
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ////////////////////////////////////////////////////////////////////////////


#ifndef _PROTEIN_FK_H
#define _PROTEIN_FK_H

#include "prob_dist.h"
#include "vec3.h"

using namespace std;

// canonical distances
const double DIST_CA_C = 1.53;
const double DIST_C_N  = 1.32;
const double DIST_N_CA = 1.46;
const double DIST_C_O  = 1.23;
		
// canonical angles
const double ANG_C_N_CA  = 121.5;
const double ANG_N_CA_C  = 111.0;
const double ANG_CA_C_N  = 117.0;
const double ANG_CA_C_O  = 121.0;
const double ANG_N_C_O   = 123.0;

const double DEG2RAD = 0.01745329252;

// phi-psi histograms
extern double PHIPSI_STD[][36];
extern double PHIPSI_STD_rowsums[];
extern double PHIPSI_GLY[][36];
extern double PHIPSI_GLY_rowsums[];
extern double PHIPSI_cPRO[][36];
extern double PHIPSI_cPRO_rowsums[];
extern double PHIPSI_tPRO[][36];
extern double PHIPSI_tPRO_rowsums[];

// omega distributions
extern double PROB_cis[];
extern double cisOMEGA_std;
extern double transOMEGA_std;


//
//
double get_dihedral(vec3<double> A, vec3<double> B, vec3<double> C, vec3<double> D) {
	double	xa, ya, za, px1, py1, pz1,
		xb, yb, zb, px2, py2, pz2,
		xc, yc, zc, px3, py3, pz3,
		np1np2, dihed;

	xa = B[0] - A[0];
	ya = B[1] - A[1];
	za = B[2] - A[2];
	
	xb = C[0] - B[0];
	yb = C[1] - B[1];
	zb = C[2] - B[2];
	
	xc = D[0] - C[0];
	yc = D[1] - C[1];
	zc = D[2] - C[2];
	
	px1 =   ya*zb - yb*za;
	py1 = - xa*zb + xb*za;
	pz1 =   xa*yb - xb*ya;
	
	px2 =   yb*zc - yc*zb;
	py2 = - xb*zc + xc*zb;
	pz2 =   xb*yc - xc*yb;
	
	np1np2 = sqrt((px1*px1+py1*py1+pz1*pz1) * (px2*px2+py2*py2+pz2*pz2));
	
	if (np1np2 == 0) 
		return -INFINITY;   // AB or CD parallel to BC
	
	dihed = acos((px1*px2+py1*py2+pz1*pz2)/np1np2);
	
	px3 =   py1*pz2 - py2*pz1;
	py3 = - px1*pz2 + px2*pz1;
	pz3 =   px1*py2 - px2*py1;
	
	if ((px3*xb + py3*yb + pz3*zb) <  0)		// sine is < 0
		dihed = -dihed;
	
	dihed = dihed * 180.0L / M_PI;			// rad -> deg
	if (dihed >= 180.0L)			// [-180.0,180.0[ range
		dihed = -180.0L;
	
	return dihed;

}


// samples from the space of allowable pseudoangles formed from three consecutive CA's
//    - use a two-gaussian estimate
//         p(angle==x) = 0.5 * N(x;92,3.5) + 0.5 * N(x;120,13)
double sample_CACACA_pseudoangle() {
	if (rand()%2 == 0) 
		return ( 92.0 + 3.5*randnorm() );
	else
		return (120.0 + 13.*randnorm() );
}



//
//  get_coord_from_angle
//
//       Find the next atom, given the preceding three:
//          c0, c1, c2 -  the coords of the three preceding atoms
//          c3         -  the new coordinate to calculate 
//          alpha      -  the dihedral c0-c1-c2-c3 
//          tau        -  the angle c1-c2-c3
//          normw      -  the distance c2-c3 
//
void get_coord_from_angle(
                  const vec3<double> &c0, 
                  const vec3<double> &c1, 
				  const vec3<double> &c2,
                  vec3<double> &c3, 
				  double alpha, double tau, double normw)
{
	double u1, u2, u3, v1, v2, v3,
	       norm, pvuv1, pvuv2, pvuv3,
	       pvvuv1, pvvuv2, pvvuv3,
	       nsa, nca, nct;
	
	u1 = c1[0] - c0[0];
	u2 = c1[1] - c0[1];
	u3 = c1[2] - c0[2];
	
	v1 = c2[0] - c1[0];
	v2 = c2[1] - c1[1];
	v3 = c2[2] - c1[2];

	// normalize v
	norm = sqrt(v1*v1+v2*v2+v3*v3);
	v1 /=  norm;
	v2 /=  norm;
	v3 /=  norm;

	// pvuv: cross-product (u*v)
	pvuv1 = u2*v3 - u3*v2;
	pvuv2 = u3*v1 - u1*v3;
	pvuv3 = u1*v2 - u2*v1;

	// normalizen pvuv
	norm = sqrt(pvuv1*pvuv1+pvuv2*pvuv2+pvuv3*pvuv3);
	pvuv1 /= norm;
	pvuv2 /= norm;
	pvuv3 /= norm;

	// pvvuv: cross-product u*(u*v)
	pvvuv1 = v2*pvuv3 - v3*pvuv2;
	pvvuv2 = v3*pvuv1 - v1*pvuv3;
	pvvuv3 = v1*pvuv2 - v2*pvuv1;
   			
	// normalize pvvuv
	norm = sqrt(pvvuv1*pvvuv1+pvvuv2*pvvuv2+pvvuv3*pvvuv3);
	pvvuv1 /= norm;
	pvvuv2 /= norm;
	pvvuv3 /= norm;

	nca = cos(DEG2RAD * alpha);
	nsa = sin(DEG2RAD * alpha);
	nct = tan(DEG2RAD * (tau-90));
	
	u1 = nca*(-pvvuv1) + nsa*pvuv1 + v1*nct;
	u2 = nca*(-pvvuv2) + nsa*pvuv2 + v2*nct;
	u3 = nca*(-pvvuv3) + nsa*pvuv3 + v3*nct;

	// renormalize s.t. norm(u) == normw
	norm = sqrt(u1*u1+u2*u2+u3*u3);
	u1 = u1*normw/norm;
	u2 = u2*normw/norm;
	u3 = u3*normw/norm;
	
	// c3 <- c2+u
	c3[0] = c2[0]+u1;
	c3[1] = c2[1]+u2;
	c3[2] = c2[2]+u3;
}

#endif

